/*
 * Moonlight|3D Copyright (c) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Mar 25, 2006
 */
package eu.moonlight3d.common.document.extensible;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * A document contribution that is a part of an extensible document.
 * Implementations of this interface form parts of the document
 * which contain information that is stored as a part of the
 * extensible document file.
 *
 * @author gregor
 */
public interface DocumentContribution {
	/**
	 * Returns a unique name describing this contribution. This name
	 * is used to refer to that particular contribution in various
	 * cases.
	 * 
	 * @return name of the contribution
	 */
	public String getName();
	
	/**
	 * Decide if the data stored in this document contribution should
	 * be retained as part of the document if it is loaded by a version 
	 * of the program that is unable to handle it (e.g. because of 
	 * missing plugins). If the data should be retained, it is saved
	 * verbatim in each new copy of the file. If it should not be retained
	 * in that case, it is discarded immediately when it is discovered
	 * while loading the file.
	 * 
	 * When a contribution's data is retained in the file, that data
	 * may be out of sync with the other data in the file and the provider
	 * of the contribution must be prepared to handle this. Depending on
	 * the type and importance of the data in that contribution it may
	 * be easier to ask for data to be discarded rather than retained. 
	 * 
	 * @return true if the data should be retained in absence of the
	 * provider of this contribution, false otherwise
	 */
	public boolean retainAlways();
	
	/**
	 * Write the data that this document contribution. The contribution
	 *  must save it into the XML DOM subtree given by contributionRoot.
	 * 
	 * @param document the document that the contributionRoot element 
	 * belongs to
	 * @param contributionRoot the root node for the data that is to be
	 * stored.
	 * @throws Exception 
	 */
	public void write(Document document, Element contributionRoot) throws Exception;
	
	/**
	 * Check if this contribution has unsaved changes. This is used
	 * to determine if the document as a whole has unsaved changes
	 * at this moment.
	 * 
	 * @return true if there are unsaved changes, false otherwise
	 */
	public boolean hasUnsavedChanges();
	
	/**
	 * Request an update of this contribution so that it can bring
	 * itself into a consistent state again after changes have 
	 * been performed on the document.
	 */
	public void update();
	
	/**
	 * Handle all cleanup that is required when the document is closed.
	 * This method gets called automatically when the document that this
	 * instance of a contribution belongs to is closed.
	 */
	public void close();
}
