/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Oct 21, 2008
 */

package eu.moonlight3d.common.document.extensible;

import java.io.File;
import java.text.MessageFormat;
import java.util.ArrayList;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.document.Document;
import eu.moonlight3d.framework.helper.PropertyContainer;
import eu.moonlight3d.framework.localization.L10N;

public class ExtensibleDocument implements Document {

	private ExtensibleDocumentFactory extensibleDocumentFactory;
	
	private ArrayList<DocumentContribution> contributions;
	private ArrayList<UnhandledContribution> unhandledContributions;
	private String typeName;
	private String filename;
	
	protected ExtensibleDocument(String typeName, ExtensibleDocumentFactory factory) {
		this.typeName=typeName;
		this.extensibleDocumentFactory=factory;
		this.contributions=new ArrayList<DocumentContribution>();
		this.unhandledContributions=new ArrayList<UnhandledContribution>();
	}
	
	protected ExtensibleDocument(String typeName, ExtensibleDocumentFactory factory, String filename) throws Exception {
		this.typeName=typeName;
		this.extensibleDocumentFactory=factory;
		this.contributions=new ArrayList<DocumentContribution>();
		this.unhandledContributions=new ArrayList<UnhandledContribution>();
		open(filename);
	}
	
	public void close() {
		for(DocumentContribution contribution : contributions) {
			contribution.close();
		}

	}

	public String getFilename() {
		return filename;
	}

	public String getTypeName() {
		return typeName;
	}

	public boolean hasUnsavedChanges() {
		for(DocumentContribution contribution : contributions) {
			if(contribution.hasUnsavedChanges()) {
				return true;
			}
		}
		return false;
	}

	public void initialize(PropertyContainer documentOptions) {
		for(DocumentContributionFactory contributionFactory : extensibleDocumentFactory.getDocumentContributionFactories()) {
			DocumentContribution contribution=contributionFactory.getNewContribution(this, documentOptions);
			contributions.add(contribution);
		}
	}
	
	private void open(String filename) throws Exception {
		this.filename=filename;
		
    	String uri=new File(filename).toURI().toString();
        org.w3c.dom.Document document=DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(uri);
        
        Element root=document.getDocumentElement();
        if(!root.getTagName().equals("extensibledoc")) {
        	throw new Exception(L10N.translate("file is not an extensible document: root tag is not extensibledoc"));
        }
        if(!root.getAttribute("fileformatversion").equals("1")) {
            throw new Exception(L10N.translate("file appears to be from an unsupported version of the program"));
        }
        if(!root.getAttribute("type").equals(typeName)) {
        	throw new Exception(L10N.translate("file does not have the correct type"));
        }
        NodeList contributionNodes=root.getElementsByTagName("contribution");


        //  read contributions
		//  TODO add as many sanity checks as possible
		for(int i=0;i<contributionNodes.getLength();i++) {
			Element contributionRoot=(Element)contributionNodes.item(i);
	
			String contributionName=contributionRoot.getAttribute("name");
	
			boolean handledContribution=false;
			for(DocumentContributionFactory factory : extensibleDocumentFactory.getDocumentContributionFactories()) {
				if(contributionName.equals(factory.getName())) {
					DocumentContribution contribution=factory.readContribution(this,document,contributionRoot);
					contributions.add(contribution);
					handledContribution=true;
					break;
				}
			}
			if(!handledContribution) {
				// check if contribution should be discarded if it stays unhandled
				boolean retain=Boolean.parseBoolean(contributionRoot.getAttribute("retain"));
				
				State state=State.getInstance();
				if(retain) {
					// retain data of contribution and warn
					UnhandledContribution unhandledContribution=new UnhandledContribution(contributionRoot);
					unhandledContributions.add(unhandledContribution);
					state.logger().warning(MessageFormat.format(L10N.translate("cannot load document contribution {0} because of missing provider; will retain affected data unchanged"),contributionName));
				} else {
					// don't retain it and warn
					state.logger().warning(MessageFormat.format(L10N.translate("cannot load document contribution {0}, discarded"),contributionName));					
				}
			}
		}
	
	}

	public void save(String filename) throws Exception {
		this.filename=filename;
		
    	org.w3c.dom.Document compositeDoc=DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
        
        // generate save file DOM tree here
        Element root=compositeDoc.createElement("extensibledoc");
        root.setAttribute("fileformatversion","1");
        root.setAttribute("type",typeName);
        compositeDoc.appendChild(root);
        
        
        for(DocumentContribution contribution : contributions) {
        	Element contributionRoot=compositeDoc.createElement("contribution");
        	contributionRoot.setAttribute("name",contribution.getName());
        	contributionRoot.setAttribute("retain",Boolean.toString(contribution.retainAlways()));
            // TODO write out contribution file format version number
        	contribution.write(compositeDoc, contributionRoot);
        	root.appendChild(contributionRoot);
        }

        for(UnhandledContribution unhandledContribution : unhandledContributions) {
        	unhandledContribution.write(root);
        }

        // set up properly formatted file output
        Source domSource=new DOMSource(compositeDoc);
        Result streamResult=new StreamResult(new File(filename));
        Transformer transformer=TransformerFactory.newInstance().newTransformer();
        transformer.setOutputProperty("indent","yes");
        transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");        
 
        // write document DOM to disk
        transformer.transform(domSource,streamResult);
	}

	public void update() {
		for(DocumentContribution contribution : contributions) {
			contribution.update();
		}
	}
}
