/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Jul 20, 2008
 */

package eu.moonlight3d.common.image.core;

import eu.moonlight3d.colour.Colour;

/**
 * A utility class for painting primtives on a Framebuffer
 *
 * @author gregor
 */
public class FramebufferPainter {
	/**
	 * Draw a single coloured pixel onto a RGB or RGBA framebuffer
	 * 
	 * @param buffer the framebuffer to paint to
	 * @param x the x coordinate of the pixel
	 * @param y the y coordinate of the pixel
	 * @param colour the colour of the pixel
	 */
	public static void drawPixel(Framebuffer buffer, int x, int y, Colour colour) {
		if(x<0 || x>=buffer.getWidth()) {
			return;
		}
		if(y<0|| y>=buffer.getHeight()) {
			return;
		}
		
		buffer.getChannel("red").putPixel(x, y, (float) colour.red);
		buffer.getChannel("green").putPixel(x, y, (float) colour.green);
		buffer.getChannel("blue").putPixel(x, y, (float) colour.blue);
		if(buffer.getChannel("alpha")!=null) {
			buffer.getChannel("alpha").putPixel(x, y, (float) colour.alpha);			
		}
	}
	
	/**
	 * Draw a straight line onto a RGB or RGBA framebuffer
	 * 
	 * @param buffer the framebuffer to paint to
	 * @param x1 the x coordinate of the line starting point
	 * @param y1 the y coordinate of the line starting point
	 * @param x2 the x coordinate of the line ending point
	 * @param y2 the y coordinate of the line ending point
	 * @param colour the colour of the line
	 */
	public static void drawLine(Framebuffer buffer, int x1, int y1, int x2, int y2, Colour colour) {
	    boolean steep = Math.abs(y2 - y1) > Math.abs(x2 - x1);
		if (steep) {
			int temp;
			temp = x1;
			x1 = y1;
			y1 = temp;

			temp = x2;
			x2 = y2;
			y2 = temp;
		}
		if (x1 > x2) {
			int temp;
			temp = x1;
			x1 = x2;
			x2 = temp;

			temp = y1;
			y1 = y2;
			y2 = temp;
		}
		double deltaX = x2 - x1;
		double deltaY = Math.abs(y2 - y1);

		double error = 0;
		double deltaError=0;
		if(deltaX!=0) {
			deltaError = deltaY / deltaX;
		}
		int yStep;
		int y = y1;
		if (y1 < y2) {
			yStep = 1;
		} else {
			yStep = -1;
		}
		for (int x = x1; x <= x2; x++) {
			if (steep) {
				drawPixel(buffer, y, x, colour);
			} else {
				drawPixel(buffer, x, y, colour);
			}
			error = error + deltaError;
			if (error >= 0.5) {
				y = y + yStep;
				error = error - 1.0;
			}
		}
	}

	/**
	 * Draw a rectangle onto a RGB or RGBA framebuffer.
	 * 
	 * @param buffer the buffer to paint to
	 * @param x1 the x coordinate of the first corner of the rectangle
	 * @param y1 the y coordinate of the first corner of the rectangle
	 * @param x2 the x coordinate of the second corner of the rectangle
	 * @param y2 the y coordinate of the second corner of the rectangle
	 * @param colour the colour of the rectangle
	 */
	public static void drawRect(Framebuffer buffer, int x1, int y1, int x2, int y2, Colour colour) {
		if(x1>x2) {
			int temp;
			temp=x2;
			x2=x1;
			x1=temp;

		}
		if(y1>y2) {
			int temp;
			temp=y2;
			y2=y1;
			y1=temp;
		}
		
		drawLine(buffer,x1,y1,x1,y2,colour);
		drawLine(buffer,x1,y1,x2,y1,colour);
		drawLine(buffer,x2,y1,x2,y2,colour);
		drawLine(buffer,x1,y2,x2,y2,colour);
	}
	
	/**
	 * Draw a filled rectangle onto a RGB or RGBA framebuffer.
	 * 
	 * @param buffer the buffer to paint to
	 * @param x1 the x coordinate of the first corner of the rectangle
	 * @param y1 the y coordinate of the first corner of the rectangle
	 * @param x2 the x coordinate of the second corner of the rectangle
	 * @param y2 the y coordinate of the second corner of the rectangle
	 * @param colour the colour of the rectangle
	 */
	public static void fillRect(Framebuffer buffer, int x1, int y1, int x2, int y2, Colour colour) {
		if(x1>x2) {
			int temp;
			temp=x2;
			x2=x1;
			x1=temp;

		}
		if(y1>y2) {
			int temp;
			temp=y2;
			y2=y1;
			y1=temp;
		}
		
		for(int x=x1;x<=x2;x++) {
			for(int y=y1;y<=y2;y++) {
				drawPixel(buffer, x, y, colour);
			}
		}
	}
}
