/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Oct 31, 2005
 */
package eu.moonlight3d.common.image.hdr;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;

import eu.moonlight3d.colour.Colour;
import eu.moonlight3d.common.image.core.Channel;
import eu.moonlight3d.common.image.core.Framebuffer;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.FileType;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.math.Helper;
import eu.moonlight3d.math.Helper.FRExpResult;

/**
 * Support for the HDR immage file format.
 *
 * @author gregor
 */
public class ImageFormat implements eu.moonlight3d.common.image.core.ImageFormat {

	public String getName() {
		return "Radiance HDR";
	}

	public FileType getFileType() {
		ArrayList<String> patterns=new ArrayList<String>();
		patterns.add("*.hdr");
		return new FileType("Radiance HDR",patterns);
	}

	public boolean isLoadable() {
		return true;
	}
	
	public boolean canLoadFile(String filename) {
		try {
			// check if file has required signature
			RandomAccessFile input=new RandomAccessFile(filename,"r");
			
			String header=input.readLine();
			if(!header.equals("#?RADIANCE")) {
				return false;
			}
			input.close();
		} catch (FileNotFoundException e) {
			return false;
		} catch (IOException e) {
			return false;
		}
		return true;
	}
	
	private void decodePixelColour(byte[] pixel, Colour colour) {
		int unsignedPixel[]=new int[4];
		for(int i=0;i<4;i++) {
			if(pixel[i]>=0) {
				unsignedPixel[i]=pixel[i];
			} else {
				unsignedPixel[i]=256+pixel[i];
			}
		}
		double exponent;
		exponent=Math.pow(2, unsignedPixel[3]-(128.0+8));
		colour.red=unsignedPixel[0]*exponent;
		colour.green=unsignedPixel[1]*exponent;
		colour.blue=unsignedPixel[2]*exponent;
	}
	
	private class PixelWriter {
		private Framebuffer target;
		private Channel redChannel;
		private Channel greenChannel;
		private Channel blueChannel;
		private boolean flipX;
		private boolean flipY;
		private boolean xFirst;

		private int index;

		public PixelWriter(Framebuffer target, boolean flipX, boolean flipY, boolean xFirst) {
			this.target=target;
			this.flipX=flipX;
			this.flipY=flipY;
			this.xFirst=xFirst;
			
			redChannel=target.getChannel("red");
			greenChannel=target.getChannel("green");
			blueChannel=target.getChannel("blue");

			index=0;
		}
		
		public void writePixels(Colour colour, int runLength) {
			for(int i=0;i<runLength;i++) {
				int x, y;
				
				if(xFirst) {
					if(flipX) {
						x=target.getWidth()-index/target.getHeight()-1;
					} else {
						x=index/target.getHeight();
					}
					if(flipY) {
						y=target.getWidth()-(index%target.getHeight())-1;
					} else {
						y=index%target.getHeight();
					}
				} else {
					if(flipX) {
						x=target.getWidth()-(index%target.getWidth())-1;
					} else {
						x=index%target.getWidth();
					}
					if(flipY) {
						y=index/target.getWidth();
					} else {
						y=target.getHeight()-index/target.getWidth()-1;
					}
				}
				
				redChannel.putPixel(x, y, (float) colour.red);
				greenChannel.putPixel(x, y, (float) colour.green);
				blueChannel.putPixel(x, y, (float) colour.blue);
				index++;
			}
		}
	}

	public Framebuffer loadFile(String filename) {
		try {
			RandomAccessFile input=new RandomAccessFile(filename,"r");
			
			String header=input.readLine();
			if(!header.equals("#?RADIANCE")) {
				State.getInstance().logger().warning(L10N.translate("HDR file signature is missing"));
				return null;
			}
			String line;
			do {
				line=input.readLine();
				
				if(line.length()>0) {
					// skip lines that are comments
					if(line.charAt(0)=='#') {
						continue;
					}
					if(line.substring(0,7).equals("FORMAT=")) {
						if(!line.substring(7).equals("32-bit_rle_rgbe")) {
							State.getInstance().logger().warning(L10N.translate("HDR file contains data in an unsupported format (not RGB)"));						
						}
					}
				}
			} while(!line.equals(""));
			
			// parse resolution string
			String resolutionString=input.readLine();
			String[] resolutionParts=resolutionString.split(" ");
			boolean flipX;
			boolean flipY;
			boolean xFirst;
			int resolutionX;
			int resolutionY;
			if(resolutionParts[0].charAt(1)=='X') {
				xFirst=true;
				
				if(resolutionParts[0].charAt(0)=='+') {
					flipX=false;
				} else {
					flipX=true;
				}
				
				if(resolutionParts[2].charAt(0)=='+') {
					flipY=false;
				} else {
					flipY=true;
				}
				
				resolutionX=Integer.parseInt(resolutionParts[1]);
				resolutionY=Integer.parseInt(resolutionParts[3]);
			} else {
				xFirst=false;

				if(resolutionParts[0].charAt(0)=='+') {
					flipY=false;
				} else {
					flipY=true;
				}
				
				if(resolutionParts[2].charAt(0)=='+') {
					flipX=false;
				} else {
					flipX=true;
				}
				
				resolutionY=Integer.parseInt(resolutionParts[1]);
				resolutionX=Integer.parseInt(resolutionParts[3]);
			}
			
			// read actual image data
			Framebuffer framebuffer=new Framebuffer(resolutionX,resolutionY);
			framebuffer.addChannel("red");
			framebuffer.addChannel("green");
			framebuffer.addChannel("blue");
			
			byte[] buffer=new byte[4];
			Colour temp=new Colour();
			PixelWriter pixelWriter=new PixelWriter(framebuffer,flipX,flipY,xFirst);
			// TODO rewrite to decode RLE segments
			int index=0;
			while(index<resolutionX*resolutionY) {
				input.read(buffer);
				decodePixelColour(buffer, temp);
				pixelWriter.writePixels(temp, 1);
				index++;
			}
			input.close();
			
			return framebuffer;
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return null;
	}

	public boolean isSaveable() {
		return true;
	}

	/**
	 * Convert a red, green, blue value tuple for a pixel to the compressed 
	 * 4 byte format that is used in HDR files.
	 * 
	 * @param red the red component of the colour
	 * @param green the green component of the colour
	 * @param blue the blue component of the colour
	 * @return the resulting compressed 4 byte representation
	 */
	private byte[] convertColour(double red, double green, double blue)
	{
		double  d;
		byte colour[]=new byte[4];
		
		if(red>green) {
			d=red;
		} else {
			d=green;
		}
		if(blue > d) {
			d = blue;
		}

		FRExpResult frexpResult=Helper.frexp(d);
		
		d=frexpResult.mantissa * 255.9999 / d;

		colour[0] = (byte) (red * d);
		colour[1] = (byte) (green * d);
		colour[2] = (byte) (blue * d);
		colour[3] = (byte) (frexpResult.exponent + 128);
		
		// do not write a special colour value that is used for RLE encoding
		if(colour[0]==255 && colour[1]==255 && colour[2]==255) {
			colour[0]=127;
			colour[1]=127;
			colour[2]=127;
			colour[3]++;
		}
		
		return colour;
	}

	public void saveFile(String filename, Framebuffer buffer) {
		try {
			RandomAccessFile outputFile=new RandomAccessFile(filename,"rw");
			
			outputFile.setLength(0);
			
			outputFile.writeBytes("#?RADIANCE\n");
			outputFile.writeBytes("# no comment\n");
			outputFile.writeBytes("FORMAT=32-bit_rle_rgbe\n");
			outputFile.writeBytes("SOFTWARE=Moonlight|3D version " + State.getInstance().getVersionInformation().getVersionString() + "\n");
			outputFile.writeBytes("\n");
			outputFile.writeBytes("-Y " + buffer.getHeight() + " +X " + buffer.getWidth() + "\n");
			
			Channel red=buffer.getChannel("red");
			Channel green=buffer.getChannel("green");
			Channel blue=buffer.getChannel("blue");
			for(int y=0;y<buffer.getHeight();y++) {
				for(int x=0;x<buffer.getWidth();x++) {
					outputFile.write(convertColour(red.getPixel(x, y),green.getPixel(x, y),blue.getPixel(x, y)));
				}
			}
			
			outputFile.close();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

}
