/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Oct 31, 2005
 */
package eu.moonlight3d.common.image.openexr;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Iterator;

import net.exrforjava.channels.Channel;
import net.exrforjava.image.Header;
import net.exrforjava.io.FrameBuffer;
import net.exrforjava.io.InputFile;
import net.exrforjava.io.OutputFile;
import eu.moonlight3d.common.image.core.Framebuffer;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.FileType;
import eu.moonlight3d.framework.localization.L10N;

public class ImageFormat implements eu.moonlight3d.common.image.core.ImageFormat {

	public String getName() {
		return "OpenEXR";
	}

	public FileType getFileType() {
		ArrayList<String> patterns=new ArrayList<String>();
		patterns.add("*.exr");
		return new FileType("OpenEXR",patterns);
	}
	
	public boolean isLoadable() {
		return true;
	}
	
	public boolean canLoadFile(String filename) {
		try {
			// check if file has required magic bytes
			RandomAccessFile input=new RandomAccessFile(filename,"r");
			
			byte[] magic=new byte[4];
			input.read(magic);
			if(magic[0]!=0x76 || magic[1]!=0x2f || magic[2]!=0x31 || magic[3]!=0x01) {
				return false;
			}
			input.close();
		} catch (FileNotFoundException e) {
			return false;
		} catch (IOException e) {
			return false;
		}
		return true;
	}

	public Framebuffer loadFile(String filename) throws IOException {
		InputFile file = new InputFile(filename);

		FrameBuffer frameBuffer = new FrameBuffer(file.getHeader().getDataWindow());
		file.setFrameBuffer(frameBuffer);

		file.readPixels();

		int width=file.getHeader().getDataWindow().getWidth();
		int height=file.getHeader().getDataWindow().getHeight();
		Framebuffer buffer=new Framebuffer(width,height);
		
		System.out.println(frameBuffer.getChannels().size());
		
//		Iterator<String> i = frameBuffer.getChannels().iterator();
//		while(i.hasNext()) {
//			String s=i.next();
//			System.out.println(s);
//		}
//		for(String channelName : frameBuffer.getChannels()) {
//			System.out.println(channelName);
//		}
		
		for(String channelName : frameBuffer.getChannels()) {
			Channel channel=frameBuffer.getChannelTypes().get(channelName);
			if(channel.xSampling!=1 || channel.ySampling!=1) {
				String warning=MessageFormat.format(L10N.translate("Channel {0} in OpenEXR file is sampled with a lower resolution and was not read."),channelName);
				State.getInstance().logger().warning(warning);
				continue;
			}
			/*
			 * R, G, B and A are well known channels which have different names in Moonlight|3D.
			 * Convert these channel names
			 */
			String internalChannelName=channelName;
			if(channelName.equals("R")) {
				internalChannelName="red";
			} else if(channelName.equals("G")) {
				internalChannelName="green";
			} else if(channelName.equals("B")) {
				internalChannelName="blue";
			} else if(channelName.equals("A")) {
				internalChannelName="alpha";
			}
			buffer.addChannel(internalChannelName);
			switch(frameBuffer.getChannelTypes().get(channelName).type) {
			case FLOAT:
			case HALF:
			{
				float[][] data = frameBuffer.getFloatData(channelName);
				float[] channelBuffer=buffer.getChannel(internalChannelName).getBuffer();
				for(int x=0;x<width;x++) {
					for(int y=0;y<height;y++) {
						channelBuffer[y*width+x]=data[y][x];
					}
				}
				break;
			}
			case UINT:
			{
				int[][] data = frameBuffer.getIntData(channelName);
				float[] channelBuffer=buffer.getChannel(internalChannelName).getBuffer();
				for(int x=0;x<width;x++) {
					for(int y=0;y<height;y++) {
						channelBuffer[y*width+x]=data[y][x]/255f;
					}
				}
				break;
			}
			}
		}
		
		return buffer;
	}
	
	public boolean isSaveable() {
		return true;
	}

	public void saveFile(String filename, Framebuffer buffer) throws IOException {
		Header header=new Header(buffer.getWidth(),buffer.getHeight());
		OutputFile outputFile=new OutputFile(filename,header);
		
		FrameBuffer frameBuffer=new FrameBuffer(outputFile.getDataWindow());
		
		for(String channelName : buffer.getChannels()) {
			float[][] data=new float[buffer.getHeight()][buffer.getWidth()];
			float[] channelBuffer=buffer.getChannel(channelName).getBuffer();
			
			for(int x=0;x<buffer.getWidth();x++) {
				for(int y=0;y<buffer.getHeight();y++) {
					data[y][x]=channelBuffer[y*buffer.getWidth()+x];
				}
			}

			String externalChannelName=channelName;
			if(channelName.equals("red")) {
				externalChannelName="R";
			} else if(channelName.equals("green")) {
				externalChannelName="G";
			} else if(channelName.equals("blue")) {
				externalChannelName="B";
			} else if(channelName.equals("alpha")) {
				externalChannelName="A";
			}
			frameBuffer.setChannel(externalChannelName, data);
		}
		
		outputFile.setFrameBuffer(frameBuffer);
		outputFile.writePixels();
		
		outputFile.close();
	}
}
