/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Feb 3, 2008
 */

package eu.moonlight3d.common.ui.colourdialog;

import com.trolltech.qt.core.QPointF;
import com.trolltech.qt.core.Qt;
import com.trolltech.qt.gui.QColor;
import com.trolltech.qt.gui.QDialog;
import com.trolltech.qt.gui.QDoubleSpinBox;
import com.trolltech.qt.gui.QGridLayout;
import com.trolltech.qt.gui.QHBoxLayout;
import com.trolltech.qt.gui.QImage;
import com.trolltech.qt.gui.QLabel;
import com.trolltech.qt.gui.QLinearGradient;
import com.trolltech.qt.gui.QMouseEvent;
import com.trolltech.qt.gui.QPaintEvent;
import com.trolltech.qt.gui.QPainter;
import com.trolltech.qt.gui.QPolygon;
import com.trolltech.qt.gui.QPushButton;
import com.trolltech.qt.gui.QTabWidget;
import com.trolltech.qt.gui.QVBoxLayout;
import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.colour.Colour;
import eu.moonlight3d.framework.localization.L10N;

/**
 * And advanced colour selection dialog.
 *
 * @author gregor
 */
public class ColourDialog extends QDialog {
	/**
	 * Flag indicating whether the alpha value should be editable, too
	 */
	private boolean editAlpha;
	/**
	 * The current colour value
	 */
	private Colour colour;
	/**
	 * Flag that is set to true when the spin box values are currently
	 * being updated from other sources
	 */
	private boolean updatingSpinboxes=false;
	
	/**
	 * Enumeration of all interpolation modes known to the ColourSlider
	 * class.
	 *
	 * @author gregor
	 */
	private enum InterpolationMode {
		/**
		 * RGB red channel colour slider 
		 */
		RGBRed,
		/**
		 * RGB green channel colour slider 
		 */
		RGBGreen,
		/**
		 * RGB blue channel colour slider 
		 */
		RGBBlue,
		/**
		 * HSV hue channel colour slider 
		 */
		HSVHue,
		/**
		 * HSV saturation channel colour slider 
		 */
		HSVSaturation,
		/**
		 * HSV value channel colour slider 
		 */
		HSVValue,
		/**
		 * Alpha channel colour slider 
		 */
		Alpha
	}

	/**
	 * A colour slider widget which displays a sliding triangle over
	 * a colour gradient indicating the variations that varying the
	 * slider's value can produce.
	 *
	 * @author gregor
	 */
	private class ColourSlider extends QWidget {
		/**
		 * The slider's interpolation mode
		 */
		private InterpolationMode interpolationMode;
		/**
		 * Flag indicating if the user is currently dragging the slider.
		 */
		private boolean dragging=false;
		/**
		 * The original value before the dragging operation started
		 */
		private double originalValue;
		/**
		 * The mouse cursor position at the start of the dragging operation
		 */
		private int startX;
		
		/**
		 * Default constructor taking the parent widget and the interpolation
		 * mode.
		 * 
		 * @param parent parent widget
		 * @param interpolationMode interpolation mode for the slider
		 */
		public ColourSlider(QWidget parent, InterpolationMode interpolationMode) {
			super(parent);
			this.interpolationMode=interpolationMode;
			setFixedSize(256,15);
		}
		
		@Override
		protected void mousePressEvent(QMouseEvent event) {
			dragging=true;
			startX=event.x();
			switch(interpolationMode) {
			case RGBRed:
				originalValue=colour.red;
				break;
			case RGBGreen:
				originalValue=colour.green;
				break;
			case RGBBlue:
				originalValue=colour.blue;
				break;
			case HSVHue:
				originalValue=colour.getHue();
				break;
			case HSVSaturation:
				originalValue=colour.getSaturation();
				break;
			case HSVValue:
				originalValue=colour.getValue();
				break;
			case Alpha:
				originalValue=colour.alpha;
				break;
			}				
		}
		
		@Override
		protected void mouseReleaseEvent(QMouseEvent event) {
			dragging=false;
		}
		
		@Override
		protected void mouseMoveEvent(QMouseEvent event) {
			if(dragging) {
				int deltaX=event.x()-startX;
				
				double valueDelta=deltaX/(double)width();
				double newValue=originalValue+valueDelta;
				
				if(newValue>1) {
					newValue=1;
				} else if(newValue<0) {
					newValue=0;
				}

				switch(interpolationMode) {
				case RGBRed:
					colour.red=newValue;
					break;
				case RGBGreen:
					colour.green=newValue;
					break;
				case RGBBlue:
					colour.blue=newValue;
					break;
				case HSVHue:
					colour.setHue(newValue);
					break;
				case HSVSaturation:
					colour.setSaturation(newValue);
					break;
				case HSVValue:
					colour.setValue(newValue);
					break;
				case Alpha:
					colour.alpha=newValue;
					break;
				}
				
				updateSpinBoxes();
				parentWidget().update();
			}
		}
		
		@Override
		protected void paintEvent(QPaintEvent event) {
			super.paintEvent(event);
			QPainter painter=new QPainter(this);
			
			QLinearGradient gradient=new QLinearGradient(new QPointF(0,0), new QPointF(width(),0));
			Colour.RGB rgb=colour.getsRGB8();
			QColor startColor;
			QColor stopColor;
			Colour startHSVColour;
			Colour stopHSVColour;
			int offset=0;

			switch(interpolationMode) {
			case RGBRed:
				startColor=new QColor(0,rgb.green,rgb.blue);
				stopColor=new QColor(255,rgb.green,rgb.blue);
				
				gradient.setColorAt(0, startColor);
				gradient.setColorAt(1, stopColor);
				painter.setBrush(gradient);
				painter.fillRect(0,0,width(),height(), painter.brush());
				
				offset=(int) (colour.red*width());
				break;
			case RGBGreen:
				startColor=new QColor(rgb.red,0,rgb.blue);
				stopColor=new QColor(rgb.red,255,rgb.blue);
				
				gradient.setColorAt(0, startColor);
				gradient.setColorAt(1, stopColor);
				painter.setBrush(gradient);
				painter.fillRect(0,0,width(),height(), painter.brush());

				offset=(int) (colour.green*width());
				break;
			case RGBBlue:
				startColor=new QColor(rgb.red,rgb.green,0);
				stopColor=new QColor(rgb.red,rgb.green,255);
				
				gradient.setColorAt(0, startColor);
				gradient.setColorAt(1, stopColor);
				painter.setBrush(gradient);
				painter.fillRect(0,0,width(),height(), painter.brush());

				offset=(int) (colour.blue*width());
				break;
			case HSVHue:
				Colour hueColour=colour.clone();
				for(int x=0;x<width();x++) {
					double hue=x/(double)width();
					hueColour.setHue(hue);
					rgb=hueColour.getsRGB8();
					painter.setPen(new QColor(rgb.red,rgb.green,rgb.blue));
					painter.drawLine(x,0,x,height());
				}
				
				offset=(int) (colour.getHue()*width());
				break;
			case HSVSaturation:
				startHSVColour=colour.clone();
				startHSVColour.setSaturation(0);
				rgb=startHSVColour.getsRGB8();
				startColor=new QColor(rgb.red,rgb.green,rgb.blue);
				stopHSVColour=colour.clone();
				stopHSVColour.setSaturation(1);
				rgb=stopHSVColour.getsRGB8();
				stopColor=new QColor(rgb.red,rgb.green,rgb.blue);
				
				gradient.setColorAt(0, startColor);
				gradient.setColorAt(1, stopColor);
				painter.setBrush(gradient);
				painter.fillRect(0,0,width(),height(), painter.brush());

				offset=(int) (colour.getSaturation()*width());
				break;
			case HSVValue:
				startHSVColour=colour.clone();
				startHSVColour.setValue(0);
				rgb=startHSVColour.getsRGB8();
				startColor=new QColor(rgb.red,rgb.green,rgb.blue);
				stopHSVColour=colour.clone();
				stopHSVColour.setValue(1);
				rgb=stopHSVColour.getsRGB8();
				stopColor=new QColor(rgb.red,rgb.green,rgb.blue);
				
				gradient.setColorAt(0, startColor);
				gradient.setColorAt(1, stopColor);
				painter.setBrush(gradient);
				painter.fillRect(0,0,width(),height(), painter.brush());

				offset=(int) (colour.getValue()*width());
				break;
			case Alpha:
				// paint background checker pattern
				painter.setBrush(new QColor(0x66,0x66,0x66));
				painter.fillRect(0,0,width(),height(),painter.brush());
				painter.setBrush(new QColor(0x99,0x99,0x99));
				for(int x=0;x<width();x+=6) {
					for(int y=0;y<height();y+=6) {
						if(((x+y)/6)%2==0) {
							painter.fillRect(x,y,6,6,painter.brush());
						}
					}
				}
				
				// paint gradient bar
				startColor=new QColor(rgb.red,rgb.green,rgb.blue,0);
				stopColor=new QColor(rgb.red,rgb.green,rgb.blue,255);
				
				gradient.setColorAt(0, startColor);
				gradient.setColorAt(1, stopColor);
				painter.setBrush(gradient);
				painter.fillRect(0,0,width(),height(), painter.brush());

				offset=(int) (colour.alpha*width());
				break;
			}
			
			// clamp offset
			if(offset>width()) {
				offset=width();
			} else if(offset<0) {
				offset=0;
			}
			
			// paint black and white position marker triangles
			painter.setPen(QColor.white);
			painter.setBrush(QColor.white);
			
			QPolygon downArrow=new QPolygon();
			downArrow.add(offset-3,0);
			downArrow.add(offset+3,0);
			downArrow.add(offset,3);			
			painter.drawPolygon(downArrow);
			
			painter.setPen(QColor.black);
			painter.setBrush(QColor.black);

			QPolygon upArrow=new QPolygon();
			upArrow.add(offset-3,height()-1);
			upArrow.add(offset+3,height()-1);
			upArrow.add(offset,height()-4);			
			painter.drawPolygon(upArrow);
			
			painter.end();
		}
	}
	
	/**
	 * Small widget showing the selected colour in a rectangular area
	 * which is split in the middle. On the left hand side the colour
	 * is shown with transparency, on the right hand side it is shown
	 * transparently over a pattern background.
	 *
	 * @author gregor
	 */
	private class ColourPreview extends QWidget {
		/**
		 * Default constructor.
		 * @param parent parent widget
		 */
		public ColourPreview(QWidget parent) {
			super(parent);
			setMinimumHeight(40);
			setMaximumHeight(60);
		}
		
		@Override
		protected void paintEvent(QPaintEvent event) {
			QPainter painter=new QPainter(this);
			
			// paint background checker pattern
			painter.setBrush(new QColor(0x66,0x66,0x66));
			painter.fillRect(0,0,width(),height(),painter.brush());
			painter.setBrush(new QColor(0x99,0x99,0x99));
			for(int x=0;x<width();x+=6) {
				for(int y=0;y<height();y+=6) {
					if(((x+y)/6)%2==0) {
						painter.fillRect(x,y,6,6,painter.brush());
					}
				}
			}
			
			Colour.RGB rgb=colour.getsRGB8();
			
			painter.setBrush(new QColor(rgb.red,rgb.green,rgb.blue,rgb.alpha));
			painter.fillRect(0, 0, width()/2,height(),painter.brush());
			painter.setBrush(new QColor(rgb.red,rgb.green,rgb.blue));
			painter.fillRect(width()/2, 0, width()-width()/2,height(),painter.brush());
			
			painter.end();
		}
	}
	
	private class HSVPlane extends  QWidget {
		private QImage hueValuePlane;
		private boolean dragging;
		
		public HSVPlane(QWidget parent) {
			super(parent);
			setFixedSize(256,256);
			dragging=false;
			
			hueValuePlane=new QImage(256,256,QImage.Format.Format_RGB32);
			
			QPainter painter=new QPainter(hueValuePlane);
			
			Colour planeColour=new Colour();
			
			for(int i=0;i<256;i++) {
				for(int j=0;j<256;j++) {
					planeColour.setFromHSV(i/256.0, 1.0, j/256.0);
					
					Colour.RGB rgb=planeColour.getsRGB8();
					
					painter.setPen(new QColor(rgb.red,rgb.green,rgb.blue));
					painter.drawPoint(i, j);
				}
			}

			painter.end();
		}
		
		private void updateColour(int mouseX, int mouseY) {
			double hue=mouseX/256.0;
			double value=mouseY/256.0;
			if(hue<0.0) {
				hue=0.0;
			} else if(hue>1.0) {
				hue=1.0;
			}
			if(value<0.0) {
				value=0.0;
			} else if(value>1.0) {
				value=1.0;
			}
			colour.setHue(hue);
			colour.setValue(value);
			
			updateSpinBoxes();
			window().update();
		}
		
		@Override
		protected void mousePressEvent(QMouseEvent event) {
			dragging=true;
			updateColour(event.x(),event.y());
		}
		
		@Override
		protected void mouseReleaseEvent(QMouseEvent event) {
			dragging=false;
		}
		
		@Override
		protected void mouseMoveEvent(QMouseEvent event) {
			if(dragging) {
				updateColour(event.x(),event.y());
			}
		}
		
		@Override
		public void paintEvent(QPaintEvent event) {
			QPainter painter=new QPainter(this);
			
			painter.drawImage(0, 0, hueValuePlane);
			
			int markX=(int)(colour.getHue()*256);
			int markY=(int)(colour.getValue()*256);
			final int markRadius=2;
			painter.setRenderHint(QPainter.RenderHint.Antialiasing);
			painter.setPen(QColor.white);
			painter.drawArc(markX-markRadius,markY-markRadius,2*markRadius,2*markRadius,0,5760);
			painter.setPen(QColor.black);
			painter.drawArc(markX-markRadius-1,markY-markRadius-1,2*markRadius+2,2*markRadius+2,0,5760);
			
			painter.end();
		}
	}
	
	private class ColourWheel extends  QWidget {
		private boolean draggingWheel;
		private boolean draggingTriangle;

		private final int RADIUS=100;
		private final int WIDTH=15;
		
		public ColourWheel(QWidget parent) {
			super(parent);
			setFixedSize(2*RADIUS,2*RADIUS);
			draggingWheel=false;
			draggingTriangle=false;
		}
		
		private double clamp(double value) {
			if(value<0.0) {
				value=0.0;
			} else if(value>1.0) {
				value=1.0;
			}
			return value;
		}
		
		private void updateHue(int mouseX, int mouseY) {
			double hue=0.5+Math.atan2((RADIUS-mouseY), (RADIUS-mouseX))/(2*Math.PI);
			hue=clamp(hue);
			colour.setHue(hue);
			
			updateSpinBoxes();
			window().update();
		}
		
		private void updateSaturationValue(int mouseX, int mouseY) {
			double saturation;
			double value;
			
			double hueAngle=colour.getHue()*2*Math.PI;
			
			double deltaX=mouseX-RADIUS;
			double deltaY=mouseY-RADIUS;
			
			double resultX=Math.cos(hueAngle)*deltaX+Math.sin(hueAngle)*deltaY;
			double resultY=-Math.sin(hueAngle)*deltaX+Math.cos(hueAngle)*deltaY;
			
		    saturation = clamp( (resultX+0.5*(RADIUS-WIDTH)) / ((RADIUS-WIDTH)*1.5));

		    double halfHeight = ((RADIUS-WIDTH) * Math.sin(Math.PI/3.0)) * (1.0 - saturation);
		    value = clamp( ((resultY+ halfHeight) / (2.0*halfHeight)));
		    
			saturation=clamp(saturation);
			value=clamp(value);
			
			colour.setSaturation(saturation);
			colour.setValue(value);
			
			updateSpinBoxes();
			window().update();
		}
		
		@Override
		protected void mousePressEvent(QMouseEvent event) {
			double distanceFromCentre;
			
			distanceFromCentre=Math.sqrt((RADIUS-event.x())*(RADIUS-event.x())+(RADIUS-event.y())*(RADIUS-event.y()));
			
			if(distanceFromCentre >= RADIUS-WIDTH && distanceFromCentre <=RADIUS) {
				draggingWheel=true;
				updateHue(event.x(),event.y());
			} else if(distanceFromCentre<RADIUS-WIDTH) {
				draggingTriangle=true;
				updateSaturationValue(event.x(),event.y());
			}
		}
		
		@Override
		protected void mouseReleaseEvent(QMouseEvent event) {
			draggingWheel=false;
			draggingTriangle=false;
		}
		
		@Override
		protected void mouseMoveEvent(QMouseEvent event) {
			if(draggingWheel) {
				updateHue(event.x(),event.y());
			}
			if(draggingTriangle) {
				updateSaturationValue(event.x(),event.y());
			}
		}
		
		private void drawCircle(QPainter painter) {
			Colour wheelColour=new Colour();

			final int numLines=360;
			for(int i=0;i<numLines;i++) {
				int lineStartX=(int) (RADIUS+RADIUS*Math.cos(i*2*Math.PI/numLines));
				int lineStartY=(int) (RADIUS+RADIUS*Math.sin(i*2*Math.PI/numLines));
				int lineStopX=(int) (RADIUS+(RADIUS-WIDTH)*Math.cos(i*2*Math.PI/numLines));
				int lineStopY=(int) (RADIUS+(RADIUS-WIDTH)*Math.sin(i*2*Math.PI/numLines));
				int nextStartX=(int) (RADIUS+RADIUS*Math.cos((i+1)*2*Math.PI/numLines));
				int nextStartY=(int) (RADIUS+RADIUS*Math.sin((i+1)*2*Math.PI/numLines));
				int nextStopX=(int) (RADIUS+(RADIUS-WIDTH)*Math.cos((i+1)*2*Math.PI/numLines));
				int nextStopY=(int) (RADIUS+(RADIUS-WIDTH)*Math.sin((i+1)*2*Math.PI/numLines));
				
				// set color
				wheelColour.setFromHSV(i/(double)numLines, 1.0, 1.0);
				Colour.RGB rgb=wheelColour.getsRGB8();
				painter.setPen(new QColor(rgb.red,rgb.green,rgb.blue));
				painter.setBrush(new QColor(rgb.red,rgb.green,rgb.blue));

				//draw polygon region
				QPolygon polygon=new QPolygon();
				polygon.add(lineStartX, lineStartY);
				polygon.add(lineStopX, lineStopY);
				polygon.add(nextStopX, nextStopY);
				polygon.add(nextStartX, nextStartY);
				
				painter.drawPolygon(polygon);
			}

			painter.setRenderHint(QPainter.RenderHint.Antialiasing);
			painter.setPen(QColor.white);
			int lineStartX=(int) (RADIUS+RADIUS*Math.cos(colour.getHue()*2*Math.PI));
			int lineStartY=(int) (RADIUS+RADIUS*Math.sin(colour.getHue()*2*Math.PI));
			int lineStopX=(int) (RADIUS+(RADIUS-WIDTH)*Math.cos(colour.getHue()*2*Math.PI));
			int lineStopY=(int) (RADIUS+(RADIUS-WIDTH)*Math.sin(colour.getHue()*2*Math.PI));
			painter.drawLine(lineStartX,lineStartY,lineStopX,lineStopY);			
			painter.setRenderHint(QPainter.RenderHint.Antialiasing,false);
		}
		
		private void drawTriangle(QPainter painter) {
			double hueAngle=colour.getHue()*2*Math.PI;

			QPointF pointA=new QPointF();
			QPointF pointB=new QPointF();
			QPointF pointC=new QPointF();
			pointA.setX(RADIUS+(RADIUS-WIDTH)*Math.cos(hueAngle));
			pointA.setY(RADIUS+(RADIUS-WIDTH)*Math.sin(hueAngle));
			pointB.setX(RADIUS+(RADIUS-WIDTH)*Math.cos(hueAngle+2.0*Math.PI/3.0));
			pointB.setY(RADIUS+(RADIUS-WIDTH)*Math.sin(hueAngle+2.0*Math.PI/3.0));
            pointC.setX(RADIUS+(RADIUS-WIDTH)*Math.cos(hueAngle+4.0*Math.PI/3.0));
			pointC.setY(RADIUS+(RADIUS-WIDTH)*Math.sin(hueAngle+4.0*Math.PI/3.0));

            int markerPositionX=(int)((1.0-colour.getSaturation())*((1.0-colour.getValue())*pointC.x()+colour.getValue()*pointB.x())+(colour.getSaturation()*pointA.x()));
            int markerPositionY=(int)((1.0-colour.getSaturation())*((1.0-colour.getValue())*pointC.y()+colour.getValue()*pointB.y())+(colour.getSaturation()*pointA.y()));
                
			Colour saturated=colour.clone();
			saturated.setSaturation(1.0);
			saturated.setValue(1.0);
			Colour.RGB saturatedRGB=saturated.getsRGB8();
			QColor rgbA=new QColor(saturatedRGB.red,saturatedRGB.green,saturatedRGB.blue);
			QColor rgbB=QColor.white;
			QColor rgbC=QColor.black;
			
			// Gouraud fill
            if(pointC.y() < pointB.y()) {
            	QPointF temp;
            	temp=pointC;
            	pointC=pointB;
            	pointB=temp;
            	QColor tempRgb;
            	tempRgb=rgbC;
            	rgbC=rgbB;
            	rgbB=tempRgb;
            }

            if(pointC.y() < pointA.y()) {
            	QPointF temp;
            	temp=pointC;
            	pointC=pointA;
            	pointA=temp;
            	QColor tempRgb;
            	tempRgb=rgbC;
            	rgbC=rgbA;
            	rgbA=tempRgb;
            }

            if(pointB.y() < pointA.y()) {
            	QPointF temp;
            	temp=pointB;
            	pointB=pointA;
            	pointA=temp;
            	QColor tempRgb; 
            	tempRgb=rgbB;
            	rgbB=rgbA;
            	rgbA=tempRgb;
            }

            if(pointA.y() == pointB.y() && pointB.x() < pointA.x()) {
            	QPointF temp;
            	temp=pointB;
            	pointB=pointA;
            	pointA=temp;
            	QColor tempRgb; 
            	tempRgb=rgbB;
            	rgbB=rgbA;
            	rgbA=tempRgb;
            }

            double dx1,dx2,dx3;
            double dr1,dr2,dr3;
            double dg1,dg2,dg3;
            double db1,db2,db3;

            QPointF pointS;
            QPointF pointE;
            QColor rgbS;
            QColor rgbE;


            if(pointB.y()-pointA.y() > 0) {
                double bay=pointB.y()-pointA.y();
                dx1=(pointB.x()-pointA.x())/bay;
                dr1=(rgbB.redF()-rgbA.redF())/bay;
                dg1=(rgbB.greenF()-rgbA.greenF())/bay;
                db1=(rgbB.blueF()-rgbA.blueF())/bay;
            } else {
                dx1=dr1=dg1=db1=0;
            }

            if(pointC.y()-pointA.y() > 0) {
                double cay=pointC.y()-pointA.y();
                dx2=(pointC.x()-pointA.x())/cay;
                dr2=(rgbC.redF()-rgbA.redF())/cay;
                dg2=(rgbC.greenF()-rgbA.greenF())/cay;
                db2=(rgbC.blueF()-rgbA.blueF())/cay;
            } else {
                dx2=dr2=dg2=db2=0;
            }

            if(pointC.y()-pointB.y() > 0) {
                double cby=pointC.y()-pointB.y();
                dx3=(pointC.x()-pointB.x())/cby;
                dr3=(rgbC.redF()-rgbB.redF())/cby;
                dg3=(rgbC.greenF()-rgbB.greenF())/cby;
                db3=(rgbC.blueF()-rgbB.blueF())/cby;
            } else {
                dx3=dr3=dg3=db3=0;
            }

            pointS=new QPointF(pointA.x(),pointA.y());
            rgbS=new QColor(rgbA);
            pointE=new QPointF(pointA.x(),pointA.y());
            rgbE=new QColor(rgbA);

            int runs = 1;
			int fill_mode = 0;

			if(dx1 == 0) {
				fill_mode = 0;
			} else if(dx1 > dx2) {
				fill_mode = 1;
				runs = 2;
			} else if(dx1 != 0) {
				fill_mode = 2;
				runs = 2;
			}

            int fill_direction_mode = 0;

            for (int current_run = 0; current_run < runs; current_run++)
            {
                double targetY = pointC.y();
                switch (fill_mode)
                {
                    case 0:
                        pointE=new QPointF(pointB.x(),pointB.y());
                        rgbE=new QColor(rgbB);
                        fill_direction_mode = 0;
                        break;
                    case 1:
                        if(current_run == 0) {
                            targetY = pointB.y();
                            fill_direction_mode = 1;
                        } else {
                            pointE=new QPointF(pointB.x(),pointB.y());
                            rgbE=new QColor(rgbB);
                            fill_direction_mode = 0;
                        }
                        break;
                    case 2:
                        if(current_run == 0) {
                            targetY = pointB.y();
                            fill_direction_mode = 2;
                        } else {
                            pointS=new QPointF(pointB.x(),pointB.y());
                            rgbS=new QColor(rgbB);
                            fill_direction_mode = 3;
                        }
                        break;
                }

                while(pointS.y() <= targetY)
                {
                    double dr, dg, db;

                    if(pointE.x()-pointS.x() > 0) {
                    	double esx=pointE.x()-pointS.x();
                        dr=(rgbE.redF()-rgbS.redF())/esx;
                        dg=(rgbE.greenF()-rgbS.greenF())/esx;
                        db=(rgbE.blueF()-rgbS.blueF())/esx;
                    } else {
                        dr=dg=db=0;
                    }
                    QPointF pointP=new QPointF(pointS.x(),pointS.y());
                    QColor rgbP=new QColor(rgbS);

                    while(pointP.x()<pointE.x()) {
                    	painter.setPen(rgbP);
                    	painter.drawPoint(pointP);
                    	pointP.setX(pointP.x()+1);
                    	rgbP.setRedF(clamp(rgbP.redF()+dr));
                    	rgbP.setGreenF(clamp(rgbP.greenF()+dg));
                    	rgbP.setBlueF(clamp(rgbP.blueF()+db));
                    }
                    
                    switch (fill_direction_mode) {
                        case 0:
                        	pointS.setX(pointS.x()+dx2);
                        	rgbS.setRedF(clamp(rgbS.redF()+dr2));
                        	rgbS.setGreenF(clamp(rgbS.greenF()+dg2));
                        	rgbS.setBlueF(clamp(rgbS.blueF()+db2));
                        	pointE.setX(pointE.x()+dx3);
                        	rgbE.setRedF(clamp(rgbE.redF()+dr3));
                        	rgbE.setGreenF(clamp(rgbE.greenF()+dg3));
                        	rgbE.setBlueF(clamp(rgbE.blueF()+db3));
                            break;
                        case 1:
                        	pointS.setX(pointS.x()+dx2);
                        	rgbS.setRedF(clamp(rgbS.redF()+dr2));
                        	rgbS.setGreenF(clamp(rgbS.greenF()+dg2));
                        	rgbS.setBlueF(clamp(rgbS.blueF()+db2));
                        	pointE.setX(pointE.x()+dx1);
                        	rgbE.setRedF(clamp(rgbE.redF()+dr1));
                        	rgbE.setGreenF(clamp(rgbE.greenF()+dg1));
                        	rgbE.setBlueF(clamp(rgbE.blueF()+db1));
                            break;
                        case 2:
                        	pointS.setX(pointS.x()+dx1);
                        	rgbS.setRedF(clamp(rgbS.redF()+dr1));
                        	rgbS.setGreenF(clamp(rgbS.greenF()+dg1));
                        	rgbS.setBlueF(clamp(rgbS.blueF()+db1));
                        	pointE.setX(pointE.x()+dx2);
                        	rgbE.setRedF(clamp(rgbE.redF()+dr2));
                        	rgbE.setGreenF(clamp(rgbE.greenF()+dg2));
                        	rgbE.setBlueF(clamp(rgbE.blueF()+db2));
                            break;
                        case 3:
                        	pointS.setX(pointS.x()+dx3);
                        	rgbS.setRedF(clamp(rgbS.redF()+dr3));
                        	rgbS.setGreenF(clamp(rgbS.greenF()+dg3));
                        	rgbS.setBlueF(clamp(rgbS.blueF()+db3));
                        	pointE.setX(pointE.x()+dx2);
                        	rgbE.setRedF(clamp(rgbE.redF()+dr2));
                        	rgbE.setGreenF(clamp(rgbE.greenF()+dg2));
                        	rgbE.setBlueF(clamp(rgbE.blueF()+db2));
                            break;
                    }
                    pointS.setY(pointS.y()+1);
                    pointE.setY(pointE.y()+1);
                }
            }
                        
			final int markRadius=3;
			painter.setRenderHint(QPainter.RenderHint.Antialiasing);
			painter.setPen(QColor.white);
			painter.drawArc(markerPositionX-markRadius,markerPositionY-markRadius,2*markRadius,2*markRadius,0,5760);
			painter.setPen(QColor.black);
			painter.drawArc(markerPositionX-markRadius-1,markerPositionY-markRadius-1,2*markRadius+2,2*markRadius+2,0,5760);
		}
		
		@Override
		public void paintEvent(QPaintEvent event) {
			QPainter painter=new QPainter(this);
			
			drawCircle(painter);
			drawTriangle(painter);
			
			painter.end();
		}
	}

	/**
	 * Tab widget containing the rgb plane, hsv plane and colour wheel
	 * selection tools
	 */
	private QTabWidget tabWidget;
	/**
	 * Colour preview widget
	 */
	private ColourPreview colourPreview;
	/**
	 * Colour slider for red component
	 */
	private ColourSlider redSlider;
	/**
	 * Spin box for red component
	 */
	private QDoubleSpinBox redSpinBox;
	/**
	 * Colour slider for green component
	 */
	private ColourSlider greenSlider;
	/**
	 * Spin box for green component
	 */
	private QDoubleSpinBox greenSpinBox;
	/**
	 * Colour slider for blue component
	 */
	private ColourSlider blueSlider;
	/**
	 * Spin box for blue component
	 */
	private QDoubleSpinBox blueSpinBox;
	/**
	 * Colour slider for hue component
	 */
	private ColourSlider hueSlider;
	/**
	 * Spin box for hue component
	 */
	private QDoubleSpinBox hueSpinBox;
	/**
	 * Colour slider for saturation component
	 */
	private ColourSlider saturationSlider;
	/**
	 * Spin box for saturation component
	 */
	private QDoubleSpinBox saturationSpinBox;
	/**
	 * Colour slider for value component
	 */
	private ColourSlider valueSlider;
	/**
	 * Spin box for value component
	 */
	private QDoubleSpinBox valueSpinBox;
	/**
	 * Colour slider for alpha component
	 */
	private ColourSlider alphaSlider;
	/**
	 * Spin box for alpha component
	 */
	private QDoubleSpinBox alphaSpinBox;
	/**
	 * The dialog's OK button
	 */
	private QPushButton okButton;
	/**
	 * The dialog's cancel button
	 */
	private QPushButton cancelButton;
	/**
	 * Flag used to remember how the dialog was closed.
	 */
	private boolean okPressed=false;
	
	/**
	 * Default constructor taking the parent window and the
	 * initial colour as arguemnt.
	 * 
	 * @param parent parent window for the dialog
	 * @param colour the initial colour that the dialog starts with
	 * @param editAlpha set to true if alpha channel should be editable
	 */
	public ColourDialog(QWidget parent, Colour colour, boolean editAlpha) {
		super(parent);
		this.editAlpha=editAlpha;
		this.colour=colour.clone();
		
		createWidgets();
	}
	
	/**
	 * Create a QDoubleSpin box for use alongside one of the colour sliders.
	 * 
	 * @return a newly created, fully initialized QDoubleSpinBox
	 */
	private QDoubleSpinBox createSpinBox() {
		QDoubleSpinBox spinBox=new QDoubleSpinBox(this);
		spinBox.setMinimum(0.0);
		spinBox.setMaximum(1.0);
		spinBox.setSingleStep(0.004);
		spinBox.setDecimals(3);
		return spinBox;
	}
	
	/**
	 * Create the dialog's widgets
	 */
	private void createWidgets() {
		setWindowTitle(L10N.translate("Select Colour"));
		
		tabWidget=new QTabWidget(this);

		QWidget hsvPlaneParent=new QWidget();
		QWidget hsvPlane=new HSVPlane(hsvPlaneParent);
		QVBoxLayout hsvPlaneLayout=new QVBoxLayout();
		hsvPlaneLayout.addWidget(hsvPlane, 0, Qt.AlignmentFlag.AlignCenter);
		hsvPlaneParent.setLayout(hsvPlaneLayout);
		
		QWidget wheelParent=new QWidget();
		QWidget wheel=new ColourWheel(wheelParent);
		QVBoxLayout wheelLayout=new QVBoxLayout();
		wheelLayout.addWidget(wheel, 0, Qt.AlignmentFlag.AlignCenter);
		wheelParent.setLayout(wheelLayout);
		
		tabWidget.addTab(hsvPlaneParent, L10N.translate("HSV"));
		tabWidget.addTab(wheelParent,L10N.translate("Wheel"));

		colourPreview=new ColourPreview(this);
		
		QGridLayout sliderLayout=new QGridLayout();
		
		QLabel redLabel=new QLabel(this);
		redLabel.setText(L10N.translate("R"));
		redSlider=new ColourSlider(this,InterpolationMode.RGBRed);
		redSpinBox=createSpinBox();
		redSpinBox.valueChanged.connect(this,"spinBoxRGBAValueChanged()");
		QLabel greenLabel=new QLabel(this);
		greenLabel.setText(L10N.translate("G"));
		greenSlider=new ColourSlider(this,InterpolationMode.RGBGreen);
		greenSpinBox=createSpinBox();
		greenSpinBox.valueChanged.connect(this,"spinBoxRGBAValueChanged()");
		QLabel blueLabel=new QLabel(this);
		blueLabel.setText(L10N.translate("B"));
		blueSlider=new ColourSlider(this,InterpolationMode.RGBBlue);
		blueSpinBox=createSpinBox();
		blueSpinBox.valueChanged.connect(this,"spinBoxRGBAValueChanged()");
		QLabel hueLabel=new QLabel(this);
		hueLabel.setText(L10N.translate("H"));
		hueSlider=new ColourSlider(this,InterpolationMode.HSVHue);
		hueSpinBox=createSpinBox();
		hueSpinBox.valueChanged.connect(this,"spinBoxHSVValueChanged()");
		QLabel saturationLabel=new QLabel(this);
		saturationLabel.setText(L10N.translate("S"));
		saturationSlider=new ColourSlider(this,InterpolationMode.HSVSaturation);
		saturationSpinBox=createSpinBox();
		saturationSpinBox.valueChanged.connect(this,"spinBoxHSVValueChanged()");
		QLabel valueLabel=new QLabel(this);
		valueLabel.setText(L10N.translate("V"));
		valueSlider=new ColourSlider(this,InterpolationMode.HSVValue);
		valueSpinBox=createSpinBox();
		valueSpinBox.valueChanged.connect(this,"spinBoxHSVValueChanged()");
		
		QLabel alphaLabel=null;
		if(editAlpha) {
			alphaLabel=new QLabel(this);
			alphaLabel.setText(L10N.translate("A"));
			alphaSlider=new ColourSlider(this,InterpolationMode.Alpha);
			alphaSpinBox=createSpinBox();
			alphaSpinBox.valueChanged.connect(this,"spinBoxRGBAValueChanged()");
		}
		
		sliderLayout.addWidget(redLabel,0,0);
		sliderLayout.addWidget(redSlider,0,1);
		sliderLayout.addWidget(redSpinBox,0,2);
		sliderLayout.addWidget(greenLabel,1,0);
		sliderLayout.addWidget(greenSlider,1,1);
		sliderLayout.addWidget(greenSpinBox,1,2);
		sliderLayout.addWidget(blueLabel,2,0);
		sliderLayout.addWidget(blueSlider,2,1);
		sliderLayout.addWidget(blueSpinBox,2,2); 
		sliderLayout.addWidget(hueLabel,3,0);
		sliderLayout.addWidget(hueSlider,3,1);
		sliderLayout.addWidget(hueSpinBox,3,2);
		sliderLayout.addWidget(saturationLabel,4,0);
		sliderLayout.addWidget(saturationSlider,4,1);
		sliderLayout.addWidget(saturationSpinBox,4,2);
		sliderLayout.addWidget(valueLabel,5,0);
		sliderLayout.addWidget(valueSlider,5,1);
		sliderLayout.addWidget(valueSpinBox,5,2);
		if(editAlpha) {
			sliderLayout.addWidget(alphaLabel,6,0);
			sliderLayout.addWidget(alphaSlider,6,1);
			sliderLayout.addWidget(alphaSpinBox,6,2);
		}

		QHBoxLayout buttonLayout=new QHBoxLayout();
		
		okButton=new QPushButton(this);
		okButton.setText(L10N.translate("OK"));
		okButton.clicked.connect(this,"onOK()");
		
		cancelButton=new QPushButton(this);
		cancelButton.setText(L10N.translate("Cancel"));
		cancelButton.clicked.connect(this,"onCancel()");

		buttonLayout.addStretch(1);
		buttonLayout.addWidget(okButton);
		buttonLayout.addWidget(cancelButton);

		QGridLayout layout=new QGridLayout();
		layout.addWidget(tabWidget,0,0,2,1);
		layout.addWidget(colourPreview,0,1);
		layout.addLayout(sliderLayout,1,1);
		layout.addLayout(buttonLayout,2,1);
		
		setLayout(layout);
		
		updateSpinBoxes();
	}
	
	/**
	 * Return whether the colour alpha value is editable
	 * @return true if alpha value is editable
	 */
	public boolean getEditAlpha() {
		return editAlpha;
	}
	
	/**
	 * Get the currently selected colour.
	 * 
	 * @return the dialog's current colour value
	 */
	public Colour getColour() {
		return colour;
	}

	/**
	 * Set the dialog's current colour value.
	 * 
	 * @param colour
	 */
	public void setColour(Colour colour) {
		this.colour = colour;
	}
	
	/**
	 * Handle click on dialog's OK button.
	 */
	protected void onOK() {
		okPressed=true;
		close();
	}
	
	/**
	 * Handle click on dialog's cancel button.
	 */
	protected void onCancel() {
		close();
	}
	
	/**
	 * Update spin box values when colour values were
	 * changed by other means.
	 */
	protected void updateSpinBoxes() {
		updatingSpinboxes=true;
		redSpinBox.setValue(colour.red);
		greenSpinBox.setValue(colour.green);
		blueSpinBox.setValue(colour.blue);
		hueSpinBox.setValue(colour.getHue());
		saturationSpinBox.setValue(colour.getSaturation());
		valueSpinBox.setValue(colour.getValue());
		if(alphaSpinBox!=null) {
			alphaSpinBox.setValue(colour.alpha);
		}
		updatingSpinboxes=false;
	}
	
	/**
	 * Update colour values because one of the RGBA spinboxes changed 
	 */
	protected void spinBoxRGBAValueChanged() {
		if(updatingSpinboxes) {
			return;
		}
		colour.red=redSpinBox.value();
		colour.green=greenSpinBox.value();
		colour.blue=blueSpinBox.value();
		if(alphaSpinBox!=null) {
			colour.alpha=alphaSpinBox.value();
		}
		update();
	}
	
	/**
	 * Update colour values because one of the HSV spinboxes changed
	 */
	protected void spinBoxHSVValueChanged() {
		if(updatingSpinboxes) {
			return;
		}
		colour.setFromHSV(hueSpinBox.value(), saturationSpinBox.value(), valueSpinBox.value());
		update();
	}
	
	/**
	 * Return, whether the user clicked OK to close
	 * the dialog.
	 * 
	 * @return true if the dialog was closed with OK, false otherwise
	 */
	public boolean getOKPressed() {
		return okPressed;
	}
}
