/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Nov 9, 2008
 */

package eu.moonlight3d.common.ui.command;

/**
 * Generic interface for undoable commands. This is the command interface
 * required by a CommandStackListener in order to support undo and redo
 * operations in an application.
 *
 * @author gregor
 */
public interface Command {
	/**
	 * Returns a very brief description of the command. This is displayed
	 * in to the user, for example in the undo and redo menu actions.
	 * 
	 * @return a brief description of the command
	 */
	public String getBriefDescription();
	
	/**
	 * Returns whether the command is capable of undoing its own actions.
	 * If it is not undoable, pushing it onto the command stack will cause
	 * the command stack to be cleared after the command executed successfully.
	 * 
	 * @return true if the command is undoable, false otherwise
	 */
	public boolean isUndoable();
	
	/**
	 * Execute the command for the first time. This method collects and validates
	 * any inputs that are required for the command. If the validation fails or
	 * the command cannot be executed for other reasons, this function returns
	 * with a false return value without having modified anything. If the command 
	 * could be executed successfully, true is returned.
	 * 
	 * Apart from input validation or other one-time preparations, the actions and
	 * the results of this method must not differ from those of redo(). Therefore
	 * it is best to actually perform the command in redo() and only leave the
	 * input gathering and checking in execute(), which calls redo() when the 
	 * checks were passed.
	 * 
	 * @return true if the command was executed successfully, false otherwise
	 */
	public boolean execute();
	
	/**
	 * Redo this command. This executes the command based on the original input
	 * values that were gathered during execute(). This function must try to 
	 * produce the same results on every invocation.
	 */
	public void redo();
	
	/**
	 * Undo the effects of this command. Every effect of this command is 
	 * reverted by this function. This function is called with the document
	 * being in the state equal to the one immediately after this command's
	 * execute() function completed successfully. This function must restore
	 * the previous state of the data exactly and reliably every time it is 
	 * invoked.
	 */
	public void undo();
}
