/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Nov 9, 2008
 */

package eu.moonlight3d.common.ui.command;

import java.util.ArrayList;
import java.util.LinkedList;

/**
 * The command stack in a command pattern implementation. This command stack
 * is used to manage undo and redo of commands issued by the user. 
 *
 * @author gregor
 */
public class CommandStack {
	private static CommandStack instance;
	
	public static CommandStack getInstance() {
		if(instance==null) {
			instance=new CommandStack();
		}
		return instance;
	}
	
	/**
	 * The commands in the command stack.
	 */
	private LinkedList<Command> commands;
	/**
	 * The maximum allowed command stack length.
	 */
	private int maxCommandStackLength;
	/**
	 * The current position of the undo/redo pointer in the command stack.
	 * A value of -1 means that the pointer is at the top of the stack
	 * where new commands are inserted.
	 */
	private int undoPosition;
	/**
	 * The list of all listeners for this command stack.
	 */
	private ArrayList<CommandStackListener> commandStackListeners;
	/**
	 * Number of functions that are iterating over the list of listeners at this moment.
	 */
	private int iteratingListeners;
	/**
	 * An updated list of listeners that is created in the event that 
	 * registerCommandStackListener() or removeCommandStackListener() is 
	 * called while inside a function that is iterating over the list of 
	 * listeners.
	 */
	private ArrayList<CommandStackListener> updatedCommandStackListeners;
	
	/**
	 * Default constructor.
	 */
	private CommandStack() {
		commands=new LinkedList<Command>();
		maxCommandStackLength=10;
		undoPosition=-1;
		commandStackListeners=new ArrayList<CommandStackListener>();
		iteratingListeners=0;
	}
	
	/**
	 * Return the current size of the command stack.
	 * 
	 * @return size of the command stack
	 */
	public int getCommandStackLength() {
		return commands.size();
	}
	
	/**
	 * Return the maximum size that the command stack is allowed
	 * to reach.
	 * 
	 * @return maximum size of the command stack
	 */
	public int getMaxCommandStackLength() {
		return maxCommandStackLength;
	}

	/**
	 * notify all registered listeners that the maximum stack size has changed.
	 */
	private void notifyStackSizeChanged() {
		iteratingListeners++;
		try {
			for(CommandStackListener listener : commandStackListeners) {
				listener.stackSizeChanged();
			}
		} finally {
			iteratingListeners--;
			if(updatedCommandStackListeners!=null) {
				commandStackListeners=updatedCommandStackListeners;
				updatedCommandStackListeners=null;
			}
		}				
	}

	/**
	 * Set the maximum command stack size.
	 * 
	 * @param maxCommandStackLength
	 */
	public void setMaxCommandStackLength(int maxCommandStackLength) {
		while(commands.size()>maxCommandStackLength) {
			commands.removeFirst();
		}
		this.maxCommandStackLength = maxCommandStackLength;
		
		notifyStackSizeChanged();
	}
	
	/**
	 * Register a command stack listener.
	 * 
	 * @param listener the listener to register
	 */
	public void registerCommandStackListener(CommandStackListener listener) {
		if(iteratingListeners==0) {
			commandStackListeners.add(listener);
		} else {
			if(updatedCommandStackListeners==null) {
				updatedCommandStackListeners=(ArrayList<CommandStackListener>) commandStackListeners.clone();
			}
			updatedCommandStackListeners.add(listener);
		}		
	}
	
	/**
	 * Unregister a command stack listener.
	 * 
	 * @param listener the listener to unregister
	 */
	public void unregisterCommandStackListener(CommandStackListener listener) {
		if(iteratingListeners==0) {
			commandStackListeners.add(listener);
		} else {
			if(updatedCommandStackListeners==null) {
				updatedCommandStackListeners=(ArrayList<CommandStackListener>) commandStackListeners.clone();
			}
			updatedCommandStackListeners.remove(listener);
		}				
	}
	
	/**
	 * notify all registered listeners that a command is about to be executed
	 * 
	 * @param command the command to be executed
	 */
	private void notifyBeforeExecute(Command command) {
		iteratingListeners++;
		try {
			for(CommandStackListener listener : commandStackListeners) {
				listener.beforeExecute(command);
			}
		} finally {
			iteratingListeners--;
			if(updatedCommandStackListeners!=null) {
				commandStackListeners=updatedCommandStackListeners;
				updatedCommandStackListeners=null;
			}
		}		
	}
	
	/**
	 * notify all registered listeners that a command was executed.
	 * 
	 * @param command the command that was executed
	 * @param success true if the execution was successful, false otherwise
	 */
	private void notifyAfterExecute(Command command, boolean success) {
		iteratingListeners++;
		try {
			for(CommandStackListener listener : commandStackListeners) {
				listener.afterExecute(command,success);
			}
		} finally {
			iteratingListeners--;
			if(updatedCommandStackListeners!=null) {
				commandStackListeners=updatedCommandStackListeners;
				updatedCommandStackListeners=null;
			}
		}
	}
	
	/**
	 * Push a command onto the command stack and execute it. If the command
	 * is not executed successfully, it will not appear on the command stack.
	 * If it is not undoable the command stack will be cleared after the 
	 * successful execution of the command.
	 * 
	 * @param command the command to push and execute
	 */
	public void pushCommand(Command command) {
		// notify listeners before undo stack is modified
		notifyBeforeExecute(command);
		
		if(undoPosition!=-1) {
			// cut stack at current position, removing redo
			while(commands.size()>undoPosition) {
				commands.remove(undoPosition);
			}
			undoPosition=-1;
		}
		
		// execute command
		
		if(!command.execute()) {
			// command failed, claiming that current state is untouched - return without appending
			notifyAfterExecute(command,false);
			return;
		}
		
		// check if we just executed an undoable command
		if(command.isUndoable()) {
			// command is undoable, append to stack, keep stack size
			commands.addLast(command);
			if(commands.size()>maxCommandStackLength) {
				commands.removeFirst();
			}
		} else {
			commands.clear();
		}
		
		// notify listeners after undo stack has a consistent state again 
		notifyAfterExecute(command,true);		
	}
	
	/**
	 * Test if there is a command that can be undone.
	 * 
	 * @return true if there is an undoable command on the stack, false otherwise
	 */
	public boolean canUndo() {
		if(commands.size()>0 && undoPosition!=0) {
			return true;
		}
		
		return false;
	}
	
	/**
	 * notify all registered listeners that a command is about to be undone.
	 * 
	 * @param command the command to be undone
	 */
	private void notifyBeforeUndo(Command command) {
		iteratingListeners++;
		try {
			for(CommandStackListener listener : commandStackListeners) {
				listener.beforeUndo(command);
			}
		} finally {
			iteratingListeners--;
			if(updatedCommandStackListeners!=null) {
				commandStackListeners=updatedCommandStackListeners;
				updatedCommandStackListeners=null;
			}
		}
	}
	
	/**
	 * notify all registered listeners that a command was undone.
	 * 
	 * @param command the command that was undone
	 */
	private void notifyAfterUndo(Command command) {
		iteratingListeners++;
		try {
			for(CommandStackListener listener : commandStackListeners) {
				listener.afterUndo(command);
			}
		} finally {
			iteratingListeners--;
			if(updatedCommandStackListeners!=null) {
				commandStackListeners=updatedCommandStackListeners;
				updatedCommandStackListeners=null;
			}
		}
	}

	/**
	 * Undo the previous command. This will undo the previous command from
	 * the undo stack as long as there is one.
	 */
	public void undo() {
		if(!canUndo()) {
			return;
		}
		
		if(undoPosition==-1) {
			Command command=commands.getLast();
			
			notifyBeforeUndo(command);
			command.undo();
			undoPosition=commands.size()-2;
			notifyAfterUndo(command);
		} else {
			Command command=commands.get(undoPosition-1);
			
			notifyBeforeUndo(command);
			command.undo();
			undoPosition--;
			notifyAfterUndo(command);
		}
	}
	
	/**
	 * Test if there is a redoable command on the stack.
	 * 
	 * @return true if there is a redoable command, false otherwise
	 */
	public boolean canRedo() {
		if(undoPosition!=-1) {
			return true;
		}
		return false;
	}
	
	/**
	 * notify all registered listeners that a command is about to be redone.
	 * 
	 * @param command the command to be redone
	 */
	private void notifyBeforeRedo(Command command) {
		iteratingListeners++;
		try {
			for(CommandStackListener listener : commandStackListeners) {
				listener.beforeRedo(command);
			}
		} finally {
			iteratingListeners--;
			if(updatedCommandStackListeners!=null) {
				commandStackListeners=updatedCommandStackListeners;
				updatedCommandStackListeners=null;
			}
		}
	}
	
	/**
	 * notify all registered listeners that a command was redone.
	 * 
	 * @param command the command that was redone
	 */
	private void notifyAfterRedo(Command command) {
		iteratingListeners++;
		try {
			for(CommandStackListener listener : commandStackListeners) {
				listener.afterRedo(command);
			}
		} finally {
			iteratingListeners--;
			if(updatedCommandStackListeners!=null) {
				commandStackListeners=updatedCommandStackListeners;
				updatedCommandStackListeners=null;
			}
		}
	}

	/**
	 * Redo the next command on the stack if possible.
	 */
	public void redo() {
		if(!canRedo()) {
			return;
		}
		
		Command command=commands.get(undoPosition);
		
		notifyBeforeRedo(command);
		command.redo();
		
		undoPosition++;
		if(undoPosition==commands.size())  {
			undoPosition=-1;
		}
		
		notifyAfterRedo(command);
	}

	/**
	 * Get the command that the next call to undo() would process.
	 * 
	 * @return the command to be undone next or null if there is none
	 */
	public Command getCurrentUndoCommand() {
		if(!canUndo()) {
			return null;
		}
		return commands.get(undoPosition-1);
	}
	
	/**
	 * Get the command that the next call to redo() would process
	 * 
	 * @return the command to be redone or null of there is none
	 */
	public Command getCurrentRedoCommand() {
		if(!canRedo()) {
			return null;
		}
		return commands.get(undoPosition);
	}
}
