/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 13, 2005
 */
package eu.moonlight3d.common.ui.propertysheet;

import java.util.ArrayList;
import java.util.HashMap;

import com.trolltech.qt.gui.QHBoxLayout;
import com.trolltech.qt.gui.QVBoxLayout;
import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.framework.helper.NamedObject;
import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.helper.PropertyContainer;

/**
 * This is the View implementation for the property editor plugin.
 * 
 * @author gregor
 */
public class PropertySheet extends QWidget {
	/**
	 * The set of property editors that is used by this property sheet.
	 */
	private PropertyEditorFactorySet propertyEditorFactorySet;
	/**
	 * The {@link ExpandBar} widget which serves as the basis for the
	 * property sheet.
	 */
	private ExpandBar expandBar;
	/**
	 * A map between categories names and all the corresponding expand
	 * bar entries.
	 */
	private HashMap<String,QWidget> expandItems;
	/**
	 * A global list of all available property editor factories. 
	 */
    private static ArrayList<PropertyEditorFactory> propertyEditorFactories;
    /**
     * A marker that is used to determine whether the property sheet definitions have 
     * been loaded yet.
     */
    private static boolean propertySheetDefintionsLoaded=false;
    /**
     * List of all property editors
     */
    private ArrayList<ValueEditor> editors;
    /**
     * The list of registered property sheet listeners
     */
    private ArrayList<PropertySheetListener> listeners;
    /**
	 * An updated list of property sheet listeners that is created in the event that 
	 * addPropertySheetListener() or removePropertySheetListener() is called while 
	 * inside a function that is iterating over the list of listeners.
     */
    private ArrayList<PropertySheetListener> updatedListeners;
	/**
	 * Number of functions that are iterating over the list of property sheet listeners 
	 * at this moment.
	 */
	private int iteratingListeners;
    /**
     * The property container that the sheet is currently displaying an editor for
     */
    private PropertyContainer container;
    
    static {
        // default property editor factories:
    	addPropertyEditorFactory(new BooleanEditorFactory());
        addPropertyEditorFactory(new ColourEditorFactory());
        addPropertyEditorFactory(new FloatEditorFactory());
        addPropertyEditorFactory(new IntegerEditorFactory());
        addPropertyEditorFactory(new StringEditorFactory());
        addPropertyEditorFactory(new VectorEditorFactory());
        
        // special property editor factories:
        addPropertyEditorFactory(new DummyEditorFactory());
        addPropertyEditorFactory(new IntegerMapFactory());
        addPropertyEditorFactory(new DirectoryListEditorFactory());
    }
    
    /**
     * Add a property editor factory to the global list of factories.
     * 
     * @param editorFactory the property editor factory to add
     */
    public static void addPropertyEditorFactory(PropertyEditorFactory editorFactory) {
        if(propertyEditorFactories==null) {
            propertyEditorFactories=new ArrayList<PropertyEditorFactory>();
        }
        propertyEditorFactories.add(editorFactory);
    }
    
    /**
     * Get the global list of property editor factories
     * 
     * @return list of property editor factories
     */
    public static ArrayList<PropertyEditorFactory> getPropertyEditorFactories() {
    	return (ArrayList<PropertyEditorFactory>) propertyEditorFactories.clone();
    }
    
	/**
	 * Default constructor for the property sheet widget
	 * 
	 * @param parent the parent of this widget
	 */
	public PropertySheet(QWidget parent) {
		super(parent);
		
		if(!propertySheetDefintionsLoaded) {
	        PropertySheetDefinition.parsePropertySheetDefinitions();			
	        propertySheetDefintionsLoaded=true;
		}
		
		expandBar=new ExpandBar(this);

		QHBoxLayout layout=new QHBoxLayout();
		layout.addWidget(expandBar);
		layout.setMargin(0);
		setLayout(layout);

		expandItems=new HashMap<String,QWidget>();
        editors=new ArrayList<ValueEditor>();
        listeners=new ArrayList<PropertySheetListener>();
        iteratingListeners=0;
        
        propertyEditorFactorySet=PropertyEditorFactorySet.getDefaultPropertyEditorFactorySet();
	}
	
	/**
	 * Set the property editor factory set that will be used by this property sheet for all
	 * future property containers.
	 * 
	 * @param propertyEditorFactorySet the property editor factory set to use
	 */
	public void setPropertyEditorFactorySet(PropertyEditorFactorySet propertyEditorFactorySet) {
		this.propertyEditorFactorySet=propertyEditorFactorySet;
	}
	
	/**
	 * Register a property sheet listener. The listener will be called whenever
	 * a change happens in the property editor.
	 * 
	 * @param listener the listener to register
	 */
	public void addPropertySheetListener(PropertySheetListener listener) {
		if(iteratingListeners==0) {
			listeners.add(listener);
		} else {
			updatedListeners=(ArrayList<PropertySheetListener>) listeners.clone();
			updatedListeners.add(listener);
		}
	}
	
	/**
	 * Unregister a property sheet listener. The listener will no longer be
	 * called from this point forward.
	 * 
	 * @param listener the listener to unregister
	 */
	public void removePropertySheetListener(PropertySheetListener listener) {
		if(iteratingListeners==0) {
			listeners.remove(listener);
		} else {
			updatedListeners=(ArrayList<PropertySheetListener>) listeners.clone();
			updatedListeners.remove(listener);
		}
	}
	
	/**
	 * Get a list of properties which are not accounted for in the property
	 * sheet definition that was used for building the property sheet.
	 * 
	 * @param container the property container for which the sheet is built
	 * @param definition the corresponding property sheet definition
	 * @return a list of all properties which are not listed in the property 
	 * sheet definition 
	 */
	private ArrayList<String> getUncoveredProperties(PropertyContainer container, PropertySheetDefinition definition) {
		ArrayList<String> uncoveredProperties=new ArrayList<String>();
		ArrayList<String> coveredProperties=definition.knownProperties();
		ArrayList<String> allProperties=container.getPropertyList();
		
		for(String propertyName : allProperties) {
			if(!coveredProperties.contains(propertyName)) {
				uncoveredProperties.add(propertyName);
			}
		}
		
		return uncoveredProperties;
	}
	
	/**
	 * Build the contents widgets of the property sheet
	 * 
	 * @param container the property container to build the widgets for
	 */
	private void buildSheet(PropertyContainer container) {
		PropertySheetDefinition sheetDefinition=null;
		
		// find matching property sheet definiton
		for(PropertySheetDefinition definition : PropertySheetDefinition.getPropertySheetDefinitions()) {
			if(container.getClass().getName().equals(definition.getTypeName())) {
				if(!definition.getInstanceName().equals("")) {
					if(container instanceof NamedObject) {
						if(((NamedObject)container).getName().equals(definition.getInstanceName())) {
							sheetDefinition=definition;
							break;
						}
					}
				} else {
					sheetDefinition=definition;
					break;
				}
			}
		}
		
		ArrayList<String> properties;
		if(sheetDefinition!=null) {
			for(PropertySheetDefinition.EditorGroup editorGroup : sheetDefinition.getEditorGroups()) {
				// build expand bar items for property groups
				ExpandItem item=new ExpandItem();
				QWidget groupComposite=new QWidget();
				QVBoxLayout layout=new QVBoxLayout();
				layout.setSpacing(1);
				layout.setMargin(2);
				groupComposite.setLayout(layout);

				// build editors for definitions
				for(PropertySheetDefinition.EditorDefinition editorDefinition : editorGroup.editors) {
					for(PropertyEditorFactory factory : propertyEditorFactories) {
						if(factory.getName().equals(editorDefinition.editor)) {
			            	ValueEditor editor = factory.generatePropertyEditor(this,groupComposite,container,editorDefinition.properties,editorDefinition.additionalParameters);
			            	layout.addWidget(editor);
			            	editors.add(editor);
							break;
						}
					}
				}
				item.setName(editorGroup.caption);
				item.setChildWidget(groupComposite);
				expandItems.put(editorGroup.caption,groupComposite);
				item.setExpanded(true);
				expandBar.addExpandItem(item);
			}

			properties=getUncoveredProperties(container,sheetDefinition);
		} else {
			properties=container.getPropertyList();
		}
		
		if(properties.size()==0) {
			// if there are no more properties left, we're done here
			return;
		}
		
		// build expandbar item for "Other Properties"
		ExpandItem item=new ExpandItem();
		QWidget groupComposite=new QWidget();
		QVBoxLayout layout=new QVBoxLayout();
		layout.setSpacing(1);
		layout.setMargin(2);
		groupComposite.setLayout(layout);

		for(int i=0;i<properties.size();i++) {
            // build user interface for this property
            Property property=container.getProperty(properties.get(i));
            if(propertyEditorFactorySet.getDefaultPropertyEditorFactory(property.getType())!=null) {
            	PropertyEditorFactory factory=propertyEditorFactorySet.getDefaultPropertyEditorFactory(property.getType());
            	ArrayList<String> propertyName=new ArrayList<String>();
            	propertyName.add(property.getName());
            	ValueEditor editor = factory.generatePropertyEditor(this,groupComposite,container,propertyName,null);
            	layout.addWidget(editor);
            	editors.add(editor);
            }
        }
		item.setName("Other Properties");
		item.setChildWidget(groupComposite);
		expandItems.put("Other Properties",groupComposite);
		item.setExpanded(true);
		expandBar.addExpandItem(item);
	}
	
	/**
	 * Set the property container whose properties should be edited.
	 * 
	 * @param container the property container to manipulate 
	 */
	public void setPropertyContainer(PropertyContainer container)
	{
		this.container=container;
		
		expandBar.clear();

		expandItems.clear();

		editors.clear();
        
		if(container!=null) {
			buildSheet(container);
		}
        
		iteratingListeners++;
		try {
			for(PropertySheetListener listener : listeners) {
				listener.propertyContainerChanged(container);
			}
		} finally {
			iteratingListeners--;
			if(updatedListeners!=null) {
				listeners=updatedListeners;
				updatedListeners=null;
			}
		}
	}
	
	/**
	 * Get the current property container or null if there is 
	 * no property container at the current time.
	 * 
	 * @return current property container or null
	 */
	public PropertyContainer getPropertyContainer() {
		return container;
	}
	
	/**
	 * Notify all registered listeners that the given property has been
	 * updated.
	 * 
	 * @param property the property which has been changed
	 */
	public void notifyPropertyUpdate(Property property) {
		iteratingListeners++;
		try {
			for(PropertySheetListener listener : listeners) {
        		listener.propertyValueUpdated(property);
			}
		} finally {
			iteratingListeners--;
			if(updatedListeners!=null) {
				listeners=updatedListeners;
				updatedListeners=null;
			}
		}
	}
	
	/**
	 * Force an update of all displayed property values.
	 */
	public void updateValues() {
		for(ValueEditor editor: editors) {
			editor.updateFromPropertyValue();
		}
	}
}
