/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Sep 24, 2005
 */
package eu.moonlight3d.common.ui.propertysheet;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import com.trolltech.qt.core.Qt;
import com.trolltech.qt.gui.QDoubleValidator;
import com.trolltech.qt.gui.QGridLayout;
import com.trolltech.qt.gui.QLabel;
import com.trolltech.qt.gui.QLineEdit;
import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.helper.PropertyContainer;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.math.Vector3D;

public class VectorEditorFactory implements PropertyEditorFactory {
	private class Parameters {
		public String labelText;
	}
	
    public class VectorValueEditor extends ValueEditor {
   		PropertySheet propertySheet;
        Property property;
        QLineEdit entry[];

        VectorValueEditor(Parameters parameters, PropertySheet sheet, QWidget parent, final Property property)
        {
            super(parent);
            propertySheet=sheet;
            this.property=property;
            
            QLabel label;
            QLabel coordinateLabel;

            QGridLayout layout=new QGridLayout();
            
            label=new QLabel(this);
            if(parameters.labelText!=null) {
            	label.setText(parameters.labelText);
            } else {
            	label.setText(property.getName());
            }
            
            layout.addWidget(label, 0, 0, 1, 0, Qt.AlignmentFlag.AlignLeft);
            
            entry=new QLineEdit[3];
            for (int i = 0; i < 3; i++) {
				coordinateLabel = new QLabel(this);
				switch (i) {
				case 0:
					coordinateLabel.setText("x");
					break;
				case 1:
					coordinateLabel.setText("y");
					break;
				case 2:
					coordinateLabel.setText("z");
					break;
				}
				
				layout.addWidget(coordinateLabel,1+i,0,Qt.AlignmentFlag.AlignLeft);

				entry[i] = new QLineEdit(this);	            
	            entry[i].setValidator(new QDoubleValidator(entry[i]));
				entry[i].setText(getPropertyValue(i));
				// note that weird little trick here: dynamically generated slot names!
				entry[i].textEdited.connect(this,"textEdited"+i+"()");
				entry[i].editingFinished.connect(this,"textEdited"+i+"()");
				layout.addWidget(entry[i],1+i,2,Qt.AlignmentFlag.AlignLeft);
			}
            
            layout.setColumnStretch(1,1);
            layout.setMargin(0);
            layout.setSpacing(2);
            setLayout(layout);
        }
        
        public void textEdited0() {
            setPropertyValue(0,entry[0].text());
          	propertySheet.notifyPropertyUpdate(property);        	
        }
        
        public void textEdited1() {
            setPropertyValue(1,entry[1].text());
          	propertySheet.notifyPropertyUpdate(property);        	        	
        }
        
        public void textEdited2() {
            setPropertyValue(2,entry[2].text());
          	propertySheet.notifyPropertyUpdate(property);        	        	
        }
        
        public String getPropertyValue(int numEntry)
        {
            try {
            	if(numEntry==0) {
                	DecimalFormat decimalFormat=new DecimalFormat();
                	decimalFormat.applyPattern("0.######");
                	return decimalFormat.format(property.getVectorValue().X1);
            	} else if(numEntry==1) {
                	DecimalFormat decimalFormat=new DecimalFormat();
                	decimalFormat.applyPattern("0.######");
                	return decimalFormat.format(property.getVectorValue().X2);
            	} else {
                	DecimalFormat decimalFormat=new DecimalFormat();
                	decimalFormat.applyPattern("0.######");
                	return decimalFormat.format(property.getVectorValue().X3);
            	}
            } catch (Exception e) {
                ExceptionDialog.run(e);
                return "";
            }
        }
    
        public void setPropertyValue(int numEntry, String value)
        {
        	double doubleValue;
            try {
            	doubleValue=NumberFormat.getInstance().parse(value).doubleValue();
                Vector3D vector=property.getVectorValue();
                if(numEntry==0) {
                    vector.X1=doubleValue;                	
                } else if(numEntry==1) {
                    vector.X2=doubleValue;                	
                } else {
                    vector.X3=doubleValue;                	
                }
                property.setValue(vector);
            } catch(NumberFormatException e) {
            	// ignore this exception
            } catch (Exception e) {
                ExceptionDialog.run(e);
            }        
        }

		@Override
		public void updateFromPropertyValue() {
			try {
			for(int i=0;i<3;i++) {
				double propertyValue;
            	if(i==0) {
                    propertyValue=property.getVectorValue().X1;            		 
            	} else if(i==1) {
                    propertyValue=property.getVectorValue().X2;            		
            	} else {
                    propertyValue=property.getVectorValue().X3;            		
            	}
            	double editValue=NumberFormat.getInstance().parse(entry[i].text()).doubleValue();
				if(editValue!=propertyValue) {
					entry[i].setText(getPropertyValue(i));
				}
			}
			} catch(Exception e) {
				ExceptionDialog.run(e);
			}
		}
    }

	public String getName() {
		return "VectorEditor";
	}

	public Object parseExtraParameters(Element parameterRoot) throws Exception {
		Parameters parameters=new Parameters();
		
		if(parameterRoot==null) {
			return parameters;
		}

		NodeList labelList=parameterRoot.getElementsByTagName("label");
		if(labelList.getLength()>1) {
			throw new Exception(L10N.translate("only one label can be defined for a vector property editor"));
		} else if(labelList.getLength()==1) {
			Element labelElement=(Element) labelList.item(0);
			parameters.labelText=labelElement.getAttribute("text");
		}
		
		return parameters;
	}

	public ValueEditor generatePropertyEditor(PropertySheet sheet, QWidget parent, PropertyContainer container, ArrayList<String> properties, Object additionalParams) {
		Property property=container.getProperty(properties.get(0));
		if(additionalParams==null) {
			additionalParams=new Parameters();
		}
        return new VectorValueEditor((Parameters)additionalParams,sheet,parent,property);
	}
}
