/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * Created on Mar 14, 2007
 */
package eu.moonlight3d.framework.document;

import eu.moonlight3d.framework.helper.FileType;
import eu.moonlight3d.framework.helper.PropertyContainer;

/**
 * A factory for a particular type of documents. It handles
 * creating empty documents and loading of existing ones.
 *
 * @author gregor
 */
public interface DocumentFactory {
	
	/**
	 * Get a unique type identification of this document type
	 * for internal use
	 * 
	 * @return a unique identifier for this document type 
	 */
	public String getTypeName();
	
	/**
	 * Return a user-readable identification of this document type.
	 * 
	 * @return name of document type
	 */
	public String getDisplayName();
	
	/**
	 * Return the file type for this document type.
	 * 
	 * @return the FileType describing this document type
	 * @see eu.moonlight3d.framework.helper.FileType
	 */
	public FileType getFileType();
	
	/**
	 * Return true if this document factory is capable of creating
	 * new empty documents. Any implementation that operates on
	 * read-only documents will probably want to return false here.
	 * 
	 * @return true if the factory is able to create new documents
	 */
	public boolean canCreateNew();
	
	/**
	 * Return true if this document factory is capable of saving
	 * documents. Any implementation that operates as a part of
	 * a document viewer or something similar might want to return
	 * false to disable saving.
	 * 
	 * @return true if saving is possible, false otherwise
	 */
	public boolean canSaveFiles();
	
	/**
	 * Create a new empty document of this type and return it.
	 * This will use default options for creating the new
	 * document.
	 *  
	 * @return a new document
	 */
	Document createNewDocument();
	
	/**
	 * Create a new empty document of this type and return it.
	 * The options passed to this function are either generated
	 * by createNewDocumentProperties() or null if the document
	 * type does not have any options.
	 *  
	 * @param options options for this document type
	 * @return a new document
	 */
	Document createNewDocument(PropertyContainer options);

	/**
	 * Create a property container containing all options the
	 * user has when creating a new, empty document of this
	 * type. The property container must be populated with
	 * valid default values. If there are no options for
	 * this document type, null is returned.
	 *  
	 * @return property container with all options for new 
	 * documents of this type
	 */
	public PropertyContainer createNewDocumentProperties();
	
	/**
	 * Do a quick check if the file with the given name can
	 * be loaded by this document factory.
	 * 
	 * @param filename the file to check
	 * @return true if the file can be opened
	 */
	public boolean canOpenDocument(String filename);
	
	/**
	 * Load the data from the file into a new document and 
	 * return it.
	 * 
	 * @param filename the name of the file to load
	 * @return the document representing the opened file
	 * @throws Exception 
	 */
	Document openDocument(String filename) throws Exception;
}
