/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package eu.moonlight3d.framework.document;

import java.util.ArrayList;
import java.util.List;

import eu.moonlight3d.framework.helper.PropertyContainer;

/**
 * This class manages all document factories and currently open 
 * documents. The document manager should only be accessed 
 * through the {@link eu.moonlight3d.framework.State} singleton.<br> 
 * 
 * At any given time there can only be one main
 * document, but the number of additional open documents 
 * is unlimited. The main document is defined as the document
 * that the user works on. 
 *
 * @author gregor
 */
public class DocumentManager {
	/**
	 * All registered document lifecycle listeners.
	 */
	private ArrayList<LifecycleListener> listeners;
	/**
	 * Number of functions that are iterating over the list of listeners at this moment.
	 */
	private int iteratingListeners;
	/**
	 * An updated list of listeners that is created in the event that addLifecycleListener()
	 * or removeLifecycleListener() is called while inside a function that is iterating
	 * over the list of listeners.
	 */
	private ArrayList<LifecycleListener> updatedListeners;
	/**
	 * All registered document factories.
	 */
	private ArrayList<DocumentFactory> documentFactories;
	/**
	 * All open documents (not including the main document). 
	 */
	private ArrayList<Document> openDocuments;
	/**
	 * The main document.
	 */
	private Document mainDocument;
	/**
	 * The name of the default document factory.
	 */
	private String defaultDocumentFactory;
	
	/**
	 * Default constructor.
	 */
	public DocumentManager() {
		listeners=new ArrayList<LifecycleListener>();
		iteratingListeners=0;
		documentFactories=new ArrayList<DocumentFactory>();
		openDocuments=new ArrayList<Document>();
	}
	
	/**
	 * Register a new document factory.
	 * 
	 * @param factory the factory to register
	 * 
	 * @see DocumentFactory
	 */
	public void registerDocumentFactory(DocumentFactory factory) {
		documentFactories.add(factory);
	}
	
	/**
	 * Unregister a document factory
	 * 
	 * @param factory the factory to unregister
	 * 
	 * @see DocumentFactory
	 */
	public void unregisterDocumentFactory(DocumentFactory factory) {
		documentFactories.remove(factory);
	}
	
	/**
	 * Return a list of registered document factories.
	 * 
	 * @return registered document factories
	 */
	public List<DocumentFactory> getDocumentFactories() {
		return (ArrayList<DocumentFactory>) documentFactories.clone();
	}
	
	/**
	 * Get a specific document factory by name.
	 * 
	 * @param type the name of the document type
	 * @return the factory corresponding to this document type
	 */
	public DocumentFactory getDocumentFactory(String type) {
		for(DocumentFactory factory : documentFactories) {
			if(factory.getTypeName().equals(type)) {
				return factory;
			}
		}
		return null;
	}
	
	/**
	 * Create a new main document of the given type.
	 * 
	 * @param type the type of the new main document
	 * @return the new main document
	 */
	public Document createNewMainDocument(String type) {
		// close current main document if neccessary
		closeMainDocument();
		// create new main document if possible
		for(DocumentFactory factory : documentFactories) {
			if(factory.getTypeName().equals(type)) {
				triggerBeforeNew();
				mainDocument=factory.createNewDocument();
				triggerAfterNew();
				mainDocument.update();
				return mainDocument;
			}
		}
		return null;
	}
	
	/**
	 * Create a new main document of the given type.
	 * The options passed to this function are either 
	 * generated by 
	 * {@link DocumentFactory#createNewDocumentProperties()} 
	 * or null if the document type does not have any options.
	 *  
	 * @param type the type of the new main document
	 * @param options options for this document type
	 * @return the new main document
	 */
	public Document createNewMainDocument(String type, PropertyContainer options) {
		// close current main document if neccessary
		closeMainDocument();
		// create new main document if possible
		for(DocumentFactory factory : documentFactories) {
			if(factory.getTypeName().equals(type)) {
				triggerBeforeNew();
				mainDocument=factory.createNewDocument(options);
				triggerAfterNew();
				mainDocument.update();
				return mainDocument;
			}
		}
		return null;
	}
	
	/**
	 * Open the file identified by the given filename as the
	 * new main document. If a main document is currently
	 * loaded, it is closed prior to any attempts to load the
	 * new file.
	 * 
	 * @param filename file name of new main document
	 * @return the new main document
	 * @throws Exception if an error happens during the loading
	 */
	public Document openMainDocument(String filename) throws Exception {
		closeMainDocument();
		DocumentFactory documentFactory=null;
		for(DocumentFactory factory : documentFactories) {
			if(factory.getFileType().matches(filename)) {
				if(factory.canOpenDocument(filename)) {
					documentFactory=factory;
				}
			}
		}
		if(documentFactory==null) {
			for(DocumentFactory factory : documentFactories) {
				if(factory.canOpenDocument(filename)) {
					documentFactory=factory;
				}
			}			
		}
		if(documentFactory==null) {
			return null;
		}
		triggerBeforeOpen(filename);
		mainDocument=documentFactory.openDocument(filename);
		triggerAfterOpen(filename);
		mainDocument.update();
		return mainDocument;
	}
	
	/**
	 * Save the current main docuemnt if it has a file name
	 * assigned to it. Otherwise the saving is aborted 
	 * silently.
	 * 
	 * @throws Exception if an error happens during the saving
	 */
	public void saveMainDocument() throws Exception {
		if(mainDocument!=null) {
			if(mainDocument.getFilename()!=null) {
				triggerBeforeSave(mainDocument.getFilename());
				mainDocument.save(mainDocument.getFilename());
				triggerAfterSave(mainDocument.getFilename());
			}
		}
	}
	
	/**
	 * Save the current main document to the given file name.
	 * This file name is also assigned to the main document.
	 * 
	 * @param filename filename to save the docuemnt to
	 * @throws Exception if an error happens during the saving
	 */
	public void saveMainDocument(String filename) throws Exception {
		if(mainDocument!=null) {
			triggerBeforeSave(filename);
			mainDocument.save(filename);
			triggerAfterSave(filename);
		}
	}

	/**
	 * Close the current main document.
	 */
	public void closeMainDocument() {
		if(mainDocument!=null) {
			triggerOnClose();
			mainDocument.close();
			mainDocument=null;
		}
	}
	
	/**
	 * Get the current main document.
	 * 
	 * @return the current main document
	 */
	public Document getMainDocument() {
		return mainDocument;
	}
	
	/**
	 * Create a new document. This document is inserted into
	 * the list of open documents.
	 * 
	 * @param type type name of the new document
	 * @return the new document
	 */
	public Document createNewDocument(String type) {
		for(DocumentFactory factory : documentFactories) {
			if(factory.getTypeName().equals(type)) {
				Document document=factory.createNewDocument();
				openDocuments.add(document);
				return document;
			}
		}
		return null;		
	}
	
	/**
	 * Create a new document. This document is inserted into
	 * the list of open documents. The options passed to this 
	 * function are either generated by 
	 * {@link DocumentFactory#createNewDocumentProperties()} 
	 * or null if the document type does not have any options.
	 *  
	 * @param type type name of the new document
	 * @param options options for this document type
	 * @return the new document
	 */
	public Document createNewDocument(String type, PropertyContainer options) {
		for(DocumentFactory factory : documentFactories) {
			if(factory.getTypeName().equals(type)) {
				Document document=factory.createNewDocument(options);
				openDocuments.add(document);
				return document;
			}
		}
		return null;		
	}
	
	/**
	 * Open the file identified by the given file name and keep
	 * the document as an open document.
	 * 
	 * @param filename the file name of the file to load 
	 * @return the new document
	 * @throws Exception if an error happens during loading
	 */
	public Document openDocument(String filename) throws Exception {
		DocumentFactory documentFactory=null;
		for(DocumentFactory factory : documentFactories) {
			if(factory.getFileType().matches(filename)) {
				if(factory.canOpenDocument(filename)) {
					documentFactory=factory;
				}
			}
		}
		if(documentFactory==null) {
			for(DocumentFactory factory : documentFactories) {
				if(factory.canOpenDocument(filename)) {
					documentFactory=factory;
				}
			}			
		}
		if(documentFactory==null) {
			return null;
		}
		Document document=documentFactory.openDocument(filename);
		openDocuments.add(document);
		return document;		
	}
	
	/**
	 * Save the given document if it has a file name assigned.
	 * Otherwise, the saving is skipped silently.
	 * 
	 * @param document the document to save
	 * @throws Exception if an error happens during saving
	 */
	public void saveDocument(Document document) throws Exception {
		if(document.getFilename()!=null) {
			document.save(document.getFilename());
		}
	}
	
	/**
	 * Save the given document to the given filename.
	 * 
	 * @param document the document to save
	 * @param filename the file to save the document to
	 * @throws Exception if an error happens during saving
	 */
	public void save(Document document, String filename) throws Exception {
		if(document.getFilename()!=null) {
			document.save(filename);
		}		
	}
	
	/**
	 * Close the given document.
	 * 
	 * @param document the document to save
	 */
	public void closeDocument(Document document) {
		document.close();
		openDocuments.remove(document);
	}
	
	/**
	 * Get a list of all documents that are currently open.
	 * This list does not include the main document.
	 * 
	 * @return list of all open documents
	 */
	public List<Document> getOpenDocuments() {
		return (ArrayList<Document>) openDocuments.clone();
	}
	
	/**
	 * Register a document lifecycle listener.
	 * 
	 * @param listener the listener to register.
	 * 
	 * @see LifecycleListener
	 */
	public void addLifecycleListener(LifecycleListener listener) {
		if(iteratingListeners==0) {
			listeners.add(listener);
		} else {
			if(updatedListeners==null) {
				updatedListeners=(ArrayList<LifecycleListener>) listeners.clone();
			}
			updatedListeners.add(listener);
		}
	}
	
	/**
	 * Unregister a lifecycle listener.
	 *
	 * @param listener the listener to unregister
	 * 
	 * @see LifecycleListener
	 */
	public void removeLifecycleListener(LifecycleListener listener) {
		if(iteratingListeners==0) {
			listeners.remove(listener);
		} else {
			if(updatedListeners==null) {
				updatedListeners=(ArrayList<LifecycleListener>) listeners.clone();
			}
			updatedListeners.remove(listener);			
		}
	}
	
	/**
	 * Set the name of default document factory.
	 * 
	 * @param defaultDocumentFactory the name of the default document factory
	 */
	public void setDefaultDocumentFactory(String defaultDocumentFactory) {
		this.defaultDocumentFactory=defaultDocumentFactory;
	}
	
	/**
	 * Get the name of the default document factory
	 * 
	 * @return name of default document factory
	 */
	public String getDefaultDocumentFactory() {
		return defaultDocumentFactory;
	}

	/**
	 * Trigger beforeNew event on all registered listeners.
	 */
	private void triggerBeforeNew() {
		iteratingListeners++;
		try {
			for(LifecycleListener listener : listeners) {
				listener.beforeNew();
			}
		} finally {
			iteratingListeners--;
			if(updatedListeners!=null) {
				listeners=updatedListeners;
				updatedListeners=null;
			}
		}
	}
	
	/**
	 * Trigger afterNew event on all registered listeners.
	 */
	private void triggerAfterNew() {
		iteratingListeners++;
		try {
			for(LifecycleListener listener : listeners) {
				listener.afterNew();
			}
		} finally {
			iteratingListeners--;
			if(updatedListeners!=null) {
				listeners=updatedListeners;
				updatedListeners=null;
			}
		}
	}
	
	/**
	 * Trigger beforeOpen event on all registered listeners.
	 * 
	 * @param filename the file name of the file that is opened
	 */
	private void triggerBeforeOpen(String filename) {
		iteratingListeners++;
		try {
			for(LifecycleListener listener : listeners) {
				listener.beforeOpen(filename);
			}
		} finally {
			iteratingListeners--;
			if(updatedListeners!=null) {
				listeners=updatedListeners;
				updatedListeners=null;
			}
		}
	}

	/**
	 * Trigger afterOpen event on all registered listeners.
	 * 
	 * @param filename the file name of the file that is opened
	 */
	private void triggerAfterOpen(String filename) {
		iteratingListeners++;
		try {
			for(LifecycleListener listener : listeners) {
				listener.afterOpen(filename);
			}
		} finally {
			iteratingListeners--;
			if(updatedListeners!=null) {
				listeners=updatedListeners;
				updatedListeners=null;
			}
		}
	}

	/**
	 * Trigger beforeSave event on all registered listeners.
	 * 
	 * @param filename the file name of the file that is saved
	 */
	private void triggerBeforeSave(String filename) {
		iteratingListeners++;
		try {
			for(LifecycleListener listener : listeners) {
				listener.beforeSave(filename);
			}
		} finally {
			iteratingListeners--;
			if(updatedListeners!=null) {
				listeners=updatedListeners;
				updatedListeners=null;
			}
		}
	}

	/**
	 * Trigger afterSave event on all registered listeners.
	 * 
	 * @param filename the file name of the file that is saved 
	 */
	private void triggerAfterSave(String filename) {
		iteratingListeners++;
		try {
			for(LifecycleListener listener : listeners) {
				listener.afterSave(filename);
			}
		} finally {
			iteratingListeners--;
			if(updatedListeners!=null) {
				listeners=updatedListeners;
				updatedListeners=null;
			}
		}
	}

	/**
	 * Trigger onClose event on all registered listeners.
	 */
	private void triggerOnClose() {
		iteratingListeners++;
		try {
			for(LifecycleListener listener : listeners) {
				listener.onClose();
			}
		} finally {
			iteratingListeners--;
			if(updatedListeners!=null) {
				listeners=updatedListeners;
				updatedListeners=null;
			}
		}
	}
}
