/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Jul 3, 2008
 */

package eu.moonlight3d.framework.helper;

/**
 * Internal base class for all array buffer classes. This
 * class defines common methods and interfaces for all of
 * these classes.
 *
 * @author gregor
 */
public abstract class ArrayBufferBase extends BufferBase {

	/**
	 * The number of arrays that are actually used, or in other
	 * words, the real length of the array of arrays.
	 */
	protected int used;
	/**
	 * The number of arrays to hold reserve space for.
	 */
	protected int reserve;

	/**
	 * Default constructor.
	 */
	public ArrayBufferBase() {
		super();
	}

	/**
	 * Resize an array that is stored in the buffer.
	 * 
	 * @param elementIndex the index of the element to be resized
	 * @param newSize the new size of the element
	 */
	public abstract void resize(int elementIndex, int newSize);

	/**
	 * Remove the array with the given index from the buffer. Note
	 * that the element indices for all subsequent elements in the
	 * buffer is implicitly decreased by one by this operation.
	 * 
	 * @param elementIndex the index of the array to be removed
	 */
	@Override
	public abstract void remove(int elementIndex);

	/**
	 * Recalculate the necessary space for the buffer to hold all
	 * the elements and provide room for expansion without having
	 * to reallocate the array too often. This function does make
	 * a tradeoff between performance and allocated space.
	 * 
	 * @param currentSize the current size of the buffer
	 * @return the calculated new size of the buffer 
	 */
	@Override
	protected int recalculateSize(int currentSize) {
		int newLength=-1;
		
		int requiredSize=used+reserve;
		if(currentSize<requiredSize) {
			if(requiredSize<100) {
				newLength=requiredSize*2;
			} else {
				newLength=(int)(requiredSize*1.1);
			}
		} else {
			// check if there is too much capacity reserved
			if(currentSize<100) {
				if(currentSize>2.5*(requiredSize)) {
					newLength=2*(requiredSize);
				}
			} else {
				if(currentSize>1.2*(requiredSize)) {
					newLength=(int) (1.1*(requiredSize));
				}
			}
		}
		
		return newLength;
	}
	

	/**
	 * Return the number of arrays in this buffer.
	 * 
	 * @return number of arrays stored in this buffer
	 */
	public int getNumArrays() {
		return used;
	}

	/**
	 * Request a certain amount of reserve space. The buffer
	 * will be resized immediately to provide enough space to 
	 * fit in the reserve. This is useful if a known large 
	 * number of elements needs to be added to this buffer at 
	 * once.
	 * 
	 * @param numElements number of arrays to reserve space for
	 * @param numArrayElements number of array elements to reserve space for
	 */
	public void setReserve(int numElements, int numArrayElements) {
		reserve=numElements;
		checkCapacity();
	}

}