/*
 * Moonlight|3D
 * Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Nov 2, 2007
 */

package eu.moonlight3d.framework.helper;

import eu.moonlight3d.framework.localization.L10N;

/**
 * Internal base class with common functionality required by all
 * buffer classes. 
 *
 * @author gregor
 */
abstract class BufferBase {
	/**
	 * The number of elements that are actually used, or in other
	 * words, the array length.
	 */
	protected int used;
	/**
	 * The number of elements to hold reserve space for.
	 */
	protected int reserve;
	
	/**
	 * Check the given element index for validity.
	 * 
	 * @param index array index to check
	 * @throws IllegalArgumentException if index is not valid
	 */
	protected void checkIndex(int index) {
		if(index < 0 || index >=used) {
			throw new IllegalArgumentException(L10N.translate("array index out of range"));
		}
	}
	
	/**
	 * Recalculate the necessary space for the buffer to hold all
	 * the elements and provide room for expansion without having
	 * to reallocate the array too often. This function does make
	 * a tradeoff between performance and allocated space.
	 * 
	 * @param currentSize the current size of the buffer
	 * @return the new size of the buffer
	 */
	protected int recalculateSize(int currentSize) {
		int newLength=-1;
		
		int requiredSize=used+reserve;
		if(currentSize<requiredSize) {
			if(requiredSize<100) {
				newLength=requiredSize*2;
			} else {
				newLength=(int)(requiredSize*1.1);
			}
		} else {
			// check if there is too much capacity reserved
			if(currentSize<100) {
				if(currentSize>2.5*(requiredSize)) {
					newLength=2*(requiredSize);
				}
			} else {
				if(currentSize>1.2*(requiredSize)) {
					newLength=(int) (1.1*(requiredSize));
				}
			}
		}
		
		return newLength;
	}
	
	/**
	 * Check the capacity of the data array and resize it
	 * if it becomes either too small or too large.
	 */
	protected abstract void checkCapacity();

	/**
	 * Remove the element at the given index from the array.
	 * Note that the element indices for all subsequent elements
	 * in the buffer is implicitly decreased by one by this operation.
	 * 
	 * @param elementIndex the index of the element to be removed
	 */
	public abstract void remove(int elementIndex);

	/**
	 * Request a certain amount of reserve space. The buffer
	 * will be resized immediately to provide enough space to 
	 * fit in the reserve. This is useful if a known large 
	 * number of elements needs to be added to this buffer at 
	 * once.
	 * 
	 * @param reserve the amount of elements to reserve space for
	 */
	public void setReserve(int reserve) {
		this.reserve=reserve;
		// ensure capacity
		checkCapacity();
	}

	/**
	 * Get the total length of the buffer, i.e. the total number of array
	 * elements stored in it.
	 * 
	 * @return number of array elements in all arrays
	 */
	public int getLength() {
		return used;
	}

}