/*
 * Moonlight|3D
 * Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Nov 2, 2007
 */

package eu.moonlight3d.framework.helper;

import java.util.Arrays;

/**
 * A buffer that acts as a fast storage for multiple individual,
 * tiny arrays while trying to avoid the overhead imposed by the
 * virtual machine on each array. 
 * <br/>
 * Arrays stored in this buffer are addressed by an element
 * index. This element index is unique for each array. However,
 * because this buffer acts like an array of arrays in some ways,
 * element indices will change as a side effect of certain 
 * operations just as one would expect from an array of arrays.
 * <br/>
 * The buffer works internally by merging the individual arrays 
 * into a single one-dimensional data array with the help of an 
 * additional index array. 
 *
 * @author gregor
 */
/*
 * A note here on the implementation: the currently implemented 
 * management scheme is absolutely naive and better schemes are
 * conceivable. For example, it should be quite possible to
 * use the reserve space as a work area, copying growing arrays
 * there as well as skip the immediate compacting of free spaces
 * in favor of a regularly running compacting if reserve space
 * becomes either to tight, too fragmented or to large and space
 * should be given up.  
 */
public class ByteArrayBuffer extends ArrayBufferBase implements Cloneable {
	/**
	 * The actual data array.
	 */
	private byte data[][];
	/**
	 * Default constructor.
	 */
	public ByteArrayBuffer() {
		data=new byte[0][];
		used=0;
		reserve=0;
	}
	
	/**
	 * Check the capacity of the data array and resize it
	 * if necessary.
	 */
	@Override
	protected void checkCapacity() {
		int newLength=recalculateSize(data.length);
		
		if(newLength>0) {
//			System.out.println("IntArrayBuffer#checkCapacity(): resizing array from " + data.length + " to " + newLength);
			data=Arrays.copyOf(data, newLength);
		}			
	}
	
	/**
	 * Append a new data array to the buffer.
	 * 
	 * @param data the data array to append
	 * @return the index of the data array in the buffer
	 */
	public int append(byte[] data) {
		used+=1;
		reserve-=1;
		if(reserve<0) {
			reserve=0;
		}
		checkCapacity();
		
		this.data[used-1]=data.clone();
		
		return used-1;
	}
	
	@Override
	public void remove(int elementIndex) {
		for(int i=elementIndex;i<used-1;i++) {
			data[i]=data[i+1];
		}
		used-=1;
		checkCapacity();
	}
	
	@Override
	public void resize(int elementIndex, int newSize) {
		data[elementIndex]=Arrays.copyOf(data[elementIndex], newSize);
	}
	
	/**
	 * Get an individual element from the array with the given index.
	 * 
	 * @param elementIndex the index of the array in the buffer
	 * @param arrayIndex the index of the element within the array
	 * @return the element within the array
	 */
	public byte getArrayElement(int elementIndex, int arrayIndex) {
		return data[elementIndex][arrayIndex];
	}
	
	/**
	 * Set the value of an individual element from the array with the given index.
	 *  
	 * @param elementIndex the index of the array in the buffer
	 * @param arrayIndex the index of the element within the array
	 * @param value the new value of the element
	 */
	public void setArrayElement(int elementIndex, int arrayIndex, byte value) {
		data[elementIndex][arrayIndex]=value;
	}
	
	/**
	 * Return a copy of the array with the given element. Please note
	 * that the return value is a copy of the data in the buffer and 
	 * changes to it will not modify the array that is stored in the
	 * buffer.
	 * 
	 * @param elementIndex the index of the array within the buffer
	 * @return a copy of the array stored in the buffer
	 */
	public byte[] getArray(int elementIndex) {
		return data[elementIndex].clone();
	}
	
	/**
	 * Replace the array that is currently stored at the given element
	 * index with the supplied one.
	 * 
	 * @param elementIndex the index of the array within the buffer
	 * @param data the data that should replace the currently stored array
	 */
	public void setArray(int elementIndex, byte[] data) {
		this.data[elementIndex]=data.clone();
	}
	
	/**
	 * Returnt the length of the array with the given index.
	 * 
	 * @param elementIndex the index of the array
	 * @return number of elements in that array
	 */
	public int getArrayLength(int elementIndex) {
		return data[elementIndex].length;
	}

	/**
	 * Get the raw buffer buffer data. Use this only if you really
	 * understand how the array buffer works internally. Use of
	 * this method is not encouraged because the internals of the
	 * buffer may change in the future because of further 
	 * optimizations that might be possible on it.
	 * 
	 * @return the raw buffer data
	 */
	public byte[][] getRawData() {
		return data;
	}	

	/**
	 * Clone this buffer.
	 */
	@Override
	public ByteArrayBuffer clone() {
		ByteArrayBuffer cloned;
		try {
			cloned = (ByteArrayBuffer) super.clone();
			cloned.data=new byte[data.length][];
			for(int i=0;i<data.length;i++) {
				if(data[i]!=null) {
					cloned.data[i]=data[i].clone();
				}
			}
			return cloned;
		} catch (CloneNotSupportedException e) {
			// this cannot happen
			e.printStackTrace();
		}
		return null;
	}
}