/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Feb 14, 2007
 */
package eu.moonlight3d.framework.helper;

import eu.moonlight3d.framework.localization.L10N;


/**
 *
 * Simple and fast String helper functions. These are simplified replacements
 * of functions found in the Java class library that achieve a much higher
 * performance than their standard counterparts because they have significantly
 * less features. They are designed as special case optimisations and not as
 * general replacements.
 *
 * @author gregor
 */
public class StringTools {
	/**
	 * Split a string into an array of substrings at occurrences of a
	 * certain token character. The resulting substrings will
	 * be the sections between occurrences of the token character in
	 * order of their occurrence in the input string. The token
	 * character will not be contained in them.
	 * 
	 * @param string the string to be split
	 * @param token the token to search for
	 * @return the array of substrings resulting from the split
	 */
	public static String[] split(String string, char token) {
		String[] result;
		
		if(string==null) {
			throw new IllegalArgumentException(L10N.translate("string is null"));
		}
		
		// token can't match if string is empty, so pack string in array and return it
		if(string.length()==0) {
			result=new String[1];
			result[0]=string;
			return result;
		}
		
		int matchPositions[]=new int[string.length()];
		int numMatches=0;
		for(int i=0;i<string.length();i++) {
			if(string.charAt(i)==token) {
				matchPositions[numMatches]=i;
				numMatches++;
			}
		}
		
		result=new String[numMatches+1];
		int lastMatch=-1;
		for(int i=0;i<numMatches;i++) {
			result[i]=string.substring(lastMatch+1, matchPositions[i]);
			lastMatch=matchPositions[i];
		}
		result[numMatches]=string.substring(lastMatch+1,string.length());
		
		return result;
	}

	/**
	 * Split a string into an array of substrings at occurrences of a
	 * certain token string. The token string is taking verbatim and
	 * not interpreted in any way (e.g. as a regular expression).
	 * The resulting substrings will be the sections between occurrences 
	 * of the token string in order of their occurrence in the input 
	 * string. The token string will not be contained in them.
	 * 
	 * @param string the string to be split
	 * @param token the token to search for
	 * @return the array of substrings resulting from the split
	 */
	public static String[] split(String string, String token) {
		String[] result;

		if(string==null) {
			throw new IllegalArgumentException(L10N.translate("string is null"));
		}
		if(token==null) {
			throw new IllegalArgumentException(L10N.translate("token is null"));
		}		
		
		// token can't match if it's longer than string, so pack string in array and return it
		if(string.length()<token.length()) {
			result=new String[1];
			result[0]=string;
			return result;
		}
		
		int matchStartPositions[]=new int[string.length()/token.length()+1];
		int matchEndPositions[]=new int[string.length()/token.length()+1];
		int numMatches=0;
		for(int i=0;i<string.length()-token.length();i++) {
			boolean match=true;
			for(int j=0;j<token.length();j++) {
				if(string.charAt(i+j)!=token.charAt(j)) {
					match=false;
					break;
				}
			}
			if(match) {
				matchStartPositions[numMatches]=i;
				matchEndPositions[numMatches]=i+token.length()-1;
				numMatches++;
			}
		}
		
		result=new String[numMatches+1];
		int lastMatchEnd=-1;
		for(int i=0;i<numMatches;i++) {
			result[i]=string.substring(lastMatchEnd+1,matchStartPositions[i]);
			lastMatchEnd=matchEndPositions[i];
		}
		result[numMatches]=string.substring(lastMatchEnd+1,string.length());
		
		return result;
	}
	
	/**
	 * Join an array of strings to a single string and put the given
	 * delimiter character between individual strings. 
	 * 
	 * @param strings the strings to join
	 * @param delimiter the delimiter
	 * @return the joined string
	 */
	public static String join(String[] strings, char delimiter) {
		String result=strings[0];
		
		if(strings==null) {
			throw new IllegalArgumentException(L10N.translate("strings is null"));
		}
		
		for(int i=0;i<strings.length;i++) {
			result+=delimiter + strings[i];
		}
		
		return result;
	}

	/**
	 * Join an array of strings to a single string and put the given
	 * delimiter string between individual strings. 
	 * 
	 * @param strings the strings to join
	 * @param delimiter the delimiter
	 * @return the joined string
	 */
	public static String join(String[] strings, String delimiter) {
		String result=strings[0];
		
		if(strings==null) {
			throw new IllegalArgumentException(L10N.translate("strings is null"));
		}
		if(delimiter==null) {
			throw new IllegalArgumentException(L10N.translate("delimiter is null"));
		}

		for(int i=0;i<strings.length;i++) {
			result+=delimiter + strings[i];
		}
		
		return result;
	}
}
