/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Nov 16, 2005
 */
package eu.moonlight3d.framework.helper;

import java.util.Vector;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import eu.moonlight3d.colour.Colour;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.math.Vector3D;

/**
 * A unified, type-safe storage class for various types of values.
 *
 * @author gregor
 */
public class Value implements Cloneable {

	/**
	 * A list of all supported value types supported by {@link Value}.
	 *
	 * @author gregor
	 */
	public enum Type {
		/**
		 * No type, only useful for stubbing the type value until it is known.
		 */
		None,
		/**
		 * A single boolean.
		 */
		Boolean,
		/**
		 * A single integer.
		 */
		Integer, 
		/**
		 * A single floating point number.
		 */
		Float, 
		/**
		 * A single string.
		 */
		String, 
		/**
		 * A single 3D vector.
		 */
		Vector3D,
		/**
		 * A single colour value.
		 */
		Colour, 
		/**
		 * An array of booleans.
		 */
		BooleanArray,
		/**
		 * An array of integers.
		 */
		IntegerArray, 
		/**
		 * An array of floats.
		 */
		FloatArray, 
		/**
		 * An array of strings.
		 */
		StringArray, 
		/**
		 * An array of 3D vectors.
		 */
		Vector3DArray
	}

	/**
	 * The current value type.
	 */
	protected Type type;
	/**
	 * The current maximal size of the values array or -1 for undefined/unlimited size
	 */
	protected int maxSize;
	/**
	 * This array stores the values of this property. We have to be careful because it isn't typed
	 * and we don't want to have mixed types in this array.
	 */
	protected Vector<Object> values;
	
	/**
	 * Default constructor.
	 */
	public Value() {
		values = new Vector<Object>();		
	}

	/**
	 * Return the type of this Value object.
	 * 
	 * @return current type
	 */
	public Type getType() {
		return type;
	}

	/**
	 * Set the type of this Value. Note that the stored value is undefined after this
	 * operation.
	 * 
	 * @param type the new type for this Value object
	 */
	public void setType(Type type) {
		this.type = type;
	
		if(type==Type.BooleanArray || type==Type.IntegerArray || type==Type.FloatArray || type==Type.StringArray || type==Type.Vector3DArray) {
			values.setSize(0);
			maxSize=0;
		} else if (type != Type.None) {
			values.setSize(1);
			maxSize = 1;
		}
	}

	/**
	 * Return the number of values stored in this Value object. This is only useful
	 * for array types. For non-array types this will always return 1.
	 * 
	 * @return the size of the value array
	 */
	public int getSize() {
		return values.size();
	}

	/**
	 * Set the maximal size for the array value types. This limits the size for the array value
	 * types unless you specify 0 which indicates an unlimited size.
	 * 
	 * @param maxSize the maximally allowed size
	 * @throws IllegalStateException if trying to set the size on a non-array type
	 */
	public void setMaxSize(int maxSize) throws IllegalStateException {
		if (type == Type.Integer || type == Type.Float || type == Type.String || type == Type.Vector3D
				|| type == Type.Colour || type == Type.None) {
			throw new IllegalStateException(L10N.translate("cannot set array size for non-array value types"));
		}
	
		this.maxSize = maxSize;
		values.setSize(maxSize);
	}
	
	/**
	 * Copy the values from the given Value object.
	 * 
	 * @param value value object to copy from
	 */
	public void copyFrom(Value value) {
		type=value.type;
		maxSize=value.maxSize;
		values=(Vector<Object>) value.values.clone();		
	}
	
	@Override
	public Value clone() {
		Value value=new Value();

		value.copyFrom(this);
		
		return value;
	}
	
	/**
	 * Remove an element from an array value.
	 * 
	 * @param index the element index of the element to be removed
	 * @throws IllegalStateException if trying to remove an element from an non-array value
	 * @throws IllegalArgumentException if trying to remove an element with an invalid index
	 */
	public void remove(int index) throws IllegalStateException, IllegalArgumentException {
		if (type == Type.Integer || type == Type.Float || type == Type.String || type == Type.Vector3D
				|| type == Type.Colour || type == Type.None) {
			throw new IllegalStateException(L10N.translate("cannot remove array element on non-array value types"));
		}
		
		if(index >= values.size() || index < 0) {
			throw new IllegalArgumentException(L10N.translate("invalid array index"));
		}
		
		values.remove(index);
	}

	/**
	 * Get the string value stored in the value object if it is
	 * is set to type String.
	 * 
	 * @return the string value stored in the Value object
	 * @throws IllegalStateException
	 */
	public String getStringValue() throws IllegalStateException {
		if (type == Type.String) {
			return (String) values.get(0);
		} else {
			throw new IllegalStateException(L10N.translate("cannot return string from non-string Value"));
		}
	}

	/**
	 * Get the 3d vector value stored in the value object if it is
	 * is set to type Vector3D.
	 * 
	 * @return the 3d vector value stored in the Value object
	 * @throws IllegalStateException
	 */
	public Vector3D getVectorValue() throws IllegalStateException {
		if (type == Type.Vector3D) {
			return ((Vector3D) values.get(0)).clone();
		} else {
			throw new IllegalStateException(L10N.translate("cannot return vector from non-vector Value"));
		}
	}

	/**
	 * Get the colour value stored in the value object if it is
	 * is set to type Colour.
	 * 
	 * @return the colour value stored in the Value object
	 * @throws IllegalStateException
	 */
	public Colour getColourValue() throws IllegalStateException {
		if (type == Type.Colour) {
			return (Colour) values.get(0);
		} else {
			throw new IllegalStateException(L10N.translate("cannot return colour from non-colour Value"));
		}
	}

	/**
	 * Get the boolean value stored in the value object if it is
	 * is set to type Boolean.
	 * 
	 * @return the boolean value stored in the Value object
	 * @throws IllegalStateException
	 */
	public boolean getBooleanValue() throws IllegalStateException {
		if (type == Type.Boolean) {
			return (Boolean) values.get(0);
		} else {
			throw new IllegalStateException(L10N.translate("cannot return boolean from non-boolean Value"));
		}
	}

	/**
	 * Get the integer value stored in the value object if it is
	 * is set to type Integer.
	 * 
	 * @return the integer value stored in the Value object
	 * @throws IllegalStateException
	 */
	public int getIntegerValue() throws IllegalStateException {
		if (type == Type.Integer) {
			return (Integer) values.get(0);
		} else {
			throw new IllegalStateException(L10N.translate("cannot return integer from non-integer Value"));
		}
	}

	/**
	 * Get the float value stored in the value object if it is
	 * is set to type float.
	 * 
	 * @return the float value stored in the Value object
	 * @throws IllegalStateException
	 */
	public double getFloatValue() throws IllegalStateException {
		if (type == Type.Float) {
			return (Double) values.get(0);
		} else {
			throw new IllegalStateException(L10N.translate("cannot return double from non-double Value"));
		}
	}

	public void setValue(String stringValue) throws IllegalStateException {
		if (type == Type.String) {
			values.set(0, stringValue);
		} else {
			throw new IllegalStateException(L10N.translate("cannot set string value on non-string Value"));
		}
	}

	public void setValue(Vector3D vectorValue) throws IllegalStateException {
		if (type == Type.Vector3D) {
			values.set(0, vectorValue.clone());
		} else {
			throw new IllegalStateException(L10N.translate("cannot set vector value on non-vector Value"));
		}
	}

	public void setValue(Colour colourValue) throws IllegalStateException {
		if (type == Type.Colour) {
			values.set(0, colourValue.clone());
		} else {
			throw new IllegalStateException(L10N.translate("cannot set colour value on non-colour Value"));
		}
	}

	public void setValue(boolean booleanValue) throws IllegalStateException {
		if (type == Type.Boolean) {
			values.set(0, booleanValue);
		} else {
			throw new IllegalStateException(L10N.translate("cannot set boolean value on non-boolean Value"));
		}
	}

	public void setValue(int integerValue) throws IllegalStateException {
		if (type == Type.Integer) {
			values.set(0, integerValue);
		} else {
			throw new IllegalStateException(L10N.translate("cannot set integer value on non-integer Value"));
		}
	}

	public void setValue(double doubleValue) throws IllegalStateException {
		if (type == Type.Float) {
			values.set(0, doubleValue);
		} else {
			throw new IllegalStateException(L10N.translate("cannot set double value on non-double Value"));
		}
	}

	public Vector3D getArrayVectorValue(int index) throws IllegalStateException {
		if (type == Type.Vector3DArray && index < values.size()) {
			return (Vector3D) values.get(index);
		} else {
			throw new IllegalStateException(L10N.translate("cannot get vector value from non-vector Value"));
		}
	}

	public String getArrayStringValue(int index) throws IllegalStateException {
		if (type == Type.StringArray && index < values.size()) {
			return (String) values.get(index);
		} else {
			throw new IllegalStateException(L10N.translate("cannot get string value from non-string Value"));
		}
	}

	public boolean getArrayBooleanValue(int index) throws IllegalStateException {
		if (type == Type.BooleanArray && index < values.size()) {
			return (Boolean) values.get(index);
		} else {
			throw new IllegalStateException(L10N.translate("cannot get boolean value from non-boolean Value"));
		}
	}

	public int getArrayIntegerValue(int index) throws IllegalStateException {
		if (type == Type.IntegerArray && index < values.size()) {
			return (Integer) values.get(index);
		} else {
			throw new IllegalStateException(L10N.translate("cannot get integer value from non-integer Value"));
		}
	}

	public double getArrayFloatValue(int index) throws IllegalStateException {
		if (type == Type.FloatArray && index < values.size()) {
			return (Double) values.get(index);
		} else {
			throw new IllegalStateException(L10N.translate("cannot get double value from non-double Value"));
		}
	}

	public void setArrayValue(int index, String stringValue) throws IllegalStateException {
		if (type == Type.StringArray && (index < maxSize || maxSize == 0)) {
			if (maxSize == 0 && index >= values.size()) {
				values.setSize(index + 1);
			}
			values.set(index, stringValue);
		} else {
			throw new IllegalStateException(L10N.translate("cannot set string value on non-string Value"));
		}
	}

	public void setArrayValue(int index, Vector3D vectorValue) throws IllegalStateException {
		if (type == Type.Vector3DArray && (index < maxSize || maxSize == 0)) {
			if (maxSize == 0 && index >= values.size()) {
				values.setSize(index + 1);
			}
			values.set(index, vectorValue);
		} else {
			throw new IllegalStateException(L10N.translate("cannot set vector value on non-vector Value"));
		}
	}

	public void setArrayValue(int index, boolean booleanValue) throws IllegalStateException {
		if (type == Type.BooleanArray && ((index < maxSize) || (maxSize == 0))) {
			if (maxSize == 0 && index >= values.size()) {
				values.setSize(index + 1);
			}
			values.set(index, booleanValue);
		} else {
			throw new IllegalStateException(L10N.translate("cannot set boolean value on non-boolean Value"));
		}
	}

	public void setArrayValue(int index, int integerValue) throws IllegalStateException {
		if (type == Type.IntegerArray && ((index < maxSize) || (maxSize == 0))) {
			if (maxSize == 0 && index >= values.size()) {
				values.setSize(index + 1);
			}
			values.set(index, integerValue);
		} else {
			throw new IllegalStateException(L10N.translate("cannot set integer value on non-integer Value"));
		}
	}

	public void setArrayValue(int index, double doubleValue) throws IllegalStateException {
		if (type == Type.FloatArray && (index < maxSize || maxSize == 0)) {
			if (maxSize == 0 && index >= values.size()) {
				values.setSize(index + 1);
			}
			values.set(index, doubleValue);
		} else {
			throw new IllegalStateException(L10N.translate("cannot set double value on non-double Value"));
		}
	}

	public void setValue(Value value) {
		type = value.type;
		maxSize = value.maxSize;
		values = new Vector();
		if (type == Type.Vector3D || type == Type.Vector3DArray || type == Type.Colour) {
			for (int i = 0; i < value.values.size(); i++) {
				if (type == Type.Vector3D || type == Type.Vector3DArray) {
					values.add(((Vector3D) value.values.get(i)).clone());
				} else if (type == Type.Colour) {
					values.add(((Colour) value.values.get(i)).clone());
				}
			}
		} else {
			// a shallow copy is enough here as the references inside the arrays
			// are replaced when values change for these types
			values = (Vector) value.values.clone();
		}
	}

	public void readFromFile(Element propertyElement, boolean preserveType) throws Exception {
        if(!propertyElement.getNodeName().equals("value")) {
            throw new Exception(L10N.translate("invalid file format: unknown tag found when expecting a <value> tag"));
        }
//        eu.moonlight3d.framework.State.getInstance().logger().info("property name: \"" + name + "\"");
        String valueType=propertyElement.getAttributes().getNamedItem("type").getNodeValue();
        
        if(valueType.equals("none")) {
            // do nothing
        } else if(valueType.equals("boolean")) {
        	if(!preserveType) {
        		setType(Value.Type.Boolean);
        	}
            NodeList valueList=propertyElement.getElementsByTagName("value");
            if(valueList.getLength()!=1) {
                throw new Exception(L10N.translate("invalid file format: expects exactly one <value> tag inside <property> tag of type boolean"));
            }
            Element valueElement=(Element)valueList.item(0);
            String value=valueElement.getAttribute("booleanvalue");
            setValue(Boolean.parseBoolean(value));
        } else if(valueType.equals("integer")) {
        	if(!preserveType) {
        		setType(Value.Type.Integer);
        	}
            NodeList valueList=propertyElement.getElementsByTagName("value");
            if(valueList.getLength()!=1) {
                throw new Exception(L10N.translate("invalid file format: expects exactly one <value> tag inside <property> tag of type integer"));
            }
            Element valueElement=(Element)valueList.item(0);
            String value=valueElement.getAttribute("integervalue");
            setValue(Integer.parseInt(value));
        } else if(valueType.equals("float")) {
        	if(!preserveType) {
        		setType(Value.Type.Float);
        	}
            NodeList valueList=propertyElement.getElementsByTagName("value");
            if(valueList.getLength()!=1) {
                throw new Exception(L10N.translate("invalid file format: expects exactly one <value> tag inside <property> tag of type float"));
            }
            Element valueElement=(Element)valueList.item(0);
            String value=valueElement.getAttribute("floatvalue");
            setValue(Double.parseDouble(value));
        } else if(valueType.equals("string")) {
        	if(!preserveType) {
        		setType(Value.Type.String);
        	}
            NodeList valueList=propertyElement.getElementsByTagName("value");
            if(valueList.getLength()!=1) {
                throw new Exception(L10N.translate("invalid file format: expects exactly one <value> tag inside <property> tag of type string"));
            }
            Element valueElement=(Element)valueList.item(0);
            String value=valueElement.getAttribute("stringvalue");
            setValue(value);
        } else if(valueType.equals("vector3d")) {
        	if(!preserveType) {
        		setType(Value.Type.Vector3D);
        	}
            NodeList valueList=propertyElement.getElementsByTagName("value");
            if(valueList.getLength()!=1) {
                throw new Exception(L10N.translate("invalid file format: expects exactly one <value> tag inside <property> tag of type vector3d"));
            }
            Element valueElement=(Element)valueList.item(0);
            String xvalue=valueElement.getAttribute("xvalue");
            String yvalue=valueElement.getAttribute("yvalue");
            String zvalue=valueElement.getAttribute("zvalue");
            eu.moonlight3d.math.Vector3D value=new eu.moonlight3d.math.Vector3D();
            value.X1=Double.parseDouble(xvalue);
            value.X2=Double.parseDouble(yvalue);
            value.X3=Double.parseDouble(zvalue);
            setValue(value);
        } else if(valueType.equals("colour")) {
        	if(!preserveType) {
        		setType(Value.Type.Colour);
        	}
            NodeList valueList=propertyElement.getElementsByTagName("value");
            if(valueList.getLength()!=1) {
                throw new Exception(L10N.translate("invalid file format: expects exactly one <value> tag inside <property> tag of type colour"));
            }
            Element valueElement=(Element)valueList.item(0);
            String red=valueElement.getAttribute("red");
            String green=valueElement.getAttribute("green");
            String blue=valueElement.getAttribute("blue");
            String alpha=valueElement.getAttribute("alpha");
            Colour value=new Colour();
            value.red=Double.parseDouble(red);
            value.green=Double.parseDouble(green);
            value.blue=Double.parseDouble(blue);
            value.alpha=Double.parseDouble(alpha);
            setValue(value);
        } else if(valueType.equals("booleanarray")) {
        	if(!preserveType) {
        		setType(Value.Type.BooleanArray);
        	}
            int length=Integer.parseInt(propertyElement.getAttribute("length"));
            NodeList valueList=propertyElement.getElementsByTagName("value");
            if(valueList.getLength()!=length) {
                throw new Exception(L10N.translate("invalid file format: number of <value> tags inside <property> doesn't match given length of booleanarray property"));
            }
            for(int k=0;k<length;k++) {
                Element valueElement=(Element)valueList.item(k);
                String value=valueElement.getAttribute("booleanvalue");
                setArrayValue(k,Boolean.parseBoolean(value));
            }
        } else if(valueType.equals("integerarray")) {
        	if(!preserveType) {
        		setType(Value.Type.IntegerArray);
        	}
            int length=Integer.parseInt(propertyElement.getAttribute("length"));
            NodeList valueList=propertyElement.getElementsByTagName("value");
            if(valueList.getLength()!=length) {
                throw new Exception(L10N.translate("invalid file format: number of <value> tags inside <property> doesn't match given length of integerarray property"));
            }
            for(int k=0;k<length;k++) {
                Element valueElement=(Element)valueList.item(k);
                String value=valueElement.getAttribute("integervalue");
                setArrayValue(k,Integer.parseInt(value));
            }
        } else if(valueType.equals("floatarray")) {
        	if(!preserveType) {
        		setType(Value.Type.FloatArray);
        	}
            int length=Integer.parseInt(propertyElement.getAttribute("length"));
            NodeList valueList=propertyElement.getElementsByTagName("value");
            if(valueList.getLength()!=length) {
                throw new Exception(L10N.translate("invalid file format: number of <value> tags inside <property> doesn't match given length of floatrarray property"));
            }
            for(int k=0;k<length;k++) {
                Element valueElement=(Element)valueList.item(k);
                String value=valueElement.getAttribute("floatvalue");
                setArrayValue(k,new Double(value));
            }
        } else if(valueType.equals("stringarray")) {
        	if(!preserveType) {
        		setType(Value.Type.StringArray);
        	}
            int length=Integer.parseInt(propertyElement.getAttribute("length"));
            NodeList valueList=propertyElement.getElementsByTagName("value");
            if(valueList.getLength()!=length) {
                throw new Exception(L10N.translate("invalid file format: number of <value> tags inside <property> doesn't match given length of stringarray property"));
            }
            for(int k=0;k<length;k++) {
                Element valueElement=(Element)valueList.item(k);
                String value=valueElement.getAttribute("stringvalue");
                setArrayValue(k,value);
            }
        } else if(valueType.equals("vector3darray")) {
        	if(!preserveType) {
        		setType(Value.Type.Vector3DArray);
        	}
            int length=Integer.parseInt(propertyElement.getAttribute("length"));
            NodeList valueList=propertyElement.getElementsByTagName("value");
            if(valueList.getLength()!=length) {
                throw new Exception(L10N.translate("invalid file format: number of <value> tags inside <property> doesn't match given length of vector3darray property"));
            }
            for(int k=0;k<length;k++) {
                Element valueElement=(Element)valueList.item(k);
                String xvalue=valueElement.getAttribute("xvalue");
                String yvalue=valueElement.getAttribute("yvalue");
                String zvalue=valueElement.getAttribute("zvalue");
                eu.moonlight3d.math.Vector3D value=new eu.moonlight3d.math.Vector3D();
                value.X1=Double.parseDouble(xvalue);
                value.X2=Double.parseDouble(yvalue);
                value.X3=Double.parseDouble(zvalue);
                setArrayValue(k,value);
            }                    
        } else {
            throw new Exception(L10N.translate("invalid file format: unknown property type encountered"));
        }    	
	}

	public void writeToFile(Document doc, Element valueContainer) throws DOMException, Exception {
        Element valueElement=doc.createElement("value");
        if(getType()==Property.Type.None) {
            valueElement.setAttribute("type","none");
        } else if(getType()==Property.Type.Boolean) {
            valueElement.setAttribute("type","boolean");
            Element value=doc.createElement("value");
            value.setAttribute("booleanvalue",Boolean.toString(getBooleanValue()));
            valueElement.appendChild(value);
        } else if(getType()==Property.Type.Integer) {
            valueElement.setAttribute("type","integer");
            Element value=doc.createElement("value");
            value.setAttribute("integervalue",Integer.toString(getIntegerValue()));
            valueElement.appendChild(value);
        } else if(getType()==Property.Type.Float) {
            valueElement.setAttribute("type","float");
            Element value=doc.createElement("value");
            value.setAttribute("floatvalue",Double.toString(getFloatValue()));
            valueElement.appendChild(value);
        } else if(getType()==Property.Type.String) {
            valueElement.setAttribute("type","string");
            Element value=doc.createElement("value");
            value.setAttribute("stringvalue",getStringValue());
            valueElement.appendChild(value);
        } else if(getType()==Property.Type.Vector3D) {
            valueElement.setAttribute("type","vector3d");
            Element value=doc.createElement("value");
            value.setAttribute("xvalue",Double.toString(getVectorValue().X1));
            value.setAttribute("yvalue",Double.toString(getVectorValue().X2));
            value.setAttribute("zvalue",Double.toString(getVectorValue().X3));
            valueElement.appendChild(value);
        } else if(getType()==Property.Type.Colour) {
            valueElement.setAttribute("type","colour");
            Element value=doc.createElement("value");
            value.setAttribute("red",Double.toString(getColourValue().red));
            value.setAttribute("green",Double.toString(getColourValue().green));
            value.setAttribute("blue",Double.toString(getColourValue().blue));
            value.setAttribute("alpha",Double.toString(getColourValue().alpha));
            valueElement.appendChild(value);
        } else if(getType()==Property.Type.BooleanArray) {
            valueElement.setAttribute("type","booleanarray");
            valueElement.setAttribute("length",Integer.toString(getSize()));
            for(int k=0;k<getSize();k++) {
                Element value=doc.createElement("value");
                value.setAttribute("booleanvalue",Boolean.toString(getArrayBooleanValue(k)));
                valueElement.appendChild(value);
            }
        } else if(getType()==Property.Type.IntegerArray) {
            valueElement.setAttribute("type","integerarray");
            valueElement.setAttribute("length",Integer.toString(getSize()));
            for(int k=0;k<getSize();k++) {
                Element value=doc.createElement("value");
                value.setAttribute("integervalue",Integer.toString(getArrayIntegerValue(k)));
                valueElement.appendChild(value);
            }
        } else if(getType()==Property.Type.FloatArray) {
            valueElement.setAttribute("type","floatarray");
            valueElement.setAttribute("length",Integer.toString(getSize()));
            for(int k=0;k<getSize();k++) {
                Element value=doc.createElement("value");
                value.setAttribute("floatvalue",Double.toString(getArrayFloatValue(k)));
                valueElement.appendChild(value);
            }
        } else if(getType()==Property.Type.StringArray) {
            valueElement.setAttribute("type","stringarray");
            valueElement.setAttribute("length",Integer.toString(getSize()));
            for(int k=0;k<getSize();k++) {
                Element value=doc.createElement("value");
                value.setAttribute("stringvalue",getArrayStringValue(k));
                valueElement.appendChild(value);
            }
        } else if(getType()==Property.Type.Vector3DArray) {
            valueElement.setAttribute("type","vector3darray");
            valueElement.setAttribute("length",Integer.toString(getSize()));
            for(int k=0;k<getSize();k++) {
                Element value=doc.createElement("value");
                value.setAttribute("xvalue",Double.toString(getArrayVectorValue(k).X1));
                value.setAttribute("yvalue",Double.toString(getArrayVectorValue(k).X2));
                value.setAttribute("zvalue",Double.toString(getArrayVectorValue(k).X3));
                valueElement.appendChild(value);
            }
        } else {
            throw new Exception(L10N.translate("encountered unknown property type while saving"));
        }
        
        valueContainer.appendChild(valueElement);
	}
}
