/*
 * Moonlight|3D
 * Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Nov 2, 2007
 */

package eu.moonlight3d.framework.helper;

import java.util.Arrays;

import eu.moonlight3d.math.Vector3D;

/**
 * A plain integer array encapsulation designed to reduce the
 * number of array reallocations during the lifetime of the 
 * array.
 * 
 * @author gregor
 */ 
public class Vector3DBuffer extends BufferBase implements Cloneable {
	/**
	 * The actual data array.
	 */
	private double data[];
	
	/**
	 * Default constructor.
	 */
	public Vector3DBuffer() {
		data=new double[30];
	}

	@Override
	protected void checkCapacity() {
		int newLength=recalculateSize(data.length);
		
		if(newLength>0) {
			data=Arrays.copyOf(data, newLength);
		}
	}
	
	/**
	 * Apend a value to the array.
	 * 
	 * @param data the value to append
	 * @return the index of the value
	 */
	public int append(Vector3D data) {
		used+=3;
		reserve-=3;
		if(reserve<0) {
			reserve=0;
		}
		// ensure capacity
		checkCapacity();
		
		this.data[used-3]=data.X1;
		this.data[used-2]=data.X2;
		this.data[used-1]=data.X3;
		return used/3;
	}
	
	@Override
	public void remove(int elementIndex) {
		for(int i=3*elementIndex;i<used-3;i++) {
			data[i]=data[i+3];
		}
		used-=3;
		// check if array could be shrinked
		checkCapacity();		
	}
	
	@Override
	public void setReserve(int reserve) {
		this.reserve=reserve*3;
		// ensure capacity
		checkCapacity();
	}
	
	/**
	 * Get the value of an array element.
	 * 
	 * @param elementIndex the index of the element to return
	 * @return the value of the element
	 */
	public Vector3D getVector(int elementIndex) {
		return new Vector3D(data[elementIndex*3],data[elementIndex*3+1],data[elementIndex*3+2]);
	}
	
	/**
	 * Set the value of an array element.
	 * 
	 * @param elementIndex the index of the element to set
	 * @param newData the new value of the element
	 */
	public void setVector(int elementIndex, Vector3D newData) {
		data[elementIndex*3]=newData.X1;
		data[elementIndex*3+1]=newData.X2;
		data[elementIndex*3+2]=newData.X3;
	}
	
	@Override
	public int getLength() {
		return used/3;
	}
	
	/**
	 * Get the raw buffer buffer data. Use this only if you really
	 * understand how the buffer works internally. Use of
	 * this method is not encouraged because the internals of the
	 * buffer may change in the future because of further 
	 * optimizations that might be possible on it.
	 * 
	 * @return the raw buffer data
	 */
	public double[] getRawData() {
		return data;
	}

	/**
	 * Clone this buffer.
	 */
	@Override
	public Vector3DBuffer clone() {
		Vector3DBuffer cloned;
		try {
			cloned = (Vector3DBuffer) super.clone();
			cloned.data=data.clone();
			return cloned;
		} catch (CloneNotSupportedException e) {
			// this cannot happen
			e.printStackTrace();
		}
		return null;
	}
}