/*
 * Moonlight|3D
 * Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Oct 4, 2007
 */

package eu.moonlight3d.framework.localization;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Locale;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import eu.moonlight3d.framework.State;

/**
 * Localization resource container and message translator. This class
 * holds is responsible for loading all known translations for the
 * default locale from the corresponding files and contains the 
 * necessary functions to translate messages using these
 * translation tables. Translations are stored with a key consisting
 * of a context and the untranslated message. Usually the context
 * is the name of the calling class. In other cases this may be
 * inappropriate, for example for strings loaded from files which
 * define user interface contributions like property sheets. In
 * this case it is recommended to use the name of the data file
 * as context. 
 *
 * @author gregor
 */
public class L10N {
	/**
	 * Container of all known localized messages
	 */
	private static HashMap<String, HashMap<String, String>> translations;

	/**
	 * Load translations from a given file and add them to the current pool of
	 * translations. This function fails silently if the file cannot be read so
	 * that the loading process stays as transparent as possible.
	 * 
	 * @param file the file to load translations from
	 */
	private static void loadTranslationFile(File file) {
		// parse Qt's ts file format here
		Document translationDoc;
		int numTranslations=0;
		
		try {
			translationDoc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);

			Element root = translationDoc.getDocumentElement();
			
			if(!"ts".equals(root.getNodeName().toLowerCase())) {
				return;
			}
			
			NodeList contexts = root.getElementsByTagName("context");
			
			for(int i=0;i<contexts.getLength();i++) {
				Element context = (Element) contexts.item(i);
				Element name = (Element) context.getElementsByTagName("name").item(0);
				String contextName=name.getChildNodes().item(0).getNodeValue();
				
				if(translations==null) {
					translations=new HashMap<String, HashMap<String,String>>();
				}
				
				HashMap<String, String> contextMap=translations.get(contextName);
				if(contextMap==null) {
					contextMap=new HashMap<String, String>();
					translations.put(contextName,contextMap);
				}
				
				NodeList messages = context.getElementsByTagName("message");
				for(int j=0;j<messages.getLength();j++) {
					Element message = (Element) messages.item(j);
					Element source = (Element) message.getElementsByTagName("source").item(0);
					Element translation = (Element) message.getElementsByTagName("translation").item(0);
					String sourceString=source.getChildNodes().item(0).getNodeValue();
					String translationString=null;
					if(translation.getChildNodes().getLength()!=0) {
						translationString=translation.getChildNodes().item(0).getNodeValue();
					}
					System.out.println(sourceString + "->" + translationString);
					if(translationString!=null) {
						contextMap.put(sourceString, translationString);
					} else {
						contextMap.put(sourceString, sourceString);						
					}
				}
			}
			
		} catch (SAXException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ParserConfigurationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		State.getInstance().logger().fine("Loaded " + numTranslations + " from " + file.getAbsolutePath());

	}
	
	/**
	 * Load translations from the specified directory.
	 * 
	 * @param directory the directory to scan for translation files
	 */
	private static void loadTranslations(File directory) {
		File[] fileList=directory.listFiles();
		
		for(File file : fileList) {
			if(file.isFile()) {
				loadTranslationFile(file);
			}
		}
	}
	
	/**
	 * Load translations for the current user locale. This funcion is invoked
	 * automatically during framework startup.
	 */
	public static void loadTranslations() {
		Locale locale=Locale.getDefault();
		
		State state=State.getInstance();
		
		String translationsBasePath=state.getInstallationDirectory() + System.getProperty("file.separator") + "translations" + System.getProperty("file.separator");
		
		File languageDirectory=new File(translationsBasePath + locale.getLanguage());
		
		if(languageDirectory.exists()) {
			loadTranslations(languageDirectory);
		}
		
		File variantDirectory=new File(translationsBasePath + locale.getLanguage() + "_" + locale.getVariant());
		
		if(variantDirectory.exists()) {
			loadTranslations(variantDirectory);
		}
	}
	
	/**
	 * Translate the given string into the current locale within
	 * the specified context. If no translation can be found in the
	 * specified context, the unmodified source message is returned.
	 * 
	 * @param context the context of the message
	 * @param message the message to translate
	 * @return the translation of the message
	 */
	public static String translate(String context, String message) {
		if(translations==null) {
			return message;
		}
		
		HashMap<String, String> contextTable=translations.get(context);
		
		if(contextTable==null) {
			return message;
		}
		
		String result=contextTable.get(message);
		
		if(result==null) {
			return message;
		}
		
		return result;
	}
	
	/**
	 * Translate the given string into the current locale within the
	 * context of the calling function. If no translation can be found,
	 * the unmodified source message is returned.
	 * 
	 * @param message the message to translate
	 * @return the translation of the message
	 */
	public static String translate(String message) {
		Throwable stackProvider=new Throwable();
		StackTraceElement[] stack=stackProvider.getStackTrace();
		
		if(stack.length<2) {
			return message;
		}
		
		String context=stack[1].getClassName();
		
		return translate(context, message);
	}
}
