/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 5, 2005
 */
package eu.moonlight3d.framework.plugins;
/**
 * \package eu.moonlight3d.framework.plugins
 * 
 * This is the plugin manager. It is responsible for loading and
 * unloading all plugins at runtime.
 * 
 * This package is part of the core program.
 */

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.exceptions.ExceptionHelper;
import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.helper.Value;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.preferences.Preferences;
import eu.moonlight3d.framework.preferences.PreferencesGroup;

/**
 * The plugin manager is responsible for loading and unloading plugins 
 * in the right order.
 * 
 * @author gregor
 */
public class Manager {
	/**
	 * Special Classloader that allows loading of plugins. This classloader
	 * is required so that the plugin classpath can be added to while the
	 * plugin definition files are loaded.
	 *
	 * @author gregor
	 */
	class ClassLoader extends URLClassLoader {
		/**
		 * Default constructor
		 */
		public ClassLoader() {
			super(new URL[0]);
		}
		
		@Override
		public void addURL(URL url) {
			super.addURL(url);
		}
	}
	
	/**
	 * Plugin wrapper that maintains plugin state and other internal information.
	 * This class acts as a wrapper around a plugin's Plugin instance so that
	 * implementation details of the plugin loader need not be exposed to this
	 * interface.
	 *
	 * @author gregor
	 */
    class PluginWrapper {
    	/**
    	 * Boolean indicating whether the plugin has already been loaded and 
    	 * initialised.
    	 */
        private boolean loaded;
        /**
         * Boolean indicating if loading and initialising the plugin has been
         * attempted already, but failed. If true, no further loading attempts
         * will be made. 
         */
        private boolean unloadable;
        /**
         * The type of the plugin's implementation of the Plugin interface.
         * This is used to instantiate the plugin class.
         */
        private Class<Plugin> pluginClass;
        /**
         * The instance of the plugin's Plugin implementation.
         */
        private Plugin plugin;
        /**
         * Reference back to the plugin manager for convenience.
         */
        private Manager manager;
        
        /**
         * Default constructor taking reference to manager and plugin class
         * type.
         * 
         * @param manager reference to manager
         * @param pluginClass plugin class type
         */
        public PluginWrapper(Manager manager, Class<Plugin> pluginClass) {
            this.manager=manager;
            this.pluginClass=pluginClass;
            unloadable=false;
        }
        
        /**
         * Instantiate the Plugin interface implementation for this plugin.
         * 
         * @throws InstantiationException
         * @throws IllegalAccessException
         */
        public void instantiate() throws InstantiationException, IllegalAccessException {
            plugin=(Plugin)pluginClass.newInstance();        	
        }
        
        /**
         * Checks, whether the dependencies of the plugin are actually satisfied
         * @return true, if dependencies are all satisfied, false otherwise
         */
        public boolean hasSatisfiesDependencies()
        {
            ArrayList<String> dependencies=plugin.getDependencies();
            
            if(dependencies==null) {
            		return true;
            }
            
            for(int i=0;i<dependencies.size();i++) {
                boolean foundDependency=false;
                for(int j=0;j<manager.plugins.size();j++) {
                    if(dependencies.get(i)==manager.plugins.get(j).plugin.getName()) {
                        foundDependency=true;
                    }
                }
                if(!foundDependency) {
                    return false;
                }
            }
            
            return true;
        }
        
        /**
         * Checks, if the plugin can be loaded right now
         * @return true, if plugin can be loaded and initialized right now, false if there are unsatisfied 
         * dependencies or dependencies aren't loaded
         */
        public boolean isLoadable()
        {
            ArrayList<String> dependencies=plugin.getDependencies();
            
            if(dependencies==null) {
            		return true;
            }
            
            if(unloadable) {
            		return false;
            }
            
            for(int i=0;i<dependencies.size();i++) {
                boolean foundDependency=false;
                for(int j=0;j<manager.plugins.size();j++) {
                    if(dependencies.get(i)==manager.plugins.get(j).plugin.getName()) {
                        if(manager.plugins.get(j).isLoaded()) {
                            foundDependency=true;
                        } else if(manager.plugins.get(j).isUnloadable()) {
                        		markUnloadable();
                        		return false;
                        }
                    }
                }
                if(!foundDependency) {
                    return false;
                }
            }
            
            return true;
        }
        
        /**
         * Indicates whether the plugin has been marked as unloadable.
         * 
         * @return true if the plugin is unloadable, false otherwise
         */
        public boolean isUnloadable() {
        		return unloadable;
        }
        
        /**
         * Mark this plugin as unloadable. When the plugin is marked, no 
         * further attempts will be made to load it.
         */
        public void markUnloadable() {
        		unloadable=true;
        }
        
        /**
         * Checks, whether there are other plugins running which are depending on this one.
         * @return true, if there is at least one other plugin running which depends on this one
         */
        public boolean hasRuntimeDependencies()
        {
            for(int i=0;i<manager.plugins.size();i++) {
                if(manager.plugins.get(i).isLoaded()) {
                    ArrayList<String> dependencies=manager.plugins.get(i).plugin.getDependencies();
                    if(dependencies==null) {
                    		continue;
                    }
                    for(int j=0;j<dependencies.size();j++) {
                        if(dependencies.get(j)==plugin.getName()) {
                            return true;
                        }
                    }
                }
            }
            
            return false;
        }

        /**
         * Indicates whether the plugin has been loaded completely and successfully.
         * 
         * @return true if the plugin was loaded completely, false otherwise
         */
        public boolean isLoaded()
        {
            return loaded;
        }
        
        /**
         * Attempt to initialise the plugin. Depending on the nature of the
         * plugin, anything may happen during plugin initialisation.
         * 
         * @throws java.lang.Throwable
         */
        public void load() throws java.lang.Throwable
        {
            plugin.load();
            loaded=true;
        }
        
        /**
         * Attempt to uninitialise the plugin. Depending on the nature of the
         * plugin, anything may happen during plugin shutdown.
         * @throws java.lang.Throwable
         */
        public void unload() throws java.lang.Throwable
        {
            plugin.unload();
            loaded=false;
        }
    }
    
    /**
     * The list of plugins found in the pluginlist files.
     */
    private ArrayList<PluginWrapper> plugins;
    /**
     * The plugin class loader.
     */
    private ClassLoader classLoader;
    
    /**
     * Default constructor.
     */
    public Manager()
    {  
        plugins=new ArrayList<PluginWrapper>();
        classLoader=new ClassLoader();
        
        // build a preferences group with the additional plugin paths
        PreferencesGroup pluginPreferences=new PreferencesGroup();
        pluginPreferences.setName("Plugins");
        Property paths=new Property(pluginPreferences);
        paths.setName("paths");
        paths.setType(Value.Type.StringArray);
        try {
			paths.setMaxSize(0);
		} catch (Exception e) {
			// this cannot happen here
			e.printStackTrace();
		}
        
        State.getInstance().getPreferencesManager().getCurrentPreferences().addPreferencesGroup(pluginPreferences);
    }
    
    /**
     * tries to load a plugin from a given class name
     * @param className the name of the class to load as a plugin
     * @throws Exception 
     */
    void getPlugin(String className) throws Exception
    {
        try {
            Class<Plugin> pluginClass=(Class<Plugin>) classLoader.loadClass(className);
            plugins.add(new PluginWrapper(this,pluginClass));
        } catch (ClassNotFoundException e) {
        	String message=MessageFormat.format(L10N.translate("failed to load plugin {0}"), className);
            throw new Exception(message,e);
        }
    }
    
    /**
     * Add a path to the plugin class path.
     * 
     * @param path the path to add
     */
    private void addToClasspath(String path) {
    	String message=L10N.translate("adding to classpath:") + " " + path;
   		State.getInstance().logger().info(message);
    	try {
			classLoader.addURL(new URL("file:" + path));
		} catch (MalformedURLException e) {
			// TODO handle exception more appropriately
			e.printStackTrace();
		}
	}
    
    /**
     * Parse a pluginlist file. This parses a pluginlist XML file, adds all 
     * plugins that are defined within it to the plugins list and all
     * classpath elements found in that file to the plugin classloader.
     * 
     * @param file the file to load
     * @throws Exception
     */
    void parsePluginList(File file) throws Exception
    {
        try {
            Document pluginListDoc=DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file.toURI().toString());
            Element root=pluginListDoc.getDocumentElement();
            NodeList classpathList, pluginList;
            
            State.getInstance().logger().info(L10N.translate("loading plugins from ") + file.toURI().toString());
            
            classpathList=root.getElementsByTagName("classpath");
            for(int i=0;i<classpathList.getLength();i++) {
            		Element classpath=(Element) classpathList.item(i);
            		String path=classpath.getAttributes().getNamedItem("location").getNodeValue();
            		addToClasspath(file.getParentFile().getCanonicalPath() + System.getProperty("file.separator") + path);
            }
            pluginList=root.getElementsByTagName("plugin");
            
            for(int i=0;i<pluginList.getLength();i++) {
                Node node=pluginList.item(i);
                
                Node attributeNode=node.getAttributes().getNamedItem("className");
                
                if(attributeNode==null) {
                    // log this
                    eu.moonlight3d.framework.State.getInstance().logger().info(L10N.translate("no className attribute in plugin"));
                    continue;
                }
                
                String pluginClassName=attributeNode.getNodeValue();
                try {
                    getPlugin(pluginClassName);
                } catch (Exception e) {
                    // log plugin load failure
                	String message=MessageFormat.format(L10N.translate("unable to load plugin {0}, reason:"), pluginClassName);
                    eu.moonlight3d.framework.State.getInstance().logger().info(message);
                    eu.moonlight3d.framework.State.getInstance().logger().info(ExceptionHelper.getFullText(e));
                }
            }
        } catch (ParserConfigurationException e) {
        	String message=MessageFormat.format(L10N.translate("unable to parse plugin list XML file {0}"), file.getAbsolutePath());
            throw new Exception(message,e);
        } catch (SAXException e) {
        	String message=MessageFormat.format(L10N.translate("unable to parse plugin list XML file {0}"), file.getAbsolutePath());
            throw new Exception(message,e);
        } catch (IOException e) {
        	String message=MessageFormat.format(L10N.translate("unable to parse plugin list XML file {0}"), file.getAbsolutePath());
            throw new Exception(message,e);
        }    
    }
    
    /**
     * Scan a plugin directory for pluginlist files and attempt to parse
     * any such files that are found.
     *  
     * @param pluginDirectory the plugin directory to scan
     * @throws Exception
     */
    void scanPluginDirectory(File pluginDirectory) throws Exception {
        File[] pluginFiles=pluginDirectory.listFiles();
        
        if(pluginFiles==null) {
        	String message=MessageFormat.format(L10N.translate("no files found in directory {0}"), pluginDirectory.getAbsolutePath());
       		State.getInstance().logger().warning(message);
            return;
        }
        
        for(int i=0;i<pluginFiles.length;i++) {
            if(pluginFiles[i].isDirectory()) {
                scanPluginDirectory(pluginFiles[i]);
            }
            if(!pluginFiles[i].getName().endsWith(".pluginlist")) {
            		continue;
            }
            parsePluginList(pluginFiles[i]);
        }
    }

    /**
     * Scan the default plugin directory and any plugin paths given in the
     * user preferences for pluginlist files and attempt to parse
     * any such files that are found.
     *  
     * @throws Exception
     */
	void scanPluginDirectory() throws Exception {
		State state=State.getInstance();
		
        // scan startup script directory/directories
        File pluginDirectory=new File(state.getInstallationDirectory() + File.separator + "plugins");
        scanPluginDirectory(pluginDirectory);
        
        File pathListFile=new File(state.getPreferencesManager().getUserPreferencesPath() + "pluginpaths");
        if(pathListFile.exists() && pathListFile.canRead()) {
        	RandomAccessFile pathList=new RandomAccessFile(state.getPreferencesManager().getUserPreferencesPath() + "pluginpaths","r");
        	String path;
        	while((path=pathList.readLine())!=null) {
        		pluginDirectory=new File(path);
        		scanPluginDirectory(pluginDirectory);        	
        	}
        	pathList.close();
        }
    }
	
	/**
	 * Instantiate all the Plugin implementations of the plugins. This step
	 * is required before plugins can be queried for their dependencies.
	 */
	void instantiatePlugins() {
		for(int i=0;i<plugins.size();i++) {
			PluginWrapper wrapper=plugins.get(i);
			boolean instantiateable=false;
			try {
				wrapper.instantiate();
				instantiateable=true;
	        } catch (InstantiationException e) {
	        	e.printStackTrace();
	        } catch (IllegalAccessException e) {
	        	e.printStackTrace();
			}
	        if(!instantiateable) {
                // log this
	        	String message=MessageFormat.format(L10N.translate("removing plugin {0} because Plugin class instance creation failed"), plugins.get(i).plugin.getName());
                State.getInstance().logger().info(message);
                plugins.remove(i);
                i--; // decrement counter because of index shift caused by remove
	        }
		}
	}
    
    /**
     * Remove those plugins from the plugin list which have unsatisfied dependencies.
     */
    void kickUnsatisfiedDependencies()
    {
        boolean dependenciesSatisfied=false;
        
        while(!dependenciesSatisfied) {
            dependenciesSatisfied=true;
            
            for(int i=0;i<plugins.size();i++) {
                if(!plugins.get(i).hasSatisfiesDependencies()) {
                    // log this
                	String message=MessageFormat.format(L10N.translate("removing plugin {0} because of unsatisfied dependencies"), plugins.get(i).plugin.getName());
                    State.getInstance().logger().info(message);
                    plugins.remove(i);
                    i--; // decrement counter because of index shift caused by remove
                    dependenciesSatisfied=false;
                }
            }
        }
    }
    
    /**
     * Actually initialises plugins up one after another while respecting their
     * interdependencies.
     * 
     * @throws Exception
     */

    void initializePlugins()
    {
        boolean pluginsSkipped=true;
        
        while(pluginsSkipped) {
            pluginsSkipped=false;
            for(int i=0;i<plugins.size();i++) {
                if(!plugins.get(i).isLoaded()) {
                    if(plugins.get(i).isLoadable()) {
                        // log this
                        State.getInstance().logger().info("loading plugin " + plugins.get(i).plugin.getName());
                        try {
                        		plugins.get(i).load();
                        } catch(Throwable e) {
                        	String message=MessageFormat.format(L10N.translate("unable to initialize plugin {0}, reason:"), plugins.get(i).plugin.getName());
                            State.getInstance().logger().info(message);
                            State.getInstance().logger().info(ExceptionHelper.getFullText(e));
                            plugins.get(i).markUnloadable();
                        }
                    } else if(!plugins.get(i).isUnloadable()){
                        pluginsSkipped=true;
                    }
                }
            }
        }
    }
    
    /**
     * Load any plugisn from the default plugin directory within the 
     * application directory and any plugin directories specified in the
     * user preferences.
     * 
     * @throws Exception 
     */
    public void loadPlugins() throws Exception
    {
        try {
//            parsePluginList();
        	scanPluginDirectory();
        	instantiatePlugins();
            kickUnsatisfiedDependencies();
            initializePlugins();
        } catch(java.lang.Throwable e) {
            throw new Exception(L10N.translate("aborting plugin loading due to an exception which occured in that process"),e);
        }
    }
    
    /**
     * Save plugin path list from preferences in a separate file, which is read on startup
     * to get this directory list because the preferences aren't yet loaded that early during
     * startup.
     * 
     * @throws IOException
     */
    private void savePluginPathList() throws IOException {
    	State state=State.getInstance();
    	
    	File pathListFile=new File(state.getPreferencesManager().getUserPreferencesPath()+ "pluginpaths");
    	RandomAccessFile pathList=new RandomAccessFile(pathListFile,"rw");
    	pathList.setLength(0);
    	
        // check if preferences group exists (it won't be present during first startup)
    	Preferences preferences=state.getPreferencesManager().getCurrentPreferences();
        if(preferences.getGroupByName("Plugins")==null) {
        	pathList.close();
        	return;
        }
        Property paths=preferences.getGroupByName("Plugins").getProperty("paths");
        for(int i=0;i<paths.getSize();i++) {
        	String path=paths.getArrayStringValue(i);
        	pathList.writeBytes(path);
        }    	
        pathList.close();
    }
    
    /**
     * Unload any plugins that were previously loaded.
     * 
     * @throws Exception
     */
    public void unloadPlugins() throws Exception
    {
        try {
            boolean pluginsSkipped=true;
            while(pluginsSkipped) {
                pluginsSkipped=false;
                for(int i=0;i<plugins.size();i++) {
                    if(plugins.get(i).hasRuntimeDependencies()) {
                        pluginsSkipped=true;
                    } else {
                        if(plugins.get(i).isLoaded()) {
                        	String message=MessageFormat.format(L10N.translate("unloading plugin {0}"), plugins.get(i).plugin.getName());
                            eu.moonlight3d.framework.State.getInstance().logger().info(message);
                            plugins.get(i).unload();
                        }
                    }
                }
            }
        } catch (java.lang.Throwable e) {
            throw new Exception(L10N.translate("aborting plugin unloading due to an exception which occured in that in that process"),e);
        }
        savePluginPathList();
    }
    
    /**
     * Get the names of all known plugins. This list includes plugins that have
     * not been loaded successfully.
     * 
     * @return list of plugins
     */
    public List<String> getPluginList()
    {
        ArrayList<String> pluginList=new ArrayList<String>();
        
        for(int i=0;i<plugins.size();i++) {
            pluginList.add(plugins.get(i).plugin.getName());
        }
        
        return pluginList;
    }
}
