/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 21, 2005
 */
package eu.moonlight3d.framework.scripts;
/**
 * \package eu.moonlight3d.framework.scripts
 * 
 * This is the management system for script interpreter plugins.
 * It is responsible for registering script interpreters and
 * providing interfaces through which scripts can be run 
 * independently of the interpreter involved. In particular,
 * this package is also responsible for running startup scripts.
 * 
 * This package is part of the core program.
 */

import java.io.File;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;

/**
 * This is the script manager. Each script language provider must register
 * with this manager. New Interpreter instances are created exclusively
 * through this class.
 * 
 * @author gregor
 */
public class Manager {
	/**
	 * Array of all registerd script languages
	 */
    private ArrayList<Language> languages;
    
    /**
     * Constructor.
     */
    public Manager()
    {
        languages=new ArrayList<Language>();
    }
    
    /**
     * Register a script language.
     * 
     * @param language the main interface for this language
     * @see Language
     */
    public void registerLanguage(Language language)
    {
        languages.add(language);
    }
    
    /**
     * unegister a script language.
     * 
     * @param name the name of the script language as given by its Language interface
     * @throws Exception if no language was registerd with this name
     * @see Language
     */
    public void unregisterLanguage(String name) throws Exception
    {
        for(int i=0;i<languages.size();i++) {
            if(languages.get(i).getLanguageName().equals(name)) {
                languages.remove(i);
                return;
            }
        }
        
        throw new Exception(L10N.translate("unable to find language with given name while attempting to unregister"));
    }

    /**
     * Return a list of names of the languages for which there
     * are interpreters available at the moment.
     * 
     * @return list of script language names
     */
    public List<String> getLanguages()
    {
        ArrayList<String> languageNames=new ArrayList<String>();
        
        for(int i=0; i<languages.size();i++) {
            languageNames.add(languages.get(i).getLanguageName());
        }
        
        return languageNames;
    }
    
    /**
     * Create and return a new Interpreter for the script language with
     * the given name.
     * 
     * @param languageName the name of the script language
     * @return a new Interpreter
     * @throws Exception if the script language name is unknown
     */
    public Interpreter createInterpreter(String languageName) throws Exception
    {
        for(int i=0;i<languages.size();i++) {
            if(languages.get(i).getLanguageName().equals(languageName)) {
                return languages.get(i).createInterpreter();
            }
        }
        
        throw new Exception(L10N.translate("unable to find language with given name while attempting to create an interpreter"));
    }
    
    /**
     * Create and return a new Interpreter for the script with the given
     * file name.
     * 
     * @param filename the file to look for an interpreter for
     * @return a new Interpreter ready to run the given file
     * @throws Exception
     */
    public Interpreter createInterpreterForFile(String filename) throws Exception
    {
        for(int i=0;i<languages.size();i++) {
            if(languages.get(i).canExecute(filename)) {
                return languages.get(i).createInterpreter();
            }
        }
        
        throw new Exception(L10N.translate("unable to find an interpreter suitable for the given file"));        
    }
    
    /**
     * Internal Method. Run startup scripts. This is only called during
     * program initialisation.
     * 
     * @see ml.ML3D
     */
    public void runStartupScripts()
    {
        // scan startup script directory/directories
        File scriptDirectory=new File("scripts/startup");
        File[] scriptFiles=scriptDirectory.listFiles();
        
        if(scriptFiles==null) {
            return;
        }
        
        for(int i=0;i<scriptFiles.length;i++) {
            if(scriptFiles[i].isDirectory()) {
                continue;
            }
            String message=MessageFormat.format(L10N.translate("trying to execute startup script {0}"), scriptFiles[i].getAbsolutePath());
            State.getInstance().logger().info(message);
            try {
                Interpreter interpreter=createInterpreterForFile(scriptFiles[i].getAbsolutePath());
                
                interpreter.runFile(scriptFiles[i].getAbsolutePath());
            } catch (Exception e) {
                // TODO log this more elaborately
                e.printStackTrace();
            }
        }
    }
}
