/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 14, 2005
 */

package eu.moonlight3d.framework.ui;

import java.lang.management.ManagementFactory;
import java.lang.management.MemoryMXBean;
import java.lang.management.MemoryUsage;
import java.util.ArrayList;
import java.util.List;


import com.trolltech.qt.core.Qt;
import com.trolltech.qt.gui.QDialog;
import com.trolltech.qt.gui.QHBoxLayout;
import com.trolltech.qt.gui.QLabel;
import com.trolltech.qt.gui.QListWidget;
import com.trolltech.qt.gui.QPushButton;
import com.trolltech.qt.gui.QTabWidget;
import com.trolltech.qt.gui.QVBoxLayout;
import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;

/**
 * Displays a simple modal About dialog for Moonlight|3D.
 * 
 * @author gregor
 */
public class AboutDialog extends QDialog {
	/**
	 * The tab widget that contains all dialog pages.
	 */
    QTabWidget tab;
    /**
     * The main page container.
     */
    QWidget mainPage;
    /**
     * The OK button of the dialog.
     */
    QPushButton okButton;
    /**
     * The plugin page container.
     */
    QWidget pluginsPage;
    /**
     * The listbox containing the list of loaded plugins.
     */
    QListWidget pluginList;
    /**
     * The framework dialog page container.
     */
    QWidget frameworkPage;
    /**
     * The memory usage dialog page container.
     */
    QWidget memoryPage;
    
    /**
     * Dialog page factory describing the application.
     */
    private static AboutDialogPageFactory mainPageFactory;
    
    /**
     * List of additional dialog page factories.
     */
    private static ArrayList<AboutDialogPageFactory> pageFactories;
    
    static {
    	pageFactories=new ArrayList<AboutDialogPageFactory>();
    }
    
    /**
     * Set the factory for the main dialog page which should describe the main
     * application.
     * 
     * @param factory the dialog page factory
     */
    public static void setMainAboutDialogPageFactory(AboutDialogPageFactory factory) {
    	mainPageFactory=factory;
    }
    
    /**
     * Register an about dialog page factory for some additional about information.
     * 
     * @param pageFactory the factory to register
     * 
     * @see AboutDialogPageFactory
     */
    public static void registerAboutDialogPageFactory(AboutDialogPageFactory pageFactory) {
    	pageFactories.add(pageFactory);
    }

    /**
     * Unregister an about dialog page factory.
     * 
     * @param pageFactory the factory to unregister.
     * 
     * @see AboutDialogPageFactory
     */
    public static void unregisterAboutDialogPageFactory(AboutDialogPageFactory pageFactory) {
    	pageFactories.remove(pageFactory);
    }

    /**
     * Constructor.
     * 
     * @param parent The parent for this (modal) dialog
     */
    public AboutDialog(QWidget parent)
    {
        super(parent);
        setWindowTitle("About " + eu.moonlight3d.framework.State.getInstance().getVersionInformation().getApplicationName());
        
        tab=new QTabWidget(parent);
        okButton=new QPushButton(this);
        okButton.setText(L10N.translate("OK"));
        okButton.clicked.connect(this,"close()");

        QVBoxLayout tabLayout=new QVBoxLayout(this);
        tabLayout.addWidget(tab,1);
        tabLayout.addWidget(okButton);
        
        setLayout(tabLayout);

        if(mainPageFactory!=null) {
        	mainPage=new QWidget(tab);
        	tab.addTab(mainPage, mainPageFactory.getName());
        	mainPageFactory.generateAboutDialogPage(mainPage);
        }

        pluginsPage=new QWidget(tab);
        tab.addTab(pluginsPage, L10N.translate("Plugins"));
        createPluginsPage(pluginsPage);
        
        frameworkPage=new QWidget(tab);
        tab.addTab(frameworkPage, L10N.translate("Framework"));
        createFrameworkPage(frameworkPage);
        
        memoryPage=new QWidget(tab);
        tab.addTab(memoryPage, L10N.translate("Memory usage"));
        createMemoryPage(memoryPage);
        
        for(AboutDialogPageFactory pageFactory : pageFactories) {
        	QWidget page=new QWidget(tab);
        	tab.addTab(page, pageFactory.getName());
        	pageFactory.generateAboutDialogPage(page);
        }
    }
    
    /**
     * Create the dialog page for the framework itself.
     *  
     * @param parent the parent widget for this dialog page
     */
    public void createFrameworkPage(QWidget parent)
    {
        QLabel label=new QLabel(parent);
        String frameworkIdentifier=State.getInstance().getFrameworkVersionInformation().getApplicationVersionString();
        label.setText(frameworkIdentifier + "\n(C) 2003-2009 The Moonlight Team\n\nGregor Mueckl\nPeter Brinkler");
    	QVBoxLayout layout=new QVBoxLayout(parent);
    	layout.addWidget(label,1,Qt.AlignmentFlag.AlignCenter);
        parent.setLayout(layout);
    }
    
    /**
     * Create a dialog page listing all plugins that are currently loaded.
     * 
     * @param parent the parent widget for this dialog page.
     */
    public void createPluginsPage(QWidget parent)
    {
        pluginList=new QListWidget(parent);

        List<String> pluginList=eu.moonlight3d.framework.State.getInstance().getPluginManager().getPluginList();
        for(int i=0;i<pluginList.size();i++) {
            this.pluginList.addItem(pluginList.get(i));
        }        
    	QHBoxLayout layout=new QHBoxLayout(parent);
    	layout.addWidget(this.pluginList,1);
        parent.setLayout(layout);
    }

    /**
     * Create a dialog page giving memory usage information.
     * 
     * @param parent the parent widget for this dialog page.
     */
    public void createMemoryPage(QWidget parent)
    {
        MemoryMXBean memoryBean=ManagementFactory.getMemoryMXBean();
        MemoryUsage heapUsage=memoryBean.getHeapMemoryUsage();
        
        QLabel memoryLabel=new QLabel(parent);
        String memoryLabelText;
        memoryLabelText=L10N.translate("max. heap memory: ") + heapUsage.getMax() + "\n";
        memoryLabelText+=L10N.translate("committed heap memory: ") + heapUsage.getCommitted() + "\n";
        memoryLabelText+=L10N.translate("used heap memory: ") + heapUsage.getUsed();
        memoryLabel.setText(memoryLabelText);
    	QVBoxLayout layout=new QVBoxLayout(parent);
    	layout.addWidget(memoryLabel,1,Qt.AlignmentFlag.AlignCenter);
        parent.setLayout(layout);
    }
}