/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Sep 2, 2006
 */
package eu.moonlight3d.framework.ui;

import java.util.ArrayList;

/**
 * Action generators dynamically generate lists of actions to be inserted
 * into the menu bar.  This is useful when the contents of a menu
 * is required to be dynamic (a listing user defined bookmarks, for example).
 * 
 * Action generators can be executed either once when the menu bar is created
 * or every time a menu is about to be shown for which the action generator
 * provides entries. The default update policy is to only update once.
 *
 * @author gregor
 */
public abstract class ActionGenerator {
	
	/**
	 * Possible update policies for the action generator.
	 */
	public enum UpdatePolicy {
		/**
		 * Update only once when menu bar is shown
		 */
		Once,
		/**
		 * Update every time the menu is displayed to the user
		 */
		Always
	}
	
	/**
	 * The unique name of the action generator
	 */
	private String name;
	/**
	 * 
	 */
	private UpdatePolicy updatePolicy;
	
	/**
	 * Create a new action generator with the given name. This sets the
	 * update policy to once when the menu bar is generated.
	 * 
	 * @param name unique name of the action generator
	 */
	public ActionGenerator(String name) {
		this.name=name;
		this.updatePolicy=UpdatePolicy.Once;
	}
	
	/**
	 * Create a new action generator with the given name
	 * 
	 * @param name unique name of the action generator
	 * @param updatePolicy the update policy to use
	 */
	public ActionGenerator(String name, UpdatePolicy updatePolicy) {
		this.name=name;
		this.updatePolicy=updatePolicy;
	}
	
	/**
	 * Get the unique name of the action generator.
	 * 
	 * @return unique name
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Get the update policy of the action generator
	 * 
	 * @return update policy
	 */
	public UpdatePolicy getUpdatePolicy() {
		return updatePolicy;
	}
	
	/**
	 * Return an ordered list of actions that is inserted into the
	 * menu in order.
	 *  
	 * @return ordered list of actions to insert
	 */
	public abstract ArrayList<Action> getActions();
}
