/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jul 6, 2005
 */
package eu.moonlight3d.framework.ui;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Set;

import com.trolltech.qt.core.QByteArray;
import com.trolltech.qt.core.QEvent;
import com.trolltech.qt.core.QObject;
import com.trolltech.qt.gui.QAction;
import com.trolltech.qt.gui.QCloseEvent;
import com.trolltech.qt.gui.QDockWidget;
import com.trolltech.qt.gui.QIcon;
import com.trolltech.qt.gui.QMainWindow;
import com.trolltech.qt.gui.QMenu;
import com.trolltech.qt.gui.QStatusBar;
import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;

/**
 * A main window instance in the Moonlight framework. Instances
 * are instances of the window definitions found in {@link FrameLayout}
 * classes. Menu bar entries, main widgets dock widgets and a
 * status bar are automatically created as defined in the FrameLayout.
 *
 * @author gregor
 */
public class ApplicationWindow extends QMainWindow {
	/**
	 * A complete copy of the frame layout definition that was 
	 * used to create this window.
	 */
    private FrameLayout frameLayout;
    
    /**
     * Map between dock widget names and concrete instances.
     */
    private HashMap<String,QDockWidget> dockWidgets;
    
    /**
     * Constructor that builds a new application window from the
     * given frame layout.
     * 
     * @param layout the layout to instanticate
     */
    public ApplicationWindow(FrameLayout layout)
    {
        super();
        setWindowTitle(layout.getTitle());
        setWindowIcon(new QIcon(State.getInstance().getInstallationDirectory() + "/icons/windowicon.png"));
        
        this.frameLayout=layout;
        dockWidgets=new HashMap<String, QDockWidget>();

        UIManager uiManager=eu.moonlight3d.framework.State.getInstance().getUIManager();

        QWidget toplevel=uiManager.getMainWidget(this, layout.getName());
        setCentralWidget(toplevel);
        
        rebuildMenuBar();
        
        setStatusBar(new QStatusBar());
        
        createDockWidgets();
        
        // restore main window state
        try {
            File stateFile=new File(State.getInstance().getPreferencesManager().getUserPreferencesPath() + frameLayout.getName() + ".state");
            FileInputStream input=new FileInputStream(stateFile);
            byte[] tempData=new byte[input.available()];
			input.read(tempData);
	        QByteArray data=new QByteArray(tempData);
	        restoreState(data);
		} catch (IOException e) {
			/* Do nothing here because file is not guaranteed to exist or be correct.
			 * This results in the use of the default layout, which should be done
			 * silently.
			 */
		}

        uiManager.registerApplicationWindow(this);
    }
    
    /**
     * This is an action generator wrapper class which is used to 
     * implement the "Always" update policy for action generators.
     * This wrapper connects to the aboutToShow slot on the generator's
     * parent menu. When this slot is triggered,  it removes the 
     * previously generated actions, forces the generator to provide a 
     * new list of actions and finally inserts the actions into the
     * parent menu at the correct position.
     *
     * @author gregor
     */
    private class ActionGeneratorWrapper extends QObject {
    	/**
    	 * The menu that the generated actions are added to
    	 */
    	private QMenu menu;
    	/**
    	 * An invisible action in the menu which serves as a marker for
    	 * the place where the regenerated actions need to be inserted.
    	 */
    	private QAction lastAction;
    	/**
    	 * The action generator that is wrapped.
    	 */
    	private ActionGenerator generator;
    	/**
    	 * The list of actions that was generated by the action generator
    	 * the last time it was invoked.
    	 */
    	private List<Action> actions;
    	
    	/**
    	 * Constructor for the wrapper.
    	 * 
    	 * @param menu the parent menu for the generator
    	 * @param generator the generator that is to be wrapped
    	 */
    	public ActionGeneratorWrapper(QMenu menu, ActionGenerator generator) {
    		super(menu);
    		this.menu=menu;
    		this.generator=generator;
    		
    		// create an invisible "last action" as insertion marker for insertAction below
    		lastAction=new QAction(menu);
    		lastAction.setVisible(false);
    		menu.addAction(lastAction);
    		
    		menu.aboutToShow.connect(this,"updateActions()");
    	}
    	
    	/**
    	 * The slot used to update the actions in the menu when the  menu is about
    	 * to be shown.
    	 */
    	public void updateActions() {
    		if(actions!=null) {
    			// remove old actions from menu
    			for(Action action : actions) {
    				menu.removeAction(action);
    			}
    		}
    		// generate new actions
    		actions=generator.getActions();
    		// add new actions to menu
    		for(Action action : actions) {
    			menu.insertAction(lastAction, action);
    		}
    	}
    }
    
    /**
     * Build the menu bar. This descends recursively into the given
     * {@link MenuNode} tree and instantiates Qt menu items from that.
     * 
     * @param menu the menu definition to instantiate
     * @param parent the parent menu entry 
     */
    private void buildMenu(MenuNode menu, QMenu parent) {
    	for(MenuNode child : menu.getChildren()) {
    		if(child.getType()==MenuNode.Type.Menu) {
    			// add submenu and recurse
    			QMenu childMenu=new QMenu(child.getName(),parent);
    			childMenu.setTearOffEnabled(true);
    			buildMenu(child,childMenu);
    			parent.addMenu(childMenu);
    		} else if(child.getType()==MenuNode.Type.ExtensionPoint) {
    			// process contributions as if they were at the current menu level
    			for(MenuNode contribution : child.getChildren()) {
   					buildMenu(contribution,parent);
    			}
    		} else if(child.getType()==MenuNode.Type.Entry) {
    			// retrieve action and append it
    			QAction action;
    			try {
					action=eu.moonlight3d.framework.State.getInstance().getUIManager().getActionByName(child.getActionName());
				} catch (Exception e) {
					// build a greyed out default action
					action=new QAction(child.getName(),parent);
					action.setDisabled(true);
				}
				action.setText(child.getName());
				parent.addAction(action);
    		} else if(child.getType()==MenuNode.Type.Generator) {
    			// retrieve generator and append generated actions
    			try {
        			ActionGenerator generator;
					generator=eu.moonlight3d.framework.State.getInstance().getUIManager().getActionGeneratorByName(child.getName());
					
					if(generator.getUpdatePolicy()==ActionGenerator.UpdatePolicy.Once) {
						for(Action action : generator.getActions()) {
							parent.addAction(action);
						}
					} else if(generator.getUpdatePolicy()==ActionGenerator.UpdatePolicy.Always) {
						// the wrapper class is a QObject with the parent menu as parent, so no need to keep the reference here:
						@SuppressWarnings("unused")
						ActionGeneratorWrapper wrapper=new ActionGeneratorWrapper(parent,generator);
					}
				} catch (Exception e) {
					// build a greyed out default action
					QAction defaultAction=new QAction(child.getName(),parent);
					defaultAction.setDisabled(true);
					parent.addAction(defaultAction);
				}
    		} else if(child.getType()==MenuNode.Type.Separator) {
    			parent.addSeparator();
    		}
    	}    	
    }

    /**
     * Rebuild the menu bar. This descends recursively into the given
     * {@link MenuNode} tree and instantiates Qt menu items from that.
     * 
     * @param menu the menu definition to instantiate 
     */
    private void rebuildMenuBar(MenuNode menu) {
    	for(MenuNode child : menu.getChildren()) {
    		if(child.getType()==MenuNode.Type.Menu) {
    			// add submenu and recurse
    			QMenu childMenu=new QMenu(child.getName(),menuBar());
    			childMenu.setTearOffEnabled(true);
    			buildMenu(child,childMenu);
    			menuBar().addMenu(childMenu);
    		} else if(child.getType()==MenuNode.Type.ExtensionPoint) {
    			// process contributions as if they were at the current menu level
    			for(MenuNode contribution : child.getChildren()) {
    				rebuildMenuBar(contribution);
    			}
    		} else if(child.getType()==MenuNode.Type.Entry) {
    			// retrieve action and append it
    			QAction action;
    			try {
					action=eu.moonlight3d.framework.State.getInstance().getUIManager().getActionByName(child.getActionName());
				} catch (Exception e) {
					// build a greyed out default action
					action=new QAction(child.getName(),menuBar());
					action.setDisabled(true);
				}
				menuBar().addAction(action);
    		} else if(child.getType()==MenuNode.Type.Generator) {
    			// retrieve generator and append generated actions
    			try {
        			ActionGenerator generator;
					generator=eu.moonlight3d.framework.State.getInstance().getUIManager().getActionGeneratorByName(child.getName());
					for(Action action : generator.getActions()) {
						menuBar().addAction(action);
					}
				} catch (Exception e) {
					// build a greyed out default action
					QAction defaultAction=new QAction(child.getName(),menuBar());
					defaultAction.setDisabled(true);
					menuBar().addAction(defaultAction);
				}
    		} else if(child.getType()==MenuNode.Type.Separator) {
    			menuBar().addSeparator();
    		}
    	}      	
    }

    /**
     * Rebuild the menu bar. This needs to be called to update the menu
     * bar to make changes in the output of any {@link ActionGenerator}
     * instances visible.
     */
    public void rebuildMenuBar() {
    	menuBar().clear();
    	
    	MenuNode menu=frameLayout.getMenu();
    	rebuildMenuBar(menu);
    }
    
    /**
     * Create dock widgets from frame layout definition.
     */
    private void createDockWidgets() {
    	for(String dockWidgetName : frameLayout.getDockWidgets()) {
    		DockWidget contentsWidget=eu.moonlight3d.framework.State.getInstance().getUIManager().getDockWidget(dockWidgetName);
    		
    		if(contentsWidget==null)  {
    			State state=State.getInstance();
    			state.logger().warning(L10N.translate("dock factory not found - dock not available: ") + dockWidgetName);
    			continue;
    		}
    		
    		QDockWidget dockWidget=new QDockWidget(contentsWidget.getTitle(),this);
    		dockWidget.setWidget(contentsWidget);
    		dockWidget.setAllowedAreas(contentsWidget.getAllowedAreas());
    		dockWidget.setFeatures(contentsWidget.getFeatures());
    		dockWidget.setObjectName(dockWidgetName);
    		addDockWidget(contentsWidget.getDefaultArea(),dockWidget);
    		dockWidgets.put(dockWidgetName, dockWidget);
    	}
    }
    
    /**
     * Get a list of all dock widget titles. These titles are used
     * to refer to these widgets.
     * 
     * @return list of dock widget names
     */
    public Set<String> getDockWidgets() {
    	return dockWidgets.keySet();
    }

    /**
     * Check if a specific dock widget is set to be hideable.
     * 
     * @param widgetName the name of the dock widget to check
     * @return true if hideable, false otherwise
     */
    public boolean isHideable(String widgetName) {
    	QDockWidget widget = dockWidgets.get(widgetName);
    	
    	if(widget==null) {
    		return false;
    	}
    	
    	return widget.features().isSet(QDockWidget.DockWidgetFeature.DockWidgetClosable);
    }
    
    /**
     * Check if a specific dock widget is currently visible.
     * 
     * @param widgetName the name of the dock widget to check
     * @return true if visible, false if hidden
     */
    public boolean isVisible(String widgetName) {
    	QDockWidget widget = dockWidgets.get(widgetName);
    	
    	if(widget==null) {
    		return false;
    	}
    	
    	return widget.isVisible();
    }
    
    /**
     * Toggle dock widget visibility for the dock widget with the
     * given title.
     * 
     * @param widgetName the widget to toggle
     */
    public void toggleDockWidgetVisibility(String widgetName) {
    	QDockWidget widget = dockWidgets.get(widgetName);
    	
    	if(widget==null) {
    		return;
    	}
    	
    	if(widget.isVisible()) {
    		widget.hide();
    	} else {
    		widget.show();
    	}
    }
    
    @Override
    public boolean event(QEvent event) {
    	if(event.type()==QEvent.Type.WindowActivate) {
    		eu.moonlight3d.framework.State.getInstance().getUIManager().setActiveApplicationWindow(this);
    	}
    	return super.event(event);
    }

    @Override
	public void closeEvent(QCloseEvent event) {
		eu.moonlight3d.framework.State.getInstance().getUIManager().unregisterApplicationWindow(this);
		// replace layout in UIManager with this layout
		State.getInstance().getUIManager().addOrReplaceLayout(frameLayout);

		// save main window state
        try {
            File stateFile=new File(State.getInstance().getPreferencesManager().getUserPreferencesPath() + frameLayout.getName() + ".state");
            stateFile.delete();
            FileOutputStream output= new FileOutputStream(stateFile);
	        QByteArray data=saveState();
			output.write(data.toByteArray());
		} catch (IOException e) {
			/* Do nothing here because a failure to save the dock widget state
			 * only means that the layout state will be reset to defaults the
			 * next time the window is opened.
			 */
		}

		disposeLater();
    }
    
}
