/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jul 14, 2005
 */
package eu.moonlight3d.framework.ui;

import java.util.ArrayList;
import java.util.List;


/**
 * Container class defining the contents of an {@link ApplicationWindow}.
 * It holds the internal name of the frame definition (wich also serves
 * as the internal name of the application window), the visible title of
 * the application window, the menu bar contents and the list of dock
 * widgets to instantiate.
 *
 * @author gregor
 */
public class FrameLayout implements Cloneable {
	/**
	 * Root {@link MenuNode} for the menu bar structure.
	 */
    private MenuNode menu;
    /**
     * The name of the frame layout definition
     */
    private String name;
    /**
     * The displayed window title for the defined window
     */
    private String title;
    /**
     * A list of dock widget factory names from which the
     * dock widgets for this window are instantiated.
     */
    private ArrayList<String> dockWidgets;
    
    /**
     * Constructor taking internal name and window title for frame layout.
     * Except from a root level menu extension point named "menubar" this
     * frame definition is empty. 
     * 
     * @param name the internal name for the frame layout
     * @param title the window title for the frame layout
     */
    public FrameLayout(String name, String title)
    {
        this.name=name;
        this.title=title;
        
        // the basic menu structure is a top level Menu with an ExtensionPoint of the same name
        menu=new MenuNode(MenuNode.Type.Menu, "menubar", "");
        menu.add(new MenuNode(MenuNode.Type.ExtensionPoint,"menubar",""));
        
        dockWidgets=new ArrayList<String>();
    }
    
    /**
     * Add a menu contribution. The menu contribution is added to the menu bar at
     * the appropriate place.
     * 
     * @param menuContribution the menu contribution to be added
     * @throws Exception if the menu node passed to this function is not a contribution
     */
    public void addMenuContribution(MenuNode menuContribution) throws Exception {
    	// add menu contribution: find extension point, insert MenuNodes
		if(menuContribution.getType()!=MenuNode.Type.Contribution) {
			throw new Exception("passed menu node is not a contribution");
		}
    	menu.addContribution(menuContribution);
    }
    
    /**
     * Get the internal name of the frame layout
     * 
     * @return internal name
     */
    public String getName()
    {
        return name;
    }
    
    /**
     * Return the window title of the frame layout
     * 
     * @return window title
     */
    public String getTitle()
    {
        return title;
    }
    
    /**
     * Return the main menu structure of the frame layout
     * 
     * @return main menu structure
     */
    public MenuNode getMenu()
    {
        return menu;
    }
    
    /**
     * Add a dock widget to this frame layout definition by name.
     * 
     * @param name the name of the dock widget factory
     */
    public void addDockWidget(String name)
    {
    	dockWidgets.add(name);
    }
    
    /**
     * Get a list of dock widgets that should be added to this
     * window.
     * 
     * @return list of dock widget factory names.
     */
    public List<String> getDockWidgets()
    {
    	return (List<String>) dockWidgets.clone();
    }
}
