/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jul 15, 2005
 */
package eu.moonlight3d.framework.ui;
/**
 * \package eu.moonlight3d.framework.ui
 * 
 * This is the core user interface framework for Moonlight|3D.
 * It contains all the code to dynamically build views and menu bars
 * from the configuration files.
 * 
 * This package is part of the core program. 
 */

import java.util.ArrayList;
import java.util.List;

import com.trolltech.qt.gui.QSyntaxHighlighter;
import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.framework.helper.Factory;

/**
 * This is the central manager class for the user interface part
 * of the framework. It keeps track of various things like 
 * toplevel application windows, actions and main widget factories.
 * 
 * This class is instanced by {@link eu.moonlight3d.framework.State} during 
 * framework initialization and can be accessed through its
 * getUIManager() method.
 * 
 * @see ApplicationWindow
 * @see Action
 * @see ActionGenerator
 * @see MainWidgetFactory
 * @see FrameLayout
 *
 * @author gregor
 */
public class UIManager {
	
	/**
	 * List of all registered frame layouts.
	 */
    private ArrayList<FrameLayout> layouts;
    
    /**
     * List of all application windows that currently exist. 
     */
    private ArrayList<ApplicationWindow> windows;
    
    /**
     * List of all registered main widget factories.
     */
	private ArrayList<MainWidgetFactory> mainWidgetFactories;

	/**
	 * List of all registered dock widget factories.
	 */
	private ArrayList<DockWidgetFactory> dockWidgetFactories; 
	
	/**
	 * List of all registered actions.
	 */
	private ArrayList<Action> actions;
	
	/**
	 * List of all registered action generators.
	 */
	private ArrayList<ActionGenerator> actionGenerators;
	
	/**
	 * List of all registered syntax highlighter factories.
	 */
	private ArrayList<Factory<QSyntaxHighlighter> > syntaxHighlighterFactories;
	
	/**
	 * Currently active application window.
	 */
	private ApplicationWindow activeWindow;
	
	/**
	 * Internal constructor - do not use outside the framework.
	 */
    public UIManager()
    {
        layouts=new ArrayList<FrameLayout>();
        windows=new ArrayList<ApplicationWindow>();
        mainWidgetFactories=new ArrayList<MainWidgetFactory>();
        dockWidgetFactories=new ArrayList<DockWidgetFactory>();
		actions=new ArrayList<Action>();
		actionGenerators=new ArrayList<ActionGenerator>();		
		syntaxHighlighterFactories=new ArrayList<Factory<QSyntaxHighlighter>>();
    }
    
    /**
     * Add a frame layout definition or replace an existing one with the same
     * name.
     * 
     * @param newLayout the new layout
     */
    public void addOrReplaceLayout(FrameLayout newLayout) {
    	for(FrameLayout layout : layouts) {
    		if(layout.getName().equals(newLayout.getName())) {
    			layouts.remove(layout);
    			break;
    		}
    	}
    	layouts.add(newLayout);
    }
    
    /**
     * Find and return the frame layout with the given name.
     *  
     * @param name the name of the frame layout to find.
     * @return the requested frame layout
     * @throws Exception if the layout does not exist
     */
    public FrameLayout getLayout(String name) throws Exception
    {
        for(int i=0;i<layouts.size();i++) {
            if(layouts.get(i).getName().equals(name)) {
                return layouts.get(i);
            }
        }
        
        throw new Exception("unable to find frame layout with given name");
    }

    /**
     * Internal method used by newly created {@link ApplicationWindow} 
     * instance to register themselves.
     * 
     * @param window the newly created application window
     */
    void registerApplicationWindow(ApplicationWindow window) {
    	windows.add(window);
    }

    /**
     * Internal method used by closed {@link ApplicationWindow}
     * instances to unregister themselves.
     *  
     * @param window the closed application window
     */
    void unregisterApplicationWindow(ApplicationWindow window)  {
    	windows.remove(window);
    }
    
    /**
     * Get a list of all application windows that are currently open.
     * 
     * @return list of all open application windows
     */
    public List<ApplicationWindow> getApplicationWindows() {
    	return (ArrayList<ApplicationWindow>) windows.clone();
    }
    
    /**
     * Get the application window that currently holds the focus.
     * 
     * @return application window which holds the focus or null
     * if the application does not hold the focus right now.
     */
    public ApplicationWindow getActiveApplicationWindow() {
    	return activeWindow;
    }

    /**
     * Set the application window that currently holds the focus.
     * This is an internal method used by {@link ApplicationWindow}
     * to track the currently active window. 
     * 
     * @param applicationWindow the window that holds the focus
     */
	void setActiveApplicationWindow(ApplicationWindow applicationWindow) {
		activeWindow=applicationWindow;
	}
	
    /**
     * Register an {@link Action}.
     * 
     * @param action the action to register
     */
	public void registerAction(Action action) {
		actions.add(action);
	}
	
	/**
	 * Unregister an {@link Action}.
	 * 
	 * @param action the action to unregister.
	 */
	public void unregisterAction(Action action) {
		actions.remove(action);
	}
	
	/**
	 * Unregister an {@link Action} by name.
	 * 
	 * @param actionName the name of the action to unregister.
	 */
	public void unregisterAction(String actionName) {
		for(Action action : actions) {
			if(action.getName().equals(actionName)) {
				actions.remove(action);
				return;
			}
		}
	}
	
	/**
	 * Return a list of all registered actions.
	 * 
	 * @return list of all registered actions
	 */
	public List<Action> getActions() {
		return (ArrayList<Action>) actions.clone();
	}
	
	/**
	 * Get the action with the given name.
	 * 
	 * @param actionName the name of the action
	 * @return the action with the given name
	 * @throws Exception if the action does not exist
	 */
	public Action getActionByName(String actionName) throws Exception {
		for(Action action : actions) {
			if(action.getName().equals(actionName)) {
				return action;
			}
		}
		throw new Exception("No action named \"" + actionName + "\" registered");
	}
	
	/**
	 * Register an {@link ActionGenerator}.
	 * 
	 * @param generator the action generator to register.
	 */
	public void registerActionGenerator(ActionGenerator generator) {
		actionGenerators.add(generator);
	}
	
	/**
	 * Unregister an {@link ActionGenerator}.
	 * 
	 * @param generator the action generator to unregister.
	 */
	public void unregisterActionGenerator(ActionGenerator generator) {
		actionGenerators.remove(generator);
	}
	
	/**
	 * Unregister an {@link ActionGenerator} by name.
	 * 
	 * @param generatorName the action generator name
	 */
	public void unregisterActionGenerator(String generatorName) {
		for(ActionGenerator generator : actionGenerators) {
			if(generator.getName().equals(generatorName)) {
				actionGenerators.remove(generator);
				return;
			}
		}
	}
	
	/**
	 * Get a list of all registered action generators.
	 * 
	 * @return list of all registered action generators.
	 */
	public List<ActionGenerator> getActionGenerators() {
		return (ArrayList<ActionGenerator>) actionGenerators.clone();
	}
	
	/**
	 * Get the action generator with the given name.
	 * 
	 * @param generatorName the action generator name
	 * @return the {@link ActionGenerator} with the given name
	 * @throws Exception if no action generator with this name exists
	 */
	public ActionGenerator getActionGeneratorByName(String generatorName) throws Exception {
		for(ActionGenerator generator : actionGenerators) {
			if(generator.getName().equals(generatorName)) {
				return generator;
			}
		}
		throw new Exception("No action generator named \"" + generatorName + "\" registered");
	}

	/**
	 * Register a {@link MainWidgetFactory}.
	 * 
	 * @param factory the main widget factory to register
	 */
	public void registerMainWidgetFactory(MainWidgetFactory factory) {
	     mainWidgetFactories.add(factory);
	}

	/**
	 * Unregister the {@link MainWidgetFactory} with the given name. 
	 * 
	 * @param name the name of the {@link MainWidgetFactory}
	 */
	public void unregisterMainWidgetFactory(String name) {
	    for(int i=0;i<mainWidgetFactories.size();i++) {
	        if(mainWidgetFactories.get(i).getName().equals(name)) {
	            mainWidgetFactories.remove(i);
	        }
	    }
	}

	/**
	 * Create and return a new main widget from the factory with the
	 * given name and the given parent as the parent widget.
	 * 
	 * @param parent the parent widget to use
	 * @param name the name of the main widget factory to use
	 * @return the newly created widget
	 */
	public QWidget getMainWidget(QWidget parent, String name) {
	    for(int i=0;i<mainWidgetFactories.size();i++) {
	        if(mainWidgetFactories.get(i).getName().equals(name)) {
	        	MainWidgetFactory factory=mainWidgetFactories.get(i);
        		return factory.createNew(parent);
	        }
	    }
	    
	    return null;
	}
	
	/**
	 * Register a  {@link Factory} for dock widgets.
	 * 
	 * @param factory the dock widget factory to register
	 */
	public void registerDockWidgetFactory(DockWidgetFactory factory) {
		dockWidgetFactories.add(factory);
	}
	
	/**
	 * Unregister a dock widget {@link Factory}.
	 * 
	 * @param factory the factory which is to be unregistered.
	 */
	public void unregisterDockWidgetFactory(DockWidgetFactory factory) {
		dockWidgetFactories.remove(factory);
	}
	
	/**
	 * Unregister a dock widget {@link Factory} with the given name.
	 * 
	 * @param factoryName the name of the factory to unregister
	 */
	public void unregisterDockWidgetFactory(String factoryName) {
		for(Factory<DockWidget> dockWidgetFactory : dockWidgetFactories) {
			if(dockWidgetFactory.getName().equals(factoryName)) {
				dockWidgetFactories.remove(dockWidgetFactory);
				return;
			}
		}
	}
	
	/**
	 * Get a dock widget {@link Factory} by name.
	 * 
	 * @param name the name of the factory
	 * @return the factory or null if no factory with the given name exists
	 */
	public DockWidget getDockWidget(String name) {
	    for(int i=0;i<dockWidgetFactories.size();i++) {
	        if(dockWidgetFactories.get(i).getName().equals(name)) {
	        	Factory<DockWidget> factory=dockWidgetFactories.get(i);
        		return factory.createNew();
	        }
	    }

	    return null;
	}
	
	/**
	 * Get a dock widget {@link Factory} by name.
	 * 
	 * @param name the name of the factory
	 * @return the dock widget title or null if no factory with the given name exists
	 */
	public String getDockWidgetTitle(String name) {
	    for(int i=0;i<dockWidgetFactories.size();i++) {
	        if(dockWidgetFactories.get(i).getName().equals(name)) {
	        	DockWidgetFactory factory=dockWidgetFactories.get(i);
        		return factory.getTitle();
	        }
	    }

	    return null;
	}
	
	/**
	 * Register a syntax highlighter factory. Note that the factory
	 * name should be the name of the language which this hihglighter 
	 * processes.
	 * 
	 * @param factory the factory to register.
	 */
	public void registerSyntaxHighlighterFactory(Factory<QSyntaxHighlighter> factory) {
		syntaxHighlighterFactories.add(factory);
	}
	
	/**
	 * Unregister the syntax highlighter factory with the given name.
	 * 
	 * @param factoryName the name of the factory to unregister
	 */
	public void unregisterSyntaxHighlighterFactory(String factoryName) {
		for(Factory<QSyntaxHighlighter> factory : syntaxHighlighterFactories) {
			if(factory.getName().equals(factoryName)) {
				syntaxHighlighterFactories.remove(factory);
				return;
			}
		}
	}
	
	/**
	 * Return a newly created syntax highlighter for the given language.
	 * 
	 * @param language the language that is to be highlighted.
	 * @return the syntax highlighter or null if there is none for the given 
	 * language.
	 */
	public QSyntaxHighlighter getSyntaxHighlighter(String language) {
		for(Factory<QSyntaxHighlighter> factory : syntaxHighlighterFactories) {
			if(factory.getName().equals(language)) {
				return (QSyntaxHighlighter) factory.createNew();
			}
		}
		return null;
	}

}
